/*
 *  $id: $
 *
 *  Copyright (C) 1998 RealNetworks, Inc.
 *  All rights reserved.
 *  
 *  http://www.real.com/devzone
 *
 *  This program contains proprietary information of RealNetworks, Inc.,
 *  and is licensed subject to restrictions on use and distribution.
 *
 *  PlayerApp.java
 *
 *  Demonstrates how to create a standalone media player app.
 */

import java.awt.*;
import java.awt.event.*;
import java.net.*;
import java.math.*;

import javax.media.*;

/**
 * Standalone media player app.
 *
 * Creates a frame window containing a media player with visual component,
 * control panel and gain control, and caching controller and progress bar.
 * The frame window is resized to the size of the video clip and player controls.
 *
 * Example usage:
 * java PlayerApp pnm://video.real.com/welcome.rm
 */
public class PlayerApp
{
    private PlayerPanel playerPanel;
    
    public static void main(String argv[])
    {
        // If a URL argument was specified:
        if (argv.length >= 1)
        {
            // Get the url argument
            String url = argv[argv.length - 1];

            // Create the PlayerApp
            PlayerApp playerApp = new PlayerApp(url);
        }
        else
        {
            // Output the command line usage
            System.err.println("Usage: java PlayerApp <URL>");
        }
    }
            
    public PlayerApp(String url)
    {
        // Create and initialize the player panel
        playerPanel = new PlayerPanel(url);
        playerPanel.setSize(400, 300);
        playerPanel.init();

        // Start the player
        playerPanel.start();
        
        // Create the frame window
        Frame frame = new Frame("PlayerApp");

        // Set the background color
        frame.setBackground(Color.lightGray);

        // Make the frame non-resizable
        frame.setResizable(false);
        
        // Add the frame window event handler
        frame.addWindowListener(
            new WindowAdapter()
            {
                public void windowClosing(WindowEvent e)
                { 
                    // Stop the player
                    playerPanel.stop();
                    
                    // Destroy the player
                    playerPanel.destroy();
                    
                    // Exit the application
                    System.exit(0); 
                }
            }
        );
        
        // Add the player panel to the frame window
        frame.add("Center", playerPanel);

        // Size and show the frame window
        frame.pack();
        frame.show();
    }
}

/**
 * Encapsulates a media player.
 *
 * Handles creating and positioning the player's visual
 * components, and responding to media events.
 */
class PlayerPanel
    extends Panel implements ControllerListener
{
    private String          url;                        // Media clip URL
    private Player          player = null;              // Media player
    private GainControl     gainControl = null;         // Gain control
    private CachingControl  cachingControl = null;      // Caching control
    
    private Component   visualComponent = null;         // Displays the media clip
    private Component   controlComponent = null;        // Displays/controls clip playback
    private Component   gainComponent = null;           // Displays/controls audio gain
    private Component   progressBarComponent = null;    // Displays download progress

    public PlayerPanel(String url)
    {
        this.url = url;
    }
    
    public void init()
    {
        // Set the background color
        setBackground(Color.lightGray);
        
        // Remove the layout manager
        setLayout(null);

        // Create the media player
        createPlayer();
    }

    /**
     * Start the media player.  
     */
    public void start()
    {
        if (player != null)
        {
            // Set the wait cursor
            setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

            // Start the media file playback
            player.start();
        }
    }

    /**
     * Stop the media file playback and release all resources.
     */
    public void stop()
    {
        if (player != null)
        {
            // Stop the media file playback
            player.stop();

            // Release the player resources
            player.deallocate();
        }
    }

    /**
     * Stop the media file playback and destroy the media player.
     */
    public void destroy()
    {
        if (player != null)
        {
            // Stop the media file playback
            player.stop();

            // Destroy the media player
            player.close();
        }
    }

    /**
     * The controllerUpdate method implements the ControllerListener interface.
     * This method is called to handle all media events.
     */
    public synchronized void controllerUpdate(ControllerEvent event)
    {
        // Handle the RealizeCompleteEvent. 
        if (event instanceof RealizeCompleteEvent)
        {
            // Create the player visual components
            createPlayerControls();

            // Position and show the player visual components
            placePlayerControls();
        }

        // Handle the PrefetchCompleteEvent. 
        else if (event instanceof PrefetchCompleteEvent)
        {
            // Reset the default cursor
            setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
        }
        
        // Handle the ResourceUnavailableEvent
        else if (event instanceof ResourceUnavailableEvent)
        {
            // Hide the progress bar, if it exists.
            if (progressBarComponent != null)
            {
                progressBarComponent.setVisible(false);
            }
            
            // Reset the default cursor
            setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
        }
        
        // Handle the CachingControlEvent
        else if (event instanceof CachingControlEvent)
        {
            // Show/hide the progress bar component.
            if (progressBarComponent != null)
            {
                long percentComplete = ((CachingControlEvent) event).getContentProgress();
                if (percentComplete == 0)
                {
                    progressBarComponent.setVisible(true);
                }
                else if (percentComplete == 100)
                {
                    progressBarComponent.setVisible(false);
                }
            }
        }
    }

    /**
     * Create the media player from the URL given in the command line.
     */
    protected void createPlayer()
    {
        MediaLocator    mediaLocator = null;    // media locator for the media file

        try
        {
            // Create a media locator from the media URL string
            mediaLocator = new MediaLocator(url);

            // Create the proper media player for this media type.
            player = Manager.createPlayer(mediaLocator);

            // If a player was created:
            if (player != null)
            {
                // Add this applet as a listener for media player events.
                player.addControllerListener(this);
            }
            else
            {
                System.err.println("Could not create player for " + mediaLocator);
            }
        }
        // Handle MalformedURLException
        catch (MalformedURLException e)
        {
            System.err.println("Invalid media file URL: " + url);
        }
        // Handle NoPlayerException
        catch(NoPlayerException e)
        {
            System.err.println("Could not find a player for " + mediaLocator);
        }
        // Handle IOException
        catch(java.io.IOException e)
        {
            System.err.println("IO Exception encountered opening " + mediaLocator);
        }
    }

    /**
     * Create the player gain control and visual components
     */
    protected void createPlayerControls()
    {
        // Get the caching control
        cachingControl = (CachingControl) player.getControl("com.real.media.RMCachingControl");
        if (cachingControl != null)
        {
            // Get the progress bar component
            progressBarComponent = cachingControl.getProgressBarComponent();
            if (progressBarComponent != null)
            {
                // Add the component to the applet window
                add(progressBarComponent);
            }
        }
 
        // Get the gain control
        gainControl = player.getGainControl();
 
        if (gainControl != null)
        {
            // Get the gain control component
            gainComponent = gainControl.getControlComponent();
            if (gainComponent != null)
            {
                // Add the component to the applet window
                add(gainComponent);
            }
        }
        
        // Get the control panel component
        controlComponent = player.getControlPanelComponent();
        if (controlComponent != null)
        {
            // Add the component to the applet window
            add(controlComponent);
        }

        // Get the visual component
        visualComponent = player.getVisualComponent();
        if (visualComponent != null)
        {
            // Add the component to the applet window
            add(visualComponent);
        }
    }
    
    /**
     * Position the player components within the player panel
     */
    protected void placePlayerControls()
    {
        // Get the visual component size:
        Dimension sizeVideo = new Dimension(0, 0);
        if (visualComponent != null)
        {
            sizeVideo = visualComponent.getSize();
        }

        // Get the control panel component size:
        Dimension sizeControlPanel = new Dimension(0, 0);
        if (controlComponent != null)
        {
            sizeControlPanel = controlComponent.getSize();
        }
        
        // Get the gain component size:
        Dimension sizeGainComponent = new Dimension(0, 0);
        if (gainComponent != null)
        {
            sizeGainComponent = gainComponent.getSize();
        }
        
        // Get the progress bar size:
        Dimension sizeProgressBar = new Dimension(0, 0); 
        if (progressBarComponent != null)
        {
            // The progress bar must be given a size before being displayed.
            // The progress bar width will be the same as the player panel.
            sizeProgressBar = new Dimension(0, 35); 
        }

        // Get the player panel rectangle
        Dimension sizePanel = new Dimension();

        // The panel width is the width of the visual component + the gain component
        // or the control panel component, whichever is largest
        sizePanel.width = sizeVideo.width + sizeGainComponent.width;
        sizePanel.width = Math.max(sizePanel.width, sizeControlPanel.width);
        
        // The panel height is the height of the control panel component + (the gain component
        // or the visual component, whichever is largest) + the progress bar component
        sizePanel.height = Math.max(sizeVideo.height, sizeGainComponent.height);
        sizePanel.height += sizeControlPanel.height;
        sizePanel.height += sizeProgressBar.height;

        // Add some space between the video clip and the player controls
        sizePanel.width += 20;
        sizePanel.height += 20;
        
        // Get the layout rectangle
        Rectangle rectLayout = new Rectangle(sizePanel);

        // Place the control panel component centered along the top edge.
        if (controlComponent != null)
        {
            controlComponent.setLocation(
                rectLayout.x + (rectLayout.width - sizeControlPanel.width) / 2,
                rectLayout.y
            );
 
            // Subtract the control panel component from the layout rect
            rectLayout.y += sizeControlPanel.height;
            rectLayout.height -= sizeControlPanel.height;
               
            // Show the control panel component
            controlComponent.setVisible(true);
        }
        
        // Place the progress bar centered along the bottom edge.
        if (progressBarComponent != null)
        {
            progressBarComponent.setBounds(
                rectLayout.x,
                rectLayout.y + (rectLayout.height - sizeProgressBar.height),
                sizePanel.width,
                sizeProgressBar.height
            );

            // Subtract the progress bar size from the layout rect
            rectLayout.height -= sizeProgressBar.height;
                
            // Show the progress bar component
            progressBarComponent.setVisible(true);
        }
 
        // Place the gain component centered along the right edge.
        if (gainComponent != null)
        {
            gainComponent.setLocation(
                rectLayout.x + rectLayout.width - sizeGainComponent.width,
                rectLayout.y + (rectLayout.height - sizeGainComponent.height) / 2
            );

            // Subtract the gain component size from the layout rect
            rectLayout.width -= sizeGainComponent.width;
            
            // Show the gain component
            gainComponent.setVisible(true);
        }
        
        // Center the visual component in the remaining area.
        if (visualComponent != null)
        {
            visualComponent.setLocation(
                rectLayout.x + (rectLayout.width - sizeVideo.width) / 2,
                rectLayout.y + (rectLayout.height - sizeVideo.height) / 2
            );
        }

        // Resize the panel and frame window
        setSize(sizePanel);
        ((Frame) getParent()).pack();
    }
}

