/*
 * $id: $
 *
 * Copyright (C) 1998 RealNetworks, Inc.
 * All rights reserved.
 *  
 * http://www.real.com/devzone
 *
 * This program contains proprietary information of RealNetworks, Inc.,
 * and is licensed subject to restrictions on use and distribution.
 *
 * PlayerApplet1.java
 *
 * Demonstrates how to create a simple media player applet
 * with a media event listener.
 */

import java.applet.*;
import java.awt.*;
import java.net.*;
import javax.media.*;

/**
 * A simple media player applet with just a visual component.
 * Plays the same media clip continuously in a loop.
 * Reads the media URL from the applet "URL" parameter tag.
 *
 * <!-- Sample HTML
 * <applet code="PlayerApplet1" width="400" height="300"> 
 * <param name="URL" value="pnm://video.real.com/welcome.rm">
 * </applet>
 * -->
 */
public class PlayerApplet1 
    extends Applet implements ControllerListener
{
    private Player      player = null;                  // Media player
    private Component   visualComponent = null;         // Displays the media clip
    
    public void init()
    {
        // Set the background color
        setBackground(Color.lightGray);
        
        // Remove the layout manager
        setLayout(null);

        // Create the media player
        createPlayer();
    }

    /**
     * Start the media file playback when entering the page.
     */
    public void start()
    {
        if (player != null)
        {
            // Set the wait cursor
            setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

            // Start the media file playback
            player.start();
        }
    }

    /**
     * Stop the media file playback and release all resources
     * before leaving the page.
     */
    public void stop()
    {
        if (player != null)
        {
            // Stop the media file playback
            player.stop();

            // Release the player resources
            player.deallocate();
        }
    }

    /**
     * Stop the media file playback and destroy the media player
     * when the applet is destroyed.
     */
    public void destroy()
    {
        if (player != null)
        {
            // Stop the media file playback
            player.stop();

            // Destroy the media player
            player.close();
        }
    }

    /**
     * The controllerUpdate method implements the ControllerListener interface.
     * This method is called to handle all media events.
     */
    public synchronized void controllerUpdate(ControllerEvent event)
    {
        // Handle the RealizeCompleteEvent. 
        if (event instanceof RealizeCompleteEvent)
        {
            // Get the media player visual component and add it to the applet.
            visualComponent = player.getVisualComponent();
            if (visualComponent != null)
            {
                add(visualComponent);

                // Center the visual component in the applet window
                Dimension sizeApplet = getSize();
                Dimension sizeVideo = visualComponent.getSize();
                visualComponent.setLocation(
                    (sizeApplet.width - sizeVideo.width) / 2,
                    (sizeApplet.height - sizeVideo.height) / 2
                );
            }
        }

        // Handle the PrefetchCompleteEvent. 
        else if (event instanceof PrefetchCompleteEvent)
        {
            // Reset the default cursor
            setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
        }
        
        // Handle the ResourceUnavailableEvent
        else if (event instanceof ResourceUnavailableEvent)
        {
            // Reset the default cursor
            setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
        }

        // Handle the EndOfMediaEvent
        else if (event instanceof EndOfMediaEvent)
        {
            // Restart playback here
            player.start();
        }
    }

    /**
     * Create the media player from the URL given in the applet parameter.
     */
    protected void createPlayer()
    {
        String          url = null;             // URL string from applet parameter
        MediaLocator    mediaLocator = null;    // media locator for the media file

        // Get the media URL string from the applet "URL" parameter.
        url = getParameter("URL");
        if (url == null)
        {
            System.err.println("Invalid URL parameter");
        }
        else
        {
            try
            {
                // Create a media locator from the media URL string
                mediaLocator = new MediaLocator(url);

                // Create the proper media player for this media type.
                player = Manager.createPlayer(mediaLocator);

                // If a player was created:
                if (player != null)
                {
                    // Add this applet as a listener for media player events.
                    player.addControllerListener(this);
                }
                else
                {
                    System.err.println("Could not create player for " + mediaLocator);
                }
            }
            // Handle MalformedURLException
            catch (MalformedURLException e)
            {
                System.err.println("Invalid media file URL: " + url);
            }
            // Handle NoPlayerException
            catch(NoPlayerException e)
            {
                System.err.println("Could not find a player for " + mediaLocator);
            }
            // Handle IOException
            catch(java.io.IOException e)
            {
                System.err.println("IO Exception encountered opening " + mediaLocator);
            }
        }
    }
}

