/*
 * $id: $
 *
 * Copyright (C) 1998 RealNetworks, Inc.
 * All rights reserved.
 *  
 * http://www.real.com/devzone
 *
 * This program contains proprietary information of RealNetworks, Inc.,
 * and is licensed subject to restrictions on use and distribution.
 *
 * PlayerApplet3.java
 *
 * Demonstrates how to extend a player applet with a caching controller 
 * and progress bar.
 */

import java.applet.*;
import java.awt.*;
import java.net.*;
import javax.media.*;

/**
 * Media player applet with a caching controller and progress bar. 
 * Reads the media URL from the applet "URL" parameter tag.
 *
 * <!-- Sample HTML
 * <applet code="PlayerApplet3" width="400" height="300"> 
 * <param name="URL" value="pnm://video.real.com/welcome.rm">
 * </applet>
 * -->
 */
public class PlayerApplet3
    extends Applet implements ControllerListener
{
    private Player          player = null;              // Media player
    private GainControl     gainControl = null;         // Gain control
    private CachingControl  cachingControl = null;      // Caching control
    
    private Component   visualComponent = null;         // Displays the media clip
    private Component   controlComponent = null;        // Displays/controls clip playback
    private Component   gainComponent = null;           // Displays/controls audio gain
    private Component   progressBarComponent = null;    // Displays download progress
    
    public void init()
    {
        // Set the background color
        setBackground(Color.lightGray);
        
        // Remove the layout manager
        setLayout(null);

        // Create the media player
        createPlayer();
    }

    /**
     * Start the media file playback when entering the page.
     */
    public void start()
    {
        if (player != null)
        {
            // Set the wait cursor
            setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

            // Start the media file playback
            player.start();
        }
    }

    /**
     * Stop the media file playback and release all resources
     * before leaving the page.
     */
    public void stop()
    {
        if (player != null)
        {
            // Stop the media file playback
            player.stop();

            // Release the player resources
            player.deallocate();
        }
    }

    /**
     * Stop the media file playback and destroy the media player
     * when the applet is destroyed.
     */
    public void destroy()
    {
        if (player != null)
        {
            // Stop the media file playback
            player.stop();

            // Destroy the media player
            player.close();
        }
    }

    /**
     * The controllerUpdate method implements the ControllerListener interface.
     * This method is called to handle all media events.
     */
    public synchronized void controllerUpdate(ControllerEvent event)
    {
        // Handle the RealizeCompleteEvent. 
        if (event instanceof RealizeCompleteEvent)
        {
            // Create the player visual components
            createPlayerControls();

            // Position the visual components
            placePlayerControls();
        }

        // Handle the PrefetchCompleteEvent. 
        else if (event instanceof PrefetchCompleteEvent)
        {
            // Reset the default cursor
            setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
        }
        
        // Handle the ResourceUnavailableEvent
        else if (event instanceof ResourceUnavailableEvent)
        {
            // Hide the progress bar, if it exists.
            if (progressBarComponent != null)
            {
                progressBarComponent.setVisible(false);
            }
            
            // Reset the default cursor
            setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
        }
        
        // Handle the CachingControlEvent
        else if (event instanceof CachingControlEvent)
        {
            // Show/hide the progress bar component.
            if (progressBarComponent != null)
            {
                long percentComplete = ((CachingControlEvent) event).getContentProgress();
                if (percentComplete == 0)
                {
                    progressBarComponent.setVisible(true);
                }
                else if (percentComplete == 100)
                {
                    progressBarComponent.setVisible(false);
                }
            }
        }
    }

    /**
     * Create the media player from the URL given in the applet parameter.
     */
    protected void createPlayer()
    {
        String          url = null;             // URL string from applet parameter
        MediaLocator    mediaLocator = null;    // media locator for the media file

        // Get the media URL string from the applet "URL" parameter.
        url = getParameter("URL");
        if (url == null)
        {
            System.err.println("Invalid URL parameter");
        }
        else
        {
            try
            {
                // Create a media locator from the media URL string
                mediaLocator = new MediaLocator(url);

                // Create the proper media player for this media type.
                player = Manager.createPlayer(mediaLocator);

                // If a player was created:
                if (player != null)
                {
                    // Add this applet as a listener for media player events.
                    player.addControllerListener(this);
                }
                else
                {
                    System.err.println("Could not create player for " + mediaLocator);
                }
            }
            // Handle MalformedURLException
            catch (MalformedURLException e)
            {
                System.err.println("Invalid media file URL: " + url);
            }
            // Handle NoPlayerException
            catch(NoPlayerException e)
            {
                System.err.println("Could not find a player for " + mediaLocator);
            }
            // Handle IOException
            catch(java.io.IOException e)
            {
                System.err.println("IO Exception encountered opening " + mediaLocator);
            }
        }
    }

    /**
     * Create the player gain control, caching control and visual components
     */
    protected void createPlayerControls()
    {
        // Get the caching control
        cachingControl = (CachingControl) player.getControl("com.real.media.RMCachingControl");
        if (cachingControl != null)
        {
            // Get the progress bar component
            progressBarComponent = cachingControl.getProgressBarComponent();
            if (progressBarComponent != null)
            {
                // Add the component to the applet window
                add(progressBarComponent);
            }
        }
 
        // Get the gain control
        gainControl = player.getGainControl();
        if (gainControl != null)
        {
            // Get the gain control component
            gainComponent = gainControl.getControlComponent();
            if (gainComponent != null)
            {
                // Add the component to the applet window
                add(gainComponent);
            }
        }
        
        // Get the control panel component
        controlComponent = player.getControlPanelComponent();
        if (controlComponent != null)
        {
            // Add the component to the applet window
            add(controlComponent);
        }

        // Get the visual component
        visualComponent = player.getVisualComponent();
        if (visualComponent != null)
        {
            // Add the component to the applet window
            add(visualComponent);
        }
    }
    
    /**
     * Position the player components within the player panel
     */
    protected void placePlayerControls()
    {
        // Get the applet window rectangle
        Rectangle rect = new Rectangle(getSize());
        
        // If the control panel component exists:
        if (controlComponent != null)
        {
            // Place the control component centered along the top edge
            Dimension size = controlComponent.getSize();
            controlComponent.setLocation(
                rect.x + (rect.width - size.width) / 2,
                rect.y
            );

            // Subtract the control component size from the applet window rect
            rect.y += size.height;
            rect.height -= size.height;
            
            // Show the control component
            controlComponent.setVisible(true);
        }
        
        // If the progress bar component exists:
        if (progressBarComponent != null)
        {
            // Place the progress bar centered along the bottom edge
            Dimension size = new Dimension(200, 35);
            progressBarComponent.setBounds(getInsets().left +
                rect.x + (rect.width - size.width) / 2, 
                rect.y + (rect.height - (size.height * 2)),
                size.width,
                size.height
            );

            // Subtract the control component size from the applet window rect
            rect.height -= size.height;
            
            // Show the control component
            progressBarComponent.setVisible(true);
        }
 
        // If the gain component exists:
        if (gainComponent != null)
        {
            // Place the gain component centered along the right edge
            Dimension size = gainComponent.getSize();
            gainComponent.setLocation(
                rect.x + rect.width - size.width,
                rect.y + (rect.height - size.height) / 2
            );

            // Subtract the control component size from the applet window rect
            rect.width -= size.width;
            
            // Show the gain component
            gainComponent.setVisible(true);
        }

        // If the visual component exists:
        if (visualComponent != null)
        {
            // Center the visual component in the remaining area
            Dimension size = visualComponent.getSize();
            visualComponent.setLocation(
                rect.x + (rect.width - size.width) / 2,
                rect.y + (rect.height - size.height) / 2
            );
        }
    }
}

