/*
 * $id: $
 *
 * Copyright (C) 1998 RealNetworks, Inc.
 * All rights reserved.
 *  
 * http://www.real.com/devzone
 *
 * This program contains proprietary information of RealNetworks, Inc.,
 * and is licensed subject to restrictions on use and distribution.
 *
 * PlayerStateApp.java
 *
 * Demonstrates media player state transitions.
 */

import java.awt.*;
import java.awt.event.*;
import java.net.*;
import javax.media.*;

import com.real.media.*;

/**
 * Standalone player app which demonstrates media player state transitions.
 *
 * Example usage:
 * java PlayerStateApp pnm://video.real.com/welcome.rm
 */
public class PlayerStateApp
{
    private PlayerPanel playerPanel;
    
    private EventWindow eventWindow;
    private Label       statusLabel;

    public static void main(String argv[])
    {
        // If a URL argument was specified:
        if (argv.length >= 1)
        {
            // Get the url argument
            String url = argv[argv.length - 1];

            // Create the PlayerState app
            PlayerStateApp playerStateApp = new PlayerStateApp(url);
        }
        else
        {
            // Output the command line usage
            System.err.println("Usage: java PlayerStateApp <URL>");
        }
    }

    public PlayerStateApp(String url)
    {
        // Create the frame window
        Frame frame = new Frame("PlayerStateApp");

        // Set the background color
        frame.setBackground(Color.lightGray);

        // Make the frame non-resizable
        frame.setResizable(false);
        
        // Create the player panel
        playerPanel = new PlayerPanel(url);
        playerPanel.setSize(400, 300);
        
        // Add the player panel to the frame window
        frame.add("Center", playerPanel);

        // Initialize the player panel
        playerPanel.init();
        
        // Add the frame window event handler
        frame.addWindowListener(
            new WindowAdapter()
            {
                public void windowClosing(WindowEvent e)
                { 
                    // Stop the player
                    playerPanel.stop();

                    // Destroy the player
                    playerPanel.destroy();

                    // Exit the application
                    System.exit(0); 
                }
            }
        );

	com.real.util.DebugOut.setDebug(true);

        // Size and show the frame window
        frame.pack();
        frame.show();
    }
}

/**
 * Encapsulates a media player.
 *
 * Handles creating and positioning the player's visual
 * components, and responding to media events.
 */
class PlayerPanel
    extends Panel implements ControllerListener, ActionListener
{
    private String          url;                        // Media clip URL
    private Player          player = null;              // Media player
    private GainControl     gainControl = null;         // Gain control
    private CachingControl  cachingControl = null;      // Caching control
    
    private Component   visualComponent = null;         // Displays the media clip
    private Component   controlComponent = null;        // Displays/controls clip playback
    private Component   gainComponent = null;           // Displays/controls audio gain
    private Component   progressBarComponent = null;    // Displays download progress

    private EventWindow eventWindow = null;     // Displays controller events
    private Label       statusLabel = null;     // Displays current player status

    private Button      realizeButton;          // Realizes the player
    private Button      prefetchButton;         // Prefetches the player
    private Button      startButton;            // Starts the player
    private Button      stopButton;             // Stops the player

    public PlayerPanel(String url)
    {
        this.url = url;
    }
    
    public void init()
    {
        // Set the background color
        setBackground(Color.lightGray);
        
        // Remove the layout manager
        setLayout(null);

        // Create the media player
        createPlayer();
        
        // Create the event window
        eventWindow = new EventWindow();

        // Create the status panel
        createStatusPanel();

        // Set the status label text
        setStatusText();
    }

    /**
     * Stop the media file playback and release all resources.
     */
    public void stop()
    {
        if (player != null)
        {
            // Stop the media file playback
            player.stop();

            // Release the player resources
            player.deallocate();
        }
    }

    /**
     * Stop the media file playback and destroy the media player.
     */
    public void destroy()
    {
        if (player != null)
        {
            // Stop the media file playback
            player.stop();

            // Destroy the media player
            player.close();
        }

        if (eventWindow != null)
        {
            // Destroy the event window
            eventWindow.dispose();
        }
    }

    /**
     * Action listener implementation
     */
    public void actionPerformed(ActionEvent event)
    {
        String cmd = event.getActionCommand();

        try
        {
            // Handle the Realize command
            if (cmd == "Realize")
            {
                player.realize();
            }

            // Handle the Prefetch command
            else if (cmd == "Prefetch")
            {
                player.prefetch();
            }

            // Handle the Start command
            else if (cmd == "Start")
            {
                player.start();
            }

            // Handle the Stop command
            else if (cmd == "Stop")
            {
                player.stop();
            }
        }
        catch (Exception e)
        {
            System.out.println("Exception in PStateApplet:\n" + e);
        }
    }

    /**
     * The controllerUpdate method implements the ControllerListener interface.
     * This method is called to handle all media events.
     */
    public synchronized void controllerUpdate(ControllerEvent event)
    {
        // Post all non-RM-specific events to the event window
        if (!(event instanceof RMOnPosChangedEvent))
        {
            eventWindow.add(event.toString());
        }
        
        // Handle the RealizeCompleteEvent. 
        if (event instanceof RealizeCompleteEvent)
        {
            // If the player isn't already started:
            if (player.getState() == Controller.Realized)
            {
                // Create the player visual components
                createPlayerControls();

                // Position and show the player visual components
                placePlayerControls();
            }
        }
        
        // Handle the ResourceUnavailableEvent
        else if (event instanceof ResourceUnavailableEvent)
        {
            // Hide the progress bar, if it exists.
            if (progressBarComponent != null)
            {
                progressBarComponent.setVisible(false);
            }
        }
        
        // Handle the CachingControlEvent
        else if (event instanceof CachingControlEvent)
        {
            // Show/hide the progress bar component.
            if (progressBarComponent != null)
            {
                long percentComplete = ((CachingControlEvent) event).getContentProgress();
                if (percentComplete == 0)
                {
                    progressBarComponent.setVisible(true);
                }
                else if (percentComplete == 100)
                {
                    progressBarComponent.setVisible(false);
                }
            }
        }

        // Set the status label text
        setStatusText();
    }

    /**
     * Create the media player from the URL given in the command line.
     */
    protected void createPlayer()
    {
        MediaLocator    mediaLocator = null;    // media locator for the media file

        try
        {
            // Create a media locator from the media URL string
            mediaLocator = new MediaLocator(url);

            // Create the proper media player for this media type.
            player = Manager.createPlayer(mediaLocator);

            // If a player was created:
            if (player != null)
            {
                // Add this applet as a listener for media player events.
                player.addControllerListener(this);
            }
            else
            {
                System.err.println("Could not create player for " + mediaLocator);
            }
        }
        // Handle MalformedURLException
        catch (MalformedURLException e)
        {
            System.err.println("Invalid media file URL: " + url);
        }
        // Handle NoPlayerException
        catch(NoPlayerException e)
        {
            System.err.println("Could not find a player for " + mediaLocator);
        }
        // Handle IOException
        catch(java.io.IOException e)
        {
            System.err.println("IO Exception encountered opening " + mediaLocator);
        }
    }

    /**
     * Create the player gain control, caching control and visual components
     */
    protected void createPlayerControls()
    {
        // Get the caching control
        cachingControl = (CachingControl) player.getControl("com.real.media.RMCachingControl");
        if (cachingControl != null)
        {
            // Get the progress bar component
            progressBarComponent = cachingControl.getProgressBarComponent();
            if (progressBarComponent != null)
            {
                // Add the component to the applet window
                add(progressBarComponent);
            }
        }
 
        // Get the gain control
        gainControl = player.getGainControl();
        if (gainControl != null)
        {
            // Get the gain control component
            gainComponent = gainControl.getControlComponent();
            if (gainComponent != null)
            {
                // Add the component to the applet window
                add(gainComponent);
            }
        }
        
        // Get the control panel component
        controlComponent = player.getControlPanelComponent();
        if (controlComponent != null)
        {
            // Add the component to the applet window
            add(controlComponent);
        }

        // Get the visual component
        visualComponent = player.getVisualComponent();
        if (visualComponent != null)
        {
            // Add the component to the applet window
            add(visualComponent);
        }
    }
    
    /**
     * Position the player components within the player panel
     */
    protected void placePlayerControls()
    {
        // Get the visual component size:
        Dimension sizeVideo = new Dimension(0, 0);
        if (visualComponent != null)
        {
            sizeVideo = visualComponent.getSize();
        }

        // Get the control panel component size:
        Dimension sizeControlPanel = new Dimension(0, 0);
        if (controlComponent != null)
        {
            sizeControlPanel = controlComponent.getSize();
        }
        
        // Get the gain component size:
        Dimension sizeGainComponent = new Dimension(0, 0);
        if (gainComponent != null)
        {
            sizeGainComponent = gainComponent.getSize();
        }
        
        // Get the progress bar size:
        Dimension sizeProgressBar = new Dimension(0, 0); 
        if (progressBarComponent != null)
        {
            // The progress bar must be given a size before being displayed.
            // The progress bar width will be the same as the player panel.
            sizeProgressBar = new Dimension(0, 35); 
        }

        // Get the player panel rectangle
        Dimension sizePanel = new Dimension();

        // The panel width is the width of the visual component + the gain component
        // or the control panel component, whichever is largest
        sizePanel.width = sizeVideo.width + sizeGainComponent.width;
        sizePanel.width = Math.max(sizePanel.width, sizeControlPanel.width);
        
        // The panel height is the height of the control panel component + (the gain component
        // or the visual component, whichever is largest) + the progress bar component
        sizePanel.height = Math.max(sizeVideo.height, sizeGainComponent.height);
        sizePanel.height += sizeControlPanel.height;
        sizePanel.height += sizeProgressBar.height;

        // Add some space between the video clip and the player controls
        sizePanel.width += 20;
        sizePanel.height += 20;
        
        // Get the layout rectangle
        Rectangle rectLayout = new Rectangle(sizePanel);

        // Place the control panel component centered along the top edge.
        if (controlComponent != null)
        {
            controlComponent.setLocation(
                rectLayout.x + (rectLayout.width - sizeControlPanel.width) / 2,
                rectLayout.y
            );
 
            // Subtract the control panel component from the layout rect
            rectLayout.y += sizeControlPanel.height;
            rectLayout.height -= sizeControlPanel.height;
               
            // Show the control panel component
            controlComponent.setVisible(true);
        }
        
        // Place the progress bar centered along the bottom edge.
        if (progressBarComponent != null)
        {
            progressBarComponent.setBounds(
                rectLayout.x,
                rectLayout.y + (rectLayout.height - sizeProgressBar.height),
                sizePanel.width,
                sizeProgressBar.height
            );

            // Subtract the progress bar size from the layout rect
            rectLayout.height -= sizeProgressBar.height;
                
            // Show the progress bar component
            progressBarComponent.setVisible(true);
        }
 
        // Place the gain component centered along the right edge.
        if (gainComponent != null)
        {
            gainComponent.setLocation(
                rectLayout.x + rectLayout.width - sizeGainComponent.width,
                rectLayout.y + (rectLayout.height - sizeGainComponent.height) / 2
            );

            // Subtract the gain component size from the layout rect
            rectLayout.width -= sizeGainComponent.width;
            
            // Show the gain component
            gainComponent.setVisible(true);
        }
        
        // Center the visual component in the remaining area.
        if (visualComponent != null)
        {
            visualComponent.setLocation(
                rectLayout.x + (rectLayout.width - sizeVideo.width) / 2,
                rectLayout.y + (rectLayout.height - sizeVideo.height) / 2
            );
        }

        // Resize the panel and frame window
        setSize(sizePanel);
        ((Frame)getParent()).pack();
    }

    /**
     * Create the status panel and add it to the frame window
     */
    protected void createStatusPanel()
    {        
        // Create the status panel
        Panel statusPanel = new Panel();
        statusPanel.setLayout(new GridLayout(0, 1));

        // Create the status label
        statusLabel = new Label();

        // Add the status label to the control panel
        statusPanel.add(statusLabel);
        
        // Create the button panel
        Panel buttonPanel = new Panel();
        
        realizeButton = new Button("Realize");
        realizeButton.addActionListener(this);
        buttonPanel.add(realizeButton);

        prefetchButton = new Button("Prefetch");
        prefetchButton.addActionListener(this);
        buttonPanel.add(prefetchButton);

        startButton = new Button("Start");
        startButton.addActionListener(this);
        buttonPanel.add(startButton);

        stopButton = new Button("Stop");
        stopButton.addActionListener(this);
        buttonPanel.add(stopButton);
        
        // Add the button panel to the control panel
        statusPanel.add(buttonPanel);
        
        // Add the control panel to the frame window
        getParent().add("South", statusPanel);
    }

    /**
     * Set the status label text
     */
    protected void setStatusText()
    {
        if ((player != null) && (statusLabel != null))
        {
            // Get the current player state
            int state = player.getState();

            // Get the text for the current state
            String statusText = "    Current State: ";
            if (state == Controller.Unrealized)
            {
                statusText += "Unrealized";
            }
            else if (state == Controller.Realizing)
            {
                statusText += "Realizing";
            }
            else if (state == Controller.Realized)
            {
                statusText += "Realized";
            }
            else if (state == Controller.Prefetching)
            {
                statusText += "Prefetching";
            }
            else if (state == Controller.Prefetched)
            {
                statusText += "Prefetched";
            }
            else if (state == Controller.Started)
            {
                statusText += "Started";
            }

            // Set the status label text
            statusLabel.setText(statusText);
        }
    }
}

/**
 * Frame window used to display Controller events 
 * as they are received.
 */
class EventWindow
    extends Frame
{
    private TextArea    eventList;

    public EventWindow()
    {
        super("Controller Events");

        // Initialize the window
        init();

        // Size and show the window
        pack();
        show();
    }

    /**
     * Adds an item to the event list
     */
    public void add(String string)
    {
        // Add the text to the text window
        eventList.setText(eventList.getText() + string + '\n');
    }

    /**
     * Creates the window controls
     */
    protected void init()
    {
        // Create the text window
        eventList = new TextArea(25, 80);
        eventList.setEditable(false);
        add("Center", eventList);
    }
}

