//
// Tutorial3.java
//
// This tutorial demonstrates how to add a "Control Panel" to your video playback 
// window (Frame). The Control Panel allows you to stop/pause/retstart and seek along 
// the video's timeline. It also shows how to use setStickySize(false) to allow 
// visualComponent to change size during a presentation.

// 
// Copyright (C) 1995,1996,1997,1998 RealNetworks, Inc.
// All rights reserved.
//


import java.awt.*;
import java.awt.event.*;

import javax.media.*;  


public class Tutorial3 extends WindowAdapter implements ControllerListener
{
	private javax.media.Player player = null;
	private String urlString = null;
	private java.awt.Component visualComponent = null;
	private java.awt.Frame frame = null;
	private java.awt.Component controlPanel = null;

	public static void main(String[] args)
	{
		System.out.println("Welcome to JMF for RealSystem G2 Tutorial3.");

		if (args.length == 0 || args[0] == null || args[0].length() == 0)
		{
			System.out.println("Tutorial3: Usage - java Tutorial3 <url>");
			return;
		}

		new Tutorial3(args[0]);
	}

	public Tutorial3(String urlString)
	{
		// save URL for later use

		this.urlString = urlString;

		// first create MediaLocator

		MediaLocator mrl = new MediaLocator(urlString);

		// create Player object 

		try
		{
			player = Manager.createPlayer(mrl);
		}
		catch (Exception e)
		{
			System.err.println("Tutorial3: Unable to create player for URL = " + urlString);
			System.err.println(e);
			return;
		}

		System.out.println("Tutorial3: beginning playback now.");

		// now we have our player, lets start it! 
		player.addControllerListener(this);
		player.start();
	}

	public void windowClosing(WindowEvent e)
	{
		// if the player is still playing, stop it before we close	
		if (player.getState() == Controller.Started)
		{
			player.stop();
		}

		frame.hide();
		player.close();
	}

	public void resizeFrame()
	{
		// get frame insets 
		int xInsets = frame.getInsets().left + frame.getInsets().right;
		int yInsets = frame.getInsets().top + frame.getInsets().bottom;

		// get the control panel 
		Dimension csize = controlPanel.getSize();

		int height = csize.height + yInsets;
		int width = csize.width + xInsets;

		if (visualComponent != null)
		{
			Dimension vsize = visualComponent.getSize();

			width = vsize.width + xInsets;
			height += vsize.height;
		}

		frame.setSize(width, height);
		frame.validate();
	}

	public void controllerUpdate(ControllerEvent e)
	{
		// realizing is done - we can get the VisualComponent now! 
		if (e instanceof RealizeCompleteEvent)
		{
			// create frame first
			frame = new Frame(urlString);
			frame.setLayout(new BorderLayout());
			frame.addNotify();

			// get frame insets 
			int xInsets = frame.getInsets().left + frame.getInsets().right;
			int yInsets = frame.getInsets().top + frame.getInsets().bottom;

			// get the control panel 
			controlPanel = player.getControlPanelComponent();
			Dimension csize = controlPanel.getSize();
			frame.add(controlPanel, "South");

			int height = csize.height + yInsets;
			int width = csize.width + xInsets;

			// handle the visual component 
			Dimension vsize = null; 
			visualComponent = player.getVisualComponent();
			if (visualComponent != null)
			{
				vsize = visualComponent.getSize();
				frame.add(visualComponent, "Center");

				width = vsize.width + xInsets;
				height += vsize.height;

				visualComponent.addComponentListener(new CompSizeAdapter(this));

				// Sticky Size demo! Call "true" and sizes won't change.
				// Disabling sticky size allows visualComponent to change size
				// as presentation requirements demand when using SMIL or RAM.
				((com.real.media.RMPlayer) player).setStickySize(false);
			}

			frame.setSize(width, height);
			frame.setVisible(true);
			frame.addWindowListener(this);
		}
		else if (e instanceof ControllerClosedEvent)
		{
			if (frame != null)
			{
				frame.dispose();
			}

			System.exit(0);
		}
	}
}

// component adapter class - listens for component size events
class CompSizeAdapter extends ComponentAdapter
{
	private Tutorial3 t3 = null;

	public CompSizeAdapter(Tutorial3 t3)
	{
		this.t3 = t3;
	}

	public void componentResized(ComponentEvent e)	
	{
		t3.resizeFrame();
	}
}
