//
// Tutorial5.java
//
// This tutorial demonstrates how to add a custom control panel and 
// retrieve and display the TAC information.
// a presentation.
// 
// Copyright (C) 1995,1996,1997,1998 RealNetworks, Inc.
// All rights reserved.
//


import java.awt.*;
import java.awt.event.*;

import javax.media.*;  


public class Tutorial5 extends WindowAdapter implements ControllerListener
{
	private javax.media.Player player = null;
	private String urlString = null;
	private java.awt.Component visualComponent = null;
	private java.awt.Frame frame = null;

	private MyControlPanel controlPanel = null;
	private long currentPercentDone;


	public static void main(String[] args)
	{
		System.out.println("Welcome to JMF for RealSystem G2 Tutorial5.");

		if (args.length == 0 || args[0] == null || args[0].length() == 0)
		{
			System.out.println("Tutorial5: Usage - java Tutorial5 <url>");
			return;
		}

		new Tutorial5(args[0]);
	}

	public Tutorial5(String urlString)
	{
		//com.real.util.DebugOut.setDebug(true);

		// save URL for later use

		this.urlString = urlString;

		// first create MediaLocator

		MediaLocator mrl = new MediaLocator(urlString);

		// create Player object 

		try
		{
			player = Manager.createPlayer(mrl);
		}
		catch (Exception e)
		{
			System.err.println("Tutorial5: Unable to create player for URL = " + urlString);
			System.err.println(e);
			return;
		}

		System.out.println("Tutorial5: beginning playback now.");

		// now we have our player, lets start it! 
		player.addControllerListener(this);
		player.start();
	}

	public void windowClosing(WindowEvent e)
	{
		// if the player is still playing, stop it before we close	
		if (player.getState() == Controller.Started)
		{
			player.stop();
		}

		frame.hide();
		player.close();
	}

	public void controllerUpdate(ControllerEvent e)
	{
		// realizing is done - we can get the VisualComponent now! 
		if (e instanceof RealizeCompleteEvent)
		{
			// create frame first
			frame = new Frame(urlString);
			frame.setLayout(new BorderLayout());
			frame.addNotify();

			// get frame insets 
			int xInsets = frame.getInsets().left + frame.getInsets().right;
			int yInsets = frame.getInsets().top + frame.getInsets().bottom;

			// get the control panel 
			controlPanel = new MyControlPanel(player, currentPercentDone);
			Dimension csize = controlPanel.getSize();
			frame.add(controlPanel, "South");

			int height = csize.height + yInsets;
			int width = csize.width + xInsets;

			// handle the visual component 
			Dimension vsize = null; 
			visualComponent = player.getVisualComponent();
			if (visualComponent != null)
			{
				vsize = visualComponent.getSize();
				frame.add(visualComponent, "Center");

				width = vsize.width + xInsets;
				height += vsize.height;
			}

			frame.setSize(width, height);
			frame.setVisible(true);
			frame.addWindowListener(this);
		}
		else if (e instanceof ControllerClosedEvent)
		{
			if (frame != null)
			{
				frame.dispose();
			}

			System.exit(0);
		}
		else if (e instanceof CachingControlEvent)
		{
			CachingControlEvent cce = (CachingControlEvent) e;
			currentPercentDone = cce.getContentProgress();

			// for file based playback, 100% buffering occurs almost
			// instantly - consequently, the RealizeComplete event 
			// has not yet been sent so no controlPanel has been
			// created yet. So we must check for a null controlPanel.
			if (controlPanel != null)
			{
				controlPanel.setBuffPercent(currentPercentDone);
			}
		}
	}
}

class MyControlPanel extends Panel implements ActionListener, ControllerListener
{
	private boolean isPaused = false;
	private com.real.media.RMPlayer player = null;

	// AWT components
	private Button playButton;
	private Button pauseButton;
	private Button stopButton;
	private TextField buffStatus;

	public void setBuffPercent(long pct)
	{
		String str = String.valueOf(pct) + (pct < 10 ?" %" :"%");
		buffStatus.setText(str);
	}

	public MyControlPanel(Player player, long currentPercentDone)
	{
		this.player = (com.real.media.RMPlayer) player;
		player.addControllerListener(this);

		setLayout(new FlowLayout());
		setVisible(true);
		setSize(400, 30);
		setBackground(new Color(192, 192, 192));

		playButton = new Button("Play");
		playButton.setSize(42, 17);
		add(playButton);

		pauseButton = new Button("Pause");
		pauseButton.setSize(42, 17);
		add(pauseButton);

		stopButton = new Button("Stop");
		stopButton.setSize(42, 17);
		add(stopButton);

		buffStatus = new TextField(String.valueOf(currentPercentDone) + (currentPercentDone < 10 ?" %" :"%"));
		buffStatus.setVisible(true);
		buffStatus.setSize(35, 17);
		buffStatus.setEditable(false);
		add(buffStatus);

		pauseButton.addActionListener(this);
		playButton.addActionListener(this);
		stopButton.addActionListener(this);

		playButton.setEnabled(false);
		pauseButton.setEnabled(false);
		stopButton.setEnabled(true);
	} 

	public void controllerUpdate(ControllerEvent e)
	{

		if (e instanceof ControllerErrorEvent)
		{
			isPaused = false;
			playButton.setEnabled(true);
			pauseButton.setEnabled(false);
			stopButton.setEnabled(false);
		}
		else if (e instanceof StartEvent)
		{
			isPaused = false;
			playButton.setEnabled(false);
			pauseButton.setEnabled(true);
			stopButton.setEnabled(true);
		}
		else if (e instanceof StopEvent)
		{
			isPaused = false;
			playButton.setEnabled(true);
			pauseButton.setEnabled(false);
			stopButton.setEnabled(false);
		}
		else if (e instanceof com.real.media.RMPauseEvent)
		{
			isPaused = true;
			playButton.setEnabled(true);
			pauseButton.setEnabled(false);
			stopButton.setEnabled(true);
		}
	}

	public void playButtonClicked(ActionEvent event)
	{
		if (isPaused)
		{
			player.beginPlay();
		}
		else
		{
			player.start();	

			// disable play button and enable stop only
			playButton.setEnabled(false);
			pauseButton.setEnabled(false);
			stopButton.setEnabled(true);
		}

		isPaused = false;
	}

	public void pauseButtonClicked(ActionEvent event)
	{
		player.pausePlay();
	}

	public void stopButtonClicked(ActionEvent event)
	{
		player.stop();
	}

	// handle button click events
	public void actionPerformed(ActionEvent event)
	{
		// see which button was clicked 
		if (event.getSource() == playButton)
		{
			playButtonClicked(event);
		}
		else if (event.getSource() == pauseButton)
		{
			pauseButtonClicked(event);
		}
		else if (event.getSource() == stopButton)
		{
			stopButtonClicked(event);
		}
	}
}
