/****************************************************************************
 *
 *  $Id: expksink.cpp,v 1.3 2000/02/08 22:50:02 glennh Exp $
 *
 *  Copyright (C) 1998 RealNetworks, Inc.
 *  All rights reserved.
 *
 *  http://www.real.com/devzone
 *
 *  This program contains proprietary information of RealNetworks, Inc.,
 *  and is licensed subject to restrictions on use and distribution.
 *
 *  expksink.cpp
 *
 *  Example Live Packet Sink plug-in for the RealMedia Architecture (RMA).
 */

/****************************************************************************
 *
 *  This is an example of a Live Packet Sink plug-in. A Live Packet Sink 
 *  server plug-in is able to receive notifications about and examine the
 *  headers and packets being sent to the server from a live source. The
 *  plugin can use this information for many purposes, including logging
 *  statistics about the live content or archiving the live streams to disk.
 *
 *  Example: expksink.cpp
 *
 *  This example plug-in will set itself up as a packet sink for a live 
 *  source as soon as that source connects to the RMA server. After this 
 *  is done, the plug-in will receive notifications for every header and 
 *  packet that are sent from the live source. It will perform a simple 
 *  logging function by adding up the sizes of all packets received. When 
 *  the live source finally terminates and disconnects from the server, 
 *  this plug-in will print out the number of streams sent by the live 
 *  source as well as the total number of bytes received on each stream.
 *
 *  This example plug-in will only act as a sink for one live source at a
 *  time, so when the first live source disconnects, the statistics will be
 *  printed to the console, and it will then be ready to handle the next 
 *  live source that connects.
 *
 *  To test this example plug-in, simply build the plug-in, place it in
 *  your RMA server plug-in directory, and connect a live source such as
 *  a RealNetworks live encoder to the server. After allowing the live 
 *  source to run briefly, stop the encoder and the statistics mentioned 
 *  above should be printed out to the server console.
 */


/****************************************************************************
 * Defines
 */
#define	INITGUID     /* Interface ID's */

/****************************************************************************
 * Includes
 */
#include <stdio.h>     /* printf, sprintf */
#include <string.h>

#include "pntypes.h" 
#include "pncom.h"     /* IUnknown */
#include "rmacomm.h"   /* IRMACommonClassFactory */
#include "rmapckts.h"  /* IRMABuffer, IRMAPacket, IRMAValues */
#include "rmaplugn.h"  /* IRMAPlugin */
#include "rmasrc.h"    /* IRMARawSinkObject */
#include "rmamon.h"    /* IRMAPNRegistry */
#include "rmafiles.h"  /* IRMARequest */
#include "expksink.h"  /* CExRegistryWatcher, CExamplePacketSink */


/****************************************************************************
 *  RMACreateInstance                                        ref:  rmaplugn.h
 *
 *  This routine creates a new instance of the CExRegistryWatcher class.
 *  It is called when the RMA server application is launched.
 */
STDAPI
RMACreateInstance(IUnknown** ppExSinkObj)
{
    *ppExSinkObj = (IUnknown*)(IRMAPlugin*)new CExRegistryWatcher();
    if (*ppExSinkObj != NULL)
    {
	(*ppExSinkObj)->AddRef();
	return PNR_OK;
    }

    return PNR_OUTOFMEMORY;
}


/****************************************************************************
 *  CExRegistryWatcher static variables                      ref:  expksink.h
 *
 *  These variables are passed to the RMA server to provide information about
 *  this plug-in. They are required to be static in order to remain valid
 *  for the lifetime of the plug-in.
 */
const char* CExRegistryWatcher::zm_pDescription = MY_DESCRIPTION;
const char* CExRegistryWatcher::zm_pCopyright   = MY_COPYRIGHT;
const char* CExRegistryWatcher::zm_pMoreInfoURL = MY_MORE_INFO_URL;


/****************************************************************************
 *  CExRegistryWatcher::CExRegistryWatcher                   ref:  expksink.h
 *
 *  Constructor
 */
CExRegistryWatcher::CExRegistryWatcher(void)
    : m_RefCount	    (0),
      m_pClassFactory	    (NULL),
      m_pRegistry	    (NULL),
      m_pPropWatch	    (NULL),
      m_ulConnID	    (0),
      m_pPacketSink	    (NULL)
{
}

/****************************************************************************
 *  CExRegistryWatcher::~CExRegistryWatcher                 ref:   expksink.h
 *
 *  Destructor
 */
CExRegistryWatcher::~CExRegistryWatcher(void)
{
    if (m_pClassFactory)
    {
	m_pClassFactory->Release();
	m_pClassFactory = NULL;
    }

    if (m_pRegistry)
    {
	m_pRegistry->Release();
	m_pRegistry = NULL;
    }

    if (m_pPropWatch)
    {
	m_pPropWatch->Release();
	m_pPropWatch = NULL;
    }

    if (m_pPacketSink)
    {
	m_pPacketSink->Release();
	m_pPacketSink = NULL;
    }
}

// IUnknown COM Interface Methods

/****************************************************************************
 *  IUnknown::AddRef                                            ref:  pncom.h
 *
 *  This routine increases the object reference count in a thread safe
 *  manner. The reference count is used to manage the lifetime of an object.
 *  This method must be explicitly called by the user whenever a new
 *  reference to an object is used.
 */
STDMETHODIMP_(UINT32)
CExRegistryWatcher::AddRef(void)
{
    return InterlockedIncrement(&m_RefCount);
}


/****************************************************************************
 *  IUnknown::Release                                           ref:  pncom.h
 *
 *  This routine decreases the object reference count in a thread safe
 *  manner, and deletes the object if no more references to it exist. It must
 *  be called explicitly by the user whenever an object is no longer needed.
 */
STDMETHODIMP_(UINT32)
CExRegistryWatcher::Release(void)
{
    if (InterlockedDecrement(&m_RefCount) > 0)
    {
	return m_RefCount;
    }

    delete this;
    return 0;
}


/****************************************************************************
 *  IUnknown::QueryInterface                                    ref:  pncom.h
 *
 *  This routine indicates which interfaces this object supports. If a given
 *  interface is supported, the object's reference count is incremented, and
 *  a reference to that interface is returned. Otherwise a NULL object and
 *  error code are returned. This method is called by other objects to
 *  discover the functionality of this object.
 */
STDMETHODIMP
CExRegistryWatcher::QueryInterface
(
    REFIID interfaceID,
    void** ppInterfaceObj
)
{
    // By definition all COM objects support the IUnknown interface
    if (IsEqualIID(interfaceID, IID_IUnknown))
    {
	AddRef();
	*ppInterfaceObj = (IUnknown*)(IRMAPlugin*)this;
	return PNR_OK;
    }

    // IRMAPlugin interface is supported
    else if (IsEqualIID(interfaceID, IID_IRMAPlugin))
    {
	AddRef();
	*ppInterfaceObj = (IRMAPlugin*)this;
	return PNR_OK;
    }

    // IRMAPropWatchResponse interface is supported
    else if (IsEqualIID(interfaceID, IID_IRMAPropWatchResponse))
    {
	AddRef();
	*ppInterfaceObj = (IRMAPropWatchResponse*)this;
	return PNR_OK;
    }

    // No other interfaces are supported
    *ppInterfaceObj = NULL;
    return PNR_NOINTERFACE;
}

/****************************************************************************
 *  IRMAPlugin::GetPluginInfo                                ref:  rmaplugn.h
 *
 *  This routine returns descriptive information about the plug-in. It is 
 *  called when the RMA server application is launched.
 */
STDMETHODIMP
CExRegistryWatcher::GetPluginInfo
(
    REF(BOOL)        bLoadMultiple,
    REF(const char*) pDescription,
    REF(const char*) pCopyright,
    REF(const char*) pMoreInfoURL,
    REF(UINT32)      versionNumber
)
{
    bLoadMultiple = FALSE;
    pDescription  = zm_pDescription;
    pCopyright    = zm_pCopyright;
    pMoreInfoURL  = zm_pMoreInfoURL;
    versionNumber = MY_PLUGIN_VERSION;

    return PNR_OK;
}


/****************************************************************************
 *  IRMAPlugin::InitPlugin                                   ref:  rmaplugn.h
 *
 *  This routine performs initialization steps such as determining if
 *  required interfaces are available. It is called when the RMA core 
 *  application is launched, and whenever a new Packet Sink object
 *  is needed.
 */
STDMETHODIMP
CExRegistryWatcher::InitPlugin(IUnknown* pContext)
{
    /*
     * Store a reference to the IRMACommonClassFactory interface which is
     * used to create commonly used RMA objects such as IRMAPacket, 
     * IRMAValues, and IRMABuffers.
     */
    pContext->QueryInterface(IID_IRMACommonClassFactory, 
	                    (void**)&m_pClassFactory);
    if (m_pClassFactory == NULL)
    {
	return PNR_NOTIMPL;
    }

    /*
     * Store a reference to the IRMAPNRegistry interface which is
     * used to retrieve useful information from the server core
     */
    pContext->QueryInterface(IID_IRMAPNRegistry, 
	                    (void**)&m_pRegistry);
    if (m_pRegistry == NULL)
    {
	return PNR_NOTIMPL;
    }

    /* 
     * Note that QueryInterface() takes care of adding a reference to the
     * interface for us. You are however responsible for releasing the
     * reference to the object when you are done using it by calling
     * the Release() routine.
     */

    // Set watches on server registry properties that we are interested in
    return SetWatches();
}

/****************************************************************************
 *  CExRegistryWatcher::LiveSourceDone                       ref:  expksink.h
 *
 *  Called by the packet sink object, when the live source is done, and
 *  we can clean up the packet sink.
 */
void
CExRegistryWatcher::LiveSourceDone(void)
{
    if (m_pPacketSink)
    {
	m_pPacketSink->Release();
	m_pPacketSink = NULL;
    }
}

/****************************************************************************
 *  CExRegistryWatcher::SetWatches                           ref:  expksink.h
 *
 *  This method sets watches on those properties of the server registry
 *  about which we want to be notified if they are modified, deleted, or
 *  have children added to them.
 */
PN_RESULT
CExRegistryWatcher::SetWatches()
{
    PN_RESULT hr = PNR_OK;

    hr = m_pRegistry->CreatePropWatch(m_pPropWatch);
    if (PNR_OK != hr)
    {
	return hr;
    }

    hr = m_pPropWatch->Init((IRMAPropWatchResponse*)this);
    if (PNR_OK != hr)
    {
	return hr;
    }

    // We just set a Watch on the "LiveConnections" composite key so
    // that we will be notified when a new child is added
    m_ulConnID = m_pPropWatch->SetWatchByName("LiveConnections");

    return hr;
}


/****************************************************************************
 *  IRMAPropWatchResponse::AddedProp                           ref:  rmamon.h
 *
 *  This method gets called when a new property is added to the server
 *  registry. The parameters identify the Property that got added as well
 *  as the COMPOSITE under which it was added.
 */
STDMETHODIMP
CExRegistryWatcher::AddedProp
(
    const UINT32 id, 
    const RMAPropType propType, 
    const UINT32 ulParentID
)
{
    PN_RESULT	hResult	= PNR_OK;
    IRMABuffer* pBuffer	= NULL;

    // Make sure the added property is a child of "LiveConnections"
    if (ulParentID == m_ulConnID)
    {
	// If we are already handling a live source, disregard any others.
	// This example will only handle one live source at any one time.
	if (m_pPacketSink)
	{
	    return PNR_OK;
	}

	// Assemble the URL of the live source so that we can request the
	// live source object with this URL
	hResult = ConstructSourceURL(id, pBuffer);
	if (PNR_OK != hResult)
	{
	    return hResult;
	}

	// Create an example packet sink to receive the headers and packets
	// from this live source
	CExamplePacketSink* pPacketSink = 
	    new CExamplePacketSink(m_pClassFactory, pBuffer, this);
	pPacketSink->AddRef();

	// Store the object pointer, so we can delete it when the live 
	// source is finished. We will know the source is finished because 
	// we will get StreamDone() calls in the packet sink for all streams 
	// in the source.
	m_pPacketSink = pPacketSink;

	pBuffer->Release();
	pBuffer = NULL;
    }

    return hResult;
}


/****************************************************************************
 *  IRMAPropWatchResponse::ModifiedProp                        ref:  rmamon.h
 *
 *  This method gets called whenever a Property being watched is modified.
 */
STDMETHODIMP
CExRegistryWatcher::ModifiedProp
(
    const UINT32 id, 
    const RMAPropType propType, 
    const UINT32 ulParentID
)
{
    return PNR_OK;
}


/****************************************************************************
 *  IRMAPropWatchResponse::DeletedProp                         ref:  rmamon.h
 *
 *  This method gets called whenever a Property being watched is deleted.
 */
STDMETHODIMP
CExRegistryWatcher::DeletedProp
(
    const UINT32 id, 
    const UINT32 ulParentID
)
{
    return PNR_OK;
}

PN_RESULT
CExRegistryWatcher::ConstructSourceURL(const UINT32 id, 
				       REF(IRMABuffer*) pURL)
{
    PN_RESULT	hr	  = PNR_OK;
    IRMABuffer* pBaseProp = NULL;
    IRMABuffer* pBuffer	  = NULL;
    char	pRegKey[MAX_DISPLAY_NAME] = {0};
    char	pOutput[MAX_DISPLAY_NAME] = {0};

    // Get the id's property name
    m_pRegistry->GetPropName(id, pBaseProp);

    // First, find the mount point for this URL, by looking in the
    // FSMount List of the server config
    sprintf(pRegKey, "%s.LongName", (const char*)pBaseProp->GetBuffer());
    hr = m_pRegistry->GetBufByName((const char*)pRegKey, pBuffer);
    if (PNR_OK != hr)
    {
	return hr;
    }
    sprintf(pRegKey, "%s.MountPoint", (const char*)pBuffer->GetBuffer());
    pBuffer->Release();

    hr = m_pRegistry->GetStrByName((const char*)pRegKey, pBuffer);
    if (PNR_OK != hr)
    {
	return hr;
    }
    strcat(pOutput, (const char*)pBuffer->GetBuffer());
    pBuffer->Release();

    // Then add the file name that was sent by the live source
    sprintf(pRegKey, "%s.FileName", (const char*)pBaseProp->GetBuffer());
    hr = m_pRegistry->GetBufByName((const char*)pRegKey, pBuffer);
    if (PNR_OK != hr)
    {
	return hr;
    }
    strcat(pOutput, (const char*)pBuffer->GetBuffer());
    pBuffer->Release();

    // Finally, create a return buffer and set it to the assembled URL
    hr = m_pClassFactory->CreateInstance(CLSID_IRMABuffer, (void**)&pURL);
    if (PNR_OK != hr)
    {
	return hr;
    }
    pURL->Set((const UCHAR*)pOutput, strlen((const char*)pOutput) + 1);

    // Cleanup
    pBaseProp->Release();
    pBaseProp = NULL;

    return hr;
}

/****************************************************************************
 *  CExamplePacketSink::CExamplePacketSink                   ref:  expksink.h
 *
 *  Constructor
 */
CExamplePacketSink::CExamplePacketSink(IRMACommonClassFactory* pClassFactory,
				       IRMABuffer* pFileName,
				       CExRegistryWatcher* pOwner)
    : m_RefCount	    (0),
      m_pOwner		    (pOwner),
      m_pClassFactory	    (pClassFactory),
      m_pFileName	    (pFileName),
      m_pSourceFinder	    (NULL),
      m_pSource		    (NULL),
      m_ulCurrentStreams    (0),
      m_ulTotalStreams	    (0),
      m_pBytesReceived	    (NULL)
{
    m_pOwner->AddRef();
    m_pClassFactory->AddRef();
    m_pFileName->AddRef();

    // Create a helper object to find the live source for us.
    // The pointer will be returned in our SourceFound() callback.
    m_pSourceFinder = 
	new CExSourceFinderResp(m_pClassFactory, m_pFileName, this);
    m_pSourceFinder->AddRef();
}

/****************************************************************************
 *  CExamplePacketSink::~CExamplePacketSink                 ref:   expksink.h
 *
 *  Destructor
 */
CExamplePacketSink::~CExamplePacketSink(void)
{
    if (m_pOwner)
    {
	m_pOwner->Release();
	m_pOwner = NULL;
    }
}

/****************************************************************************
 *  CExamplePacketSink::Done                                 ref:  expksink.h
 *
 *  Cleanup
 */
void
CExamplePacketSink::Done(void)
{
    if (m_pClassFactory)
    {
	m_pClassFactory->Release();
	m_pClassFactory = NULL;
    }

    if (m_pFileName)
    {
	m_pFileName->Release();
	m_pFileName = NULL;
    }

    if (m_pSourceFinder)
    {
	m_pSourceFinder->Release();
	m_pSourceFinder = NULL;
    }

    if (m_pSource)
    {
	m_pSource->Done();

	m_pSource->Release();
	m_pSource = NULL;
    }

    if (m_pBytesReceived)
    {
	delete []m_pBytesReceived;
	m_pBytesReceived = NULL;
    }

    // This should result in our deletion, since this is the 
    // last object that has a reference on us...
    m_pOwner->LiveSourceDone();
}

/****************************************************************************
 *  IRMARawSinkObject::InitDone                                ref:  rmasrc.h
 *
 *  This method is called in response to the IRMARawSourceObject::Init()
 *  call.
 */
STDMETHODIMP
CExamplePacketSink::InitDone
(
    PN_RESULT status
)
{
    if (PNR_OK != status)
    {
	return PNR_OK;
    }

    // Request the file header from the live source
    m_pSource->GetFileHeader();

    return PNR_OK;
}

/****************************************************************************
 *  IRMARawSinkObject::FileHeaderReady                         ref:  rmasrc.h
 *
 *  This method is called in response to the 
 *  IRMARawSourceObject::GetFileHeader() call.
 */
STDMETHODIMP
CExamplePacketSink::FileHeaderReady
(
    PN_RESULT   status,
    IRMAValues* pHeader
)
{
    if (PNR_OK != status)
    {
        return PNR_OK;
    }

    // Find out how many streams are in this source
    pHeader->GetPropertyULONG32("StreamCount", m_ulTotalStreams);

    // Create an array to store the number of Bytes received on each stream
    m_pBytesReceived = new UINT32[m_ulTotalStreams];

    for (UINT32 i = 0; i < m_ulTotalStreams; i++)
    {
	// Initialize the number of Bytes received
	m_pBytesReceived[i] = 0;

	// Request the stream header for this stream
	m_pSource->GetStreamHeader((UINT16)i);
    }

    return PNR_OK;
}

/****************************************************************************
 *  IRMARawSinkObject::StreamHeaderReady                       ref:  rmasrc.h
 *
 *  This method is called in response to the 
 *  IRMARawSourceObject::GetStreamHeader() call.
 */
STDMETHODIMP
CExamplePacketSink::StreamHeaderReady
(
    PN_RESULT   status,
    IRMAValues* pValues
)
{
    m_ulCurrentStreams++;

    if (m_ulCurrentStreams == m_ulTotalStreams)
    {
	for (UINT32 i = 0; i < m_ulTotalStreams; i++)
	{
	    // Request notifications for all packets in this stream
	    m_pSource->StartPackets((UINT16)i);
	}
    }

    return PNR_OK;
}

/****************************************************************************
 *  IRMARawSinkObject::PacketReady                             ref:  rmasrc.h
 *
 *  This method is called for every packet received from the live source,
 *  but only if the IRMARawSourceObject::StartPackets() call has been made
 *  for the relevant stream.
 */
STDMETHODIMP
CExamplePacketSink::PacketReady
(
    PN_RESULT   status,
    IRMAPacket* pPacket
)
{
    if (PNR_OK != status)
    {
        return PNR_OK;
    }

    IRMABuffer* pBuffer = pPacket->GetBuffer();

    m_pBytesReceived[pPacket->GetStreamNumber()] += pBuffer->GetSize();

    pBuffer->Release();
    pBuffer = NULL;

    return PNR_OK;
}

/****************************************************************************
 *  IRMARawSinkObject::StreamDone                              ref:  rmasrc.h
 *
 *  This method is called when a live stream is complete.
 */
STDMETHODIMP
CExamplePacketSink::StreamDone
(
    UINT16 uStreamNumber
)
{
    m_ulCurrentStreams--;

    if (m_ulCurrentStreams == 0)
    {
	// All streams have finished, so print out our collected statistics 
	printf("\n");
	printf("Live Source Done\n");
	printf("================\n");
	printf("Total Streams Received: %ld\n", m_ulTotalStreams);
	for (UINT32 i = 0; i < m_ulTotalStreams; i++)
	{
	    printf("Stream %ld: %ld Bytes Received\n", i, m_pBytesReceived[i]);
	}
	printf("\n");

	// Begin the cleanup process
	Done();
    }

    return PNR_OK;
}


/****************************************************************************
 *  CExamplePacketSink::SourceFound                          ref:  expksink.h
 *
 *  This method gets called by the CExSourceFinderResponse object when it
 *  has completed its search for the live source.
 */
void
CExamplePacketSink::SourceFound
(
    PN_RESULT status,
    IRMARawSourceObject* pSource
)
{
    if (PNR_OK == status)
    {
	m_pSource = pSource;
	m_pSource->AddRef();

	// Initialize the source, passing ourselves as a notification sink
	m_pSource->Init((IUnknown*)(IRMARawSinkObject*)this);
    }
}


// IUnknown COM Interface Methods

/****************************************************************************
 *  IUnknown::AddRef                                            ref:  pncom.h
 *
 *  This routine increases the object reference count in a thread safe
 *  manner. The reference count is used to manage the lifetime of an object.
 *  This method must be explicitly called by the user whenever a new
 *  reference to an object is used.
 */
STDMETHODIMP_(UINT32)
CExamplePacketSink::AddRef(void)
{
    return InterlockedIncrement(&m_RefCount);
}


/****************************************************************************
 *  IUnknown::Release                                           ref:  pncom.h
 *
 *  This routine decreases the object reference count in a thread safe
 *  manner, and deletes the object if no more references to it exist. It must
 *  be called explicitly by the user whenever an object is no longer needed.
 */
STDMETHODIMP_(UINT32)
CExamplePacketSink::Release(void)
{
    if (InterlockedDecrement(&m_RefCount) > 0)
    {
	return m_RefCount;
    }

    delete this;
    return 0;
}


/****************************************************************************
 *  IUnknown::QueryInterface                                    ref:  pncom.h
 *
 *  This routine indicates which interfaces this object supports. If a given
 *  interface is supported, the object's reference count is incremented, and
 *  a reference to that interface is returned. Otherwise a NULL object and
 *  error code are returned. This method is called by other objects to
 *  discover the functionality of this object.
 */
STDMETHODIMP
CExamplePacketSink::QueryInterface
(
    REFIID interfaceID,
    void** ppInterfaceObj
)
{
    // By definition all COM objects support the IUnknown interface
    if (IsEqualIID(interfaceID, IID_IUnknown))
    {
	AddRef();
	*ppInterfaceObj = (IUnknown*)(IRMAPlugin*)this;
	return PNR_OK;
    }

    // IRMARawSinkObject interface is supported
    else if (IsEqualIID(interfaceID, IID_IRMARawSinkObject))
    {
	AddRef();
	*ppInterfaceObj = (IRMARawSinkObject*)this;
	return PNR_OK;
    }

    // No other interfaces are supported
    *ppInterfaceObj = NULL;
    return PNR_NOINTERFACE;
}


/****************************************************************************
 *  CExSourceFinderResp::CExSourceFinderResp                 ref:  expksink.h
 *
 *  Constructor
 */
CExSourceFinderResp::CExSourceFinderResp
(
    IRMACommonClassFactory* pClassFactory,
    IRMABuffer* pFileName,
    CExamplePacketSink* pOwner
)
    : m_RefCount	    (0),
      m_pOwner		    (pOwner),
      m_pClassFactory	    (pClassFactory),
      m_pFileName	    (pFileName),
      m_pSourceFinder	    (NULL)
{
    PN_RESULT hr = PNR_OK;
    
    m_pOwner->AddRef();
    m_pClassFactory->AddRef();
    m_pFileName->AddRef();

    // Find the live source with the IRMASourceFinderObject interface
    hr = m_pClassFactory->CreateInstance(CLSID_IRMASourceFinderObject, 
					 (void**)&m_pSourceFinder);
    if (PNR_OK == hr)
    {
	m_pSourceFinder->Init((IRMASourceFinderResponse*)this);
    }
}

/****************************************************************************
 *  CExSourceFinderResp::~CExSourceFinderResp               ref:   expksink.h
 *
 *  Destructor
 */
CExSourceFinderResp::~CExSourceFinderResp(void)
{
    if (m_pOwner)
    {
	m_pOwner->Release();
	m_pOwner = NULL;
    }

    if (m_pClassFactory)
    {
	m_pClassFactory->Release();
	m_pClassFactory = NULL;
    }

    if (m_pFileName)
    {
	m_pFileName->Release();
	m_pFileName = NULL;
    }

    if (m_pSourceFinder)
    {
	m_pSourceFinder->Done();

	m_pSourceFinder->Release();
	m_pSourceFinder = NULL;
    }
}

/****************************************************************************
 *  IRMASourceFinderResponse::InitDone                         ref:  rmasrc.h
 *
 *  This method is called in response to the IRMASourceFinderObject::Init()
 *  call.
 */
STDMETHODIMP
CExSourceFinderResp::InitDone
(
    PN_RESULT status
)
{
    PN_RESULT hr = PNR_OK;
    IRMARequest* pRequest = NULL;

    if (PNR_OK != status)
    {
	// Init failure, so we can't even attempt to find the source
	m_pOwner->SourceFound(status, (IRMARawSourceObject*)NULL);
	m_pOwner->Release();
	m_pOwner = NULL;
	return PNR_OK;
    }

    // Create a request object
    hr = m_pClassFactory->CreateInstance(CLSID_IRMARequest, 
	(void**)&pRequest);
    if (PNR_OK != hr)
    {
	return PNR_OK;
    }

    // Attempt to locate the live source object, by passing the URL 
    // we constructed from the server registry to the source finder
    pRequest->SetURL((const char*)m_pFileName->GetBuffer());
    m_pSourceFinder->Find(pRequest);
    pRequest->Release();
    pRequest = NULL;

    return PNR_OK;
}

/****************************************************************************
 *  IRMASourceFinderResponse::FindDone                         ref:  rmasrc.h
 *
 *  This method gets called when the Source Finder is done searching for
 *  the requested live stream source. If successful, the second parameter
 *  will contain an IUnknown* to the live source object itself.
 */
STDMETHODIMP
CExSourceFinderResp::FindDone
(
    PN_RESULT status,
    IUnknown* pUnknown
)
{
    PN_RESULT hResult = PNR_OK;
    IRMARawSourceObject* pSource = NULL;

    if (status != PNR_OK)
    {
	return PNR_OK;
    }

    // Query for the IRMARawSourceObject so we can request that 
    // we be notified about headers and packets
    hResult = pUnknown->QueryInterface(IID_IRMARawSourceObject, 
	(void**)&pSource);

    // Notify our owner that our job is done. The first parameter will 
    // tell the owner if pSource is a valid Source pointer or not.
    m_pOwner->SourceFound(hResult, pSource);

    if (PNR_OK == hResult)
    {
	pSource->Release();
	pSource = NULL;
    }

    // Our job is done, so we will just release our reference 
    // on our owner and wait around to be destroyed.
    m_pOwner->Release();
    m_pOwner = NULL;

    return PNR_OK;
}


// IUnknown COM Interface Methods

/****************************************************************************
 *  IUnknown::AddRef                                            ref:  pncom.h
 *
 *  This routine increases the object reference count in a thread safe
 *  manner. The reference count is used to manage the lifetime of an object.
 *  This method must be explicitly called by the user whenever a new
 *  reference to an object is used.
 */
STDMETHODIMP_(UINT32)
CExSourceFinderResp::AddRef(void)
{
    return InterlockedIncrement(&m_RefCount);
}


/****************************************************************************
 *  IUnknown::Release                                           ref:  pncom.h
 *
 *  This routine decreases the object reference count in a thread safe
 *  manner, and deletes the object if no more references to it exist. It must
 *  be called explicitly by the user whenever an object is no longer needed.
 */
STDMETHODIMP_(UINT32)
CExSourceFinderResp::Release(void)
{
    if (InterlockedDecrement(&m_RefCount) > 0)
    {
	return m_RefCount;
    }

    delete this;
    return 0;
}


/****************************************************************************
 *  IUnknown::QueryInterface                                    ref:  pncom.h
 *
 *  This routine indicates which interfaces this object supports. If a given
 *  interface is supported, the object's reference count is incremented, and
 *  a reference to that interface is returned. Otherwise a NULL object and
 *  error code are returned. This method is called by other objects to
 *  discover the functionality of this object.
 */
STDMETHODIMP
CExSourceFinderResp::QueryInterface
(
    REFIID interfaceID,
    void** ppInterfaceObj
)
{
    // By definition all COM objects support the IUnknown interface
    if (IsEqualIID(interfaceID, IID_IUnknown))
    {
	AddRef();
	*ppInterfaceObj = (IUnknown*)(IRMAPlugin*)this;
	return PNR_OK;
    }

    // IRMASourceFinderResponse interface is supported
    else if (IsEqualIID(interfaceID, IID_IRMASourceFinderResponse))
    {
	AddRef();
	*ppInterfaceObj = (IRMASourceFinderResponse*)this;
	return PNR_OK;
    }

    // No other interfaces are supported
    *ppInterfaceObj = NULL;
    return PNR_NOINTERFACE;
}
