/****************************************************************************
 *  $Id: expksink.h,v 1.1 1998/12/09 22:27:00 darrens Exp $
 *
 *  Copyright (C) 1998 RealNetworks, Inc.
 *  All rights reserved.
 *  
 *  http://www.real.com/devzone
 *
 *  This program contains proprietary information of RealNetworks, Inc.,
 *  and is licensed subject to restrictions on use and distribution.
 *
 *  expksink.h
 * 
 *  Example Live Packet Sink plug-in for the RealMedia Architecture (RMA).
 */
#ifndef _EXPKSINK_H_
#define _EXPKSINK_H_


/****************************************************************************
 * Constants
 */
#define MY_PLUGIN_VERSION    0
#define MY_DESCRIPTION       "Example Live Packet Sink Plug-in"
#define MY_COPYRIGHT         "(c) 1998 RealNetworks, All rights reserved."
#define MY_MORE_INFO_URL     "http://www.real.com"

/****************************************************************************
 * Forward Declarations
 */
class CExamplePacketSink;
class CExSourceFinderResp;

/****************************************************************************
 *
 *  The classes that follow inherit the minimum number of interfaces 
 *  required to create an RMA Live Packet Sink plug-in. 
 *
 *  The IRMAPropWatchResponse interface is used by the plugin to find out
 *  when a live source connects to the RMA server. This monitoring interface
 *  will provide a notification when the live source information is written
 *  into the server registry, so that the plugin can find the live source
 *  and register itself as a packet sink for that source. For more 
 *  information about this interface, see the Example Monitoring Plugin in
 *  the RMA SDK.
 *
 *  The IRMARawSinkObject interface provides notifications about all packets 
 *  that are served from a live source connected to the RMA server, as well 
 *  as the relevant file and stream headers. 
 *  
 *  The IRMASourceFinderResponse interface notifies us when the live source
 *  we requested it to find has been located. We ask the SourceFinder to
 *  locate a live source by passing it the FileName, which we retrieved from
 *  the server registry. It then calls us back on this interface when the
 *  find operation is complete.
 *
 *  The IRMAPlugin interface is required by all plug-ins to handle 
 *  fundamental plug-in operations. 
 *
 *  Since this is a COM object, this class also inherits COM's IUnknown 
 *  interface to handle object reference counting and interface query.
 */

/****************************************************************************
 *
 *  CExRegistryWatcher Class
 *
 *  This class watches the server registry for new Live Connections to the
 *  server. When a Live Connection is detected, an instance of the 
 *  CExamplePacketSink class is created to act as a packet sink for that
 *  particular live source. When the CExamplePacketSink object indicates
 *  that the live source is done, the object is cleaned up, and this class
 *  continues watching the registry for the next live source to connect.
 */
class CExRegistryWatcher : public IRMAPropWatchResponse,
                           public IRMAPlugin
{
public:
    CExRegistryWatcher (void);

    void LiveSourceDone(void);

    /************************************************************************
     *  IRMAPropWatchResponse Interface Methods                ref:  rmamon.h
     */
    STDMETHOD(AddedProp)		(THIS_
					const UINT32	    id,
					const RMAPropType   propType,
					const UINT32	    ulParentID);
    STDMETHOD(ModifiedProp)		(THIS_
					const UINT32	    id,
					const RMAPropType   propType,
					const UINT32	    ulParentID);
    STDMETHOD(DeletedProp)		(THIS_
					const UINT32	    id,
					const UINT32	    ulParentID);


    /************************************************************************
     *  IRMAPlugin Interface Methods                         ref:  rmaplugn.h
     */
    STDMETHOD(GetPluginInfo)
	    (THIS_ 
	      REF(BOOL)        bLoadMultiple,
	      REF(const char*) pDescription,
	      REF(const char*) pCopyright,
	      REF(const char*) pMoreInfoURL,
	      REF(UINT32)      versionNumber
	    );
    STDMETHOD(InitPlugin) (THIS_ IUnknown* pContext);


    /************************************************************************
     *  IUnknown COM Interface Methods                          ref:  pncom.h
     */
    STDMETHOD (QueryInterface ) (THIS_ REFIID ID, void** ppInterfaceObj);
    STDMETHOD_(UINT32, AddRef ) (THIS);
    STDMETHOD_(UINT32, Release) (THIS);


private:
    /****** Private Class Methods ******************************************/
    ~CExRegistryWatcher(void);

    PN_RESULT SetWatches	 (void);
    PN_RESULT ConstructSourceURL (const UINT32 id, REF(IRMABuffer*) pURL);

    /****** Private Class Variables ****************************************/
    INT32		    m_RefCount;		// Object's reference count
    IRMACommonClassFactory* m_pClassFactory;	// Creates common RMA classes
    IRMAPNRegistry*	    m_pRegistry;	// Accesses server registry
    IRMAPropWatch*	    m_pPropWatch;
    UINT32		    m_ulConnID;
    CExamplePacketSink*	    m_pPacketSink;

    /****** Private Static Class Variables *********************************/
    static const char*      zm_pDescription;
    static const char*      zm_pCopyright;
    static const char*      zm_pMoreInfoURL;

    PRIVATE_DESTRUCTORS_ARE_NOT_A_CRIME       // Avoids GCC compiler warning
};


/****************************************************************************
 *
 *  CExamplePacketSink Class
 *
 *  This class acts as the actual packet sink for the live source, receiving
 *  notifications when each header or packet is received by the server. When
 *  an object of this class is first created, it is passed the FileName of
 *  the live source. Immediately, a CExSourceFinderResp object is created to
 *  handle the task of searching for the live source in the server. Once
 *  that helper object signals that it has found the live source, this class
 *  sets itself up as a packet sink, and waits for callback notifications
 *  for headers and packets.
 */
class CExamplePacketSink : public IRMARawSinkObject
{
public:
    CExamplePacketSink (IRMACommonClassFactory* pClassFactory,
			IRMABuffer* pFileName,
			CExRegistryWatcher* pOwner);

    void SourceFound(PN_RESULT status, IRMARawSourceObject* pSource);
    void Done(void);

    /************************************************************************
     *  IRMARawSinkObject Interface Methods                    ref:  rmasrc.h
     */
    STDMETHOD(InitDone)			(THIS_
    					PN_RESULT   status);
    STDMETHOD(FileHeaderReady)		(THIS_
    					PN_RESULT   status,
    					IRMAValues* pHeader);
    STDMETHOD(StreamHeaderReady)	(THIS_
					PN_RESULT   status,
					IRMAValues* pHeader);
    STDMETHOD(PacketReady)		(THIS_
					PN_RESULT   status,
					IRMAPacket* pPacket);
    STDMETHOD(StreamDone)		(THIS_
					UINT16      unStreamNumber);

    /************************************************************************
     *  IUnknown COM Interface Methods                          ref:  pncom.h
     */
    STDMETHOD (QueryInterface ) (THIS_ REFIID ID, void** ppInterfaceObj);
    STDMETHOD_(UINT32, AddRef ) (THIS);
    STDMETHOD_(UINT32, Release) (THIS);


private:
    /****** Private Class Methods ******************************************/
    ~CExamplePacketSink(void);

    /****** Private Class Variables ****************************************/
    INT32		    m_RefCount;		// Object's reference count
    CExRegistryWatcher*	    m_pOwner;
    CExSourceFinderResp*    m_pSourceFinder;	// Locates the live source
    IRMACommonClassFactory* m_pClassFactory;	// Creates common RMA classes
    IRMABuffer*		    m_pFileName;	// The live source file name
    IRMARawSourceObject*    m_pSource;		// The live source itself
    UINT32		    m_ulCurrentStreams; // Current # of live streams
    UINT32		    m_ulTotalStreams;	// Total streams in file
    UINT32*		    m_pBytesReceived;

    PRIVATE_DESTRUCTORS_ARE_NOT_A_CRIME       // Avoids GCC compiler warning
};


/****************************************************************************
 *
 *  CExSourceFinderResp Class
 *
 *  This class handles the task of finding a live source with a particular
 *  FileName in the server core. The FileName is passed into the constructor
 *  and the IRMASourceFinderObject interface is used to search for the 
 *  source. This class signals that its job is complete by passing the 
 *  pointer to the source object back to its creator (the CExamplePacketSink
 *  object).
 */
class CExSourceFinderResp : public IRMASourceFinderResponse
{
public:
    CExSourceFinderResp (IRMACommonClassFactory* pClassFactory,
			 IRMABuffer* pFileName,
			 CExamplePacketSink* pOwner);

    /************************************************************************
     *  IRMASourceFinderResponse Interface Methods             ref:  rmasrc.h
     */
    STDMETHOD(InitDone)			(THIS_
    					PN_RESULT   status);
    STDMETHOD(FindDone)			(THIS_
					PN_RESULT   status,
					IUnknown*   pUnknown);

    /************************************************************************
     *  IUnknown COM Interface Methods                          ref:  pncom.h
     */
    STDMETHOD (QueryInterface ) (THIS_ REFIID ID, void** ppInterfaceObj);
    STDMETHOD_(UINT32, AddRef ) (THIS);
    STDMETHOD_(UINT32, Release) (THIS);


private:
    /****** Private Class Methods ******************************************/
    ~CExSourceFinderResp(void);

    /****** Private Class Variables ****************************************/
    INT32		    m_RefCount;		// Object's reference count
    CExamplePacketSink*	    m_pOwner;
    IRMACommonClassFactory* m_pClassFactory;	// Creates common RMA classes
    IRMABuffer*		    m_pFileName;	// The live source file name
    IRMASourceFinderObject* m_pSourceFinder;	// Locates the live source

    PRIVATE_DESTRUCTORS_ARE_NOT_A_CRIME       // Avoids GCC compiler warning
};


#endif  /* ifndef _EXPKSINK_H_ */
