/****************************************************************************
 * 
 *  $Id: pcmrendr.cpp,v 1.9 1998/07/14 03:33:01 ehodge Exp $
 *
 *  Copyright (C) 1995,1996,1997 RealNetworks, Inc.
 *  All rights reserved.
 *  
 *  http://www.real.com/devzone
 *
 *  This program contains proprietary information of RealNetworks, Inc.,
 *  and is licensed subject to restrictions on use and distribution.
 * 
 *  pcmrendr.cpp
 *
 *
 *  PCM audio Renderer for RealMedia Architecture (RMA).  
 *
 */


/****************************************************************************
 * Defines
 */
#define INITGUID     /* Interface ID's */
#define STRICT


/****************************************************************************
 * Includes
 */
#include <stdio.h>
#include <signal.h>

#ifdef _UNIX
    #include <sys/types.h>
    #include <sys/stat.h>
    #include <unistd.h>
    #include <fcntl.h>
#endif

#include "pntypes.h"
#include "pncom.h"
#include "rmapckts.h"
#include "rmafiles.h"
#include "rmacore.h"
#include "rmarendr.h"
#include "rmaplugn.h"
#include "rmaausvc.h"
#include "rmacomm.h"

#include "fivemque.h"
#include "pcmrendr.h"
#include "netswap.h"
#include "pcmrendr.ver"


/****************************************************************************
 *  RMACreateInstance                                        ref:  rmaplugn.h
 *
 *  This routine creates a new instance of the CPCMRenderer class.
 *  It is called when the RMA core application is launched, and whenever
 *  a stream associated with this plug-in needs to be rendered.
 */
STDAPI
RMACreateInstance(IUnknown** ppIUnknown)
{
    *ppIUnknown = (IUnknown*)(IRMAPlugin*)new CPCMRenderer();
    if (*ppIUnknown)
    {
	(*ppIUnknown)->AddRef();
	return PNR_OK;
    }
    return PNR_OUTOFMEMORY;
}


/****************************************************************************
 *  CPCMRenderer static variables                           ref:  pcmrendr.h
 *
 *  These variables are passed to the RMA core to provide information about
 *  this plug-in. They are required to be static in order to remain valid
 *  for the lifetime of the plug-in.
 */
const char* CPCMRenderer::zm_pDescription    = "PCM Renderer Plugin";
const char* CPCMRenderer::zm_pCopyright      = "(c) 1995,1996,1997 RealNetworks, All rights reserved.";
const char* CPCMRenderer::zm_pMoreInfoURL    = "http://www.real.com";
const char* CPCMRenderer::zm_ppStreamMimeTypes[] = { "audio/x-pn-wav", NULL };


/****************************************************************************
 *  CPCMRenderer::CPCMRenderer                               ref:  pcmrendr.h
 *
 *  Constructor
 */
CPCMRenderer::CPCMRenderer()
	: m_lRefCount(0)
	, m_pContext(NULL)
        , m_pStream(NULL)
        , m_pPlayer(NULL)
        , m_pHeader(NULL)
        , m_pLastPacket(NULL)
        , m_ulLastTime(0)
	, m_logfile(0)
	, m_pcmfile(0)
	, m_pAudioPlayer(0)
	, m_pAudioStream(0)
	, m_pCommonClassFactory(0)
	, m_bSwapSampleBytes (FALSE)
	, m_bZeroOffsetPCM (FALSE)
	, m_bIsGapInStreaming(TRUE)
	, m_bInSeekMode(FALSE)
{ 
}


/****************************************************************************
 *  IRMAPlugin::GetPluginInfo                                ref:  rmaplugn.h
 *
 *  This routine returns descriptive information about the plug-in, most
 *  of which is used in the About box of the user interface. It is called
 *  when the RMA core application is launched.
 */
STDMETHODIMP
CPCMRenderer::GetPluginInfo
(
    REF(BOOL)        /*OUT*/ bLoadMultiple,
    REF(const char*) /*OUT*/ pDescription,
    REF(const char*) /*OUT*/ pCopyright,
    REF(const char*) /*OUT*/ pMoreInfoURL,
    REF(UINT32)      /*OUT*/ ulVersionNumber
)
{
    bLoadMultiple = TRUE;   // Must be true for file formats.

    pDescription    = zm_pDescription;
    pCopyright	    = zm_pCopyright;
    pMoreInfoURL    = zm_pMoreInfoURL;
    ulVersionNumber = TARVER_ULONG32_VERSION;

    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::GetRendererInfo                            ref:  rmarendr.h
 *
 *  This routine returns crucial information required to associate this
 *  plug-in with a given stream MIME type. This information tells the RMA
 *  core which Renderer to use to display a particular type of stream. This
 *  method is called when the RMA core application is launched.
 */
STDMETHODIMP
CPCMRenderer::GetRendererInfo
(
    REF(const char**) /*OUT*/ pStreamMimeTypes,
    REF(UINT32)       /*OUT*/ unInitialGranularity
)
{
    pStreamMimeTypes = zm_ppStreamMimeTypes;
    unInitialGranularity = 50;  // OnTimeSync() every 50ms (this is the same as
				// the TIME_PER_PACKET in wvffplin.cpp
    m_ulGranularity = unInitialGranularity;
    return PNR_OK;
}


/****************************************************************************
 *  IRMAPlugin::InitPlugin                                   ref:  rmaplugn.h
 *
 *  This routine performs initialization steps such as determining if
 *  required interfaces are available. It is called when the RMA core 
 *  application is launched, and whenever a stream associated with this
 *  plug-in needs to be rendered.
 */
STDMETHODIMP
CPCMRenderer::InitPlugin(IUnknown* /*IN*/ pContext)
{
    m_pContext = pContext;
    m_pContext->AddRef();
    
    if (PNR_OK !=
	m_pContext->QueryInterface(IID_IRMACommonClassFactory, (void**)&m_pCommonClassFactory))
    {
	return PNR_INVALID_PARAMETER;
    }

    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::StartStream                                ref:  rmarendr.h
 *
 *  The RMA core calls this routine to provide access to the stream and
 *  player. It is called when the plug-in is being initialized.
 */
STDMETHODIMP
CPCMRenderer::StartStream
(
    IRMAStream*	    pStream,
    IRMAPlayer*	    pPlayer
)
{
    // Save for later use!
    m_pStream  = pStream;
    m_pPlayer  = pPlayer;

    if (m_pStream ) m_pStream->AddRef();
    if (m_pPlayer ) m_pPlayer->AddRef();

    // get interface to audio player
    if ( PNR_OK != m_pPlayer->QueryInterface(IID_IRMAAudioPlayer, (void**) &m_pAudioPlayer ))
    {
        return !PNR_OK;
    }

    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnHeader                                   ref:  rmarendr.h
 *
 *  This routine is passed the stream header object created by the associated
 *  file format plug-in. Use the methods defined in the IRMAValues interface
 *  to access the stream header data. This method is called from the RMA
 *  core when the plug-in is being initialized.
 */
STDMETHODIMP
CPCMRenderer::OnHeader(IRMAValues* pHeader)
{
    // wave format stuff
    PCMHEADER* pPcmFormatRecord;
    PCMHEADER  tmpPcmFormatRecord;

    // Keep this for later use...
    m_pHeader = pHeader;
    m_pHeader->AddRef();

    IRMABuffer* pOpaqueData = 0;
    UINT32 ulStreamNumber;
    UINT32 ulMaxBitRate;
    UINT32 ulAvgBitRate;
    UINT32 ulMaxPacketSize;
    UINT32 ulAvgPacketSize;
    UINT32 ulStartTime;
    UINT32 ulPreroll;
    UINT32 ulDuration;
    IRMABuffer* pStreamName = 0;
    IRMABuffer* pMimeType = 0;

    pHeader->GetPropertyBuffer ("OpaqueData",       pOpaqueData);
    pHeader->GetPropertyULONG32("StreamNumber",     ulStreamNumber);
    pHeader->GetPropertyULONG32("MaxBitRate",       ulMaxBitRate);
    pHeader->GetPropertyULONG32("AvgBitRate",       ulAvgBitRate);
    pHeader->GetPropertyULONG32("MaxPacketSize",    ulMaxPacketSize);
    pHeader->GetPropertyULONG32("AvgPacketSize",    ulAvgPacketSize);
    pHeader->GetPropertyULONG32("StartTime",        ulStartTime);
    pHeader->GetPropertyULONG32("Preroll",          ulPreroll);
    pHeader->GetPropertyULONG32("Duration",         ulDuration);
    pHeader->GetPropertyCString("StreamName",       pStreamName);
    pHeader->GetPropertyCString("MimeType",         pMimeType);

    BOOL bBigEndian = TestBigEndian(); //False if Intel

    AudioPCMHEADER audioPCMHEADER;
    BOOL bIsOldHeader = FALSE;
    if(pOpaqueData->GetSize() != audioPCMHEADER.static_size())
    {
	/*don't attempt to touch the bytes beyond the end of the buffer:*/
	bIsOldHeader = TRUE;
    }
    else
    {
	audioPCMHEADER.unpack(pOpaqueData->GetBuffer(), pOpaqueData->GetSize());
    }
    /*Check if this is from an old (preview release of G2 / rmasdk beta-9
     * or earlier) file format:
     */
    if(bIsOldHeader  ||
	    audioPCMHEADER.usVersion != 0  ||
	    audioPCMHEADER.usMagicNumberTag !=
	    AUDIO_PCMHEADER_MAGIC_NUMBER)
    {
	pPcmFormatRecord = (PCMHEADER*) pOpaqueData->GetBuffer();

	/*Make sure endianness is ok by swapping bytes if and only if
	 * usChannels is > 0xff; don't just count on this being net-
	 * endian as some preview-release file formats did not send
	 * the data net-endian:
	 */
	if(pPcmFormatRecord->usChannels > 0xff)
	{
    	    SwapWordBytes((UINT16*)&pPcmFormatRecord->usFormatTag, 1);
    	    SwapWordBytes((UINT16*)&pPcmFormatRecord->usChannels, 1);
    	    SwapDWordBytes((UINT32*)&pPcmFormatRecord->ulSamplesPerSec, 1);
    	    SwapWordBytes((UINT16*)&pPcmFormatRecord->usBitsPerSample, 1);
    	    SwapWordBytes((UINT16*)&pPcmFormatRecord->usSampleEndianness, 1);
	}
    }
    else /*is valid pAudioPCMHEADER and is already in the correct byte-order
	  * as performed by audioPCMHEADER.unpack():
	  */
    {
	pPcmFormatRecord = &tmpPcmFormatRecord;
	switch(audioPCMHEADER.usVersion)
	{
	    case 0:
	    {
		/*PCM format stuff starts 4 bytes into the buffer (which is
		 * right after audioPCMHEADER.usVersion and
		 * audioPCMHEADER.usMagicNumberTag:
		 */
    		pPcmFormatRecord->usFormatTag = audioPCMHEADER.usFormatTag;
    		pPcmFormatRecord->usChannels = audioPCMHEADER.usChannels;
    		pPcmFormatRecord->ulSamplesPerSec =
			audioPCMHEADER.ulSamplesPerSec;
    		pPcmFormatRecord->usBitsPerSample =
			audioPCMHEADER.usBitsPerSample;
    		pPcmFormatRecord->usSampleEndianness =
			audioPCMHEADER.usSampleEndianness;
		break;
	    }
	    default:
		PN_RELEASE(pOpaqueData);
		PN_RELEASE(pStreamName);
		PN_RELEASE(pMimeType);
		return PNR_UNEXPECTED;
	}
    }


    if (pPcmFormatRecord->usFormatTag == PN_PCM_ZERO_OFFSET &&
	pPcmFormatRecord->usBitsPerSample == 8)
    {
	m_bZeroOffsetPCM = TRUE;
    }

    if (pPcmFormatRecord->usBitsPerSample == 16 &&
	pPcmFormatRecord->usSampleEndianness == PN_PCM_BIGENDIAN) 
    {
	m_bSwapSampleBytes = TRUE;
    }

    m_audioFmt.uChannels 		= pPcmFormatRecord->usChannels;
    m_audioFmt.uBitsPerSample 	= pPcmFormatRecord->usBitsPerSample;
    m_audioFmt.ulSamplesPerSec 	= pPcmFormatRecord->ulSamplesPerSec;
    m_audioFmt.uMaxBlockSize  	= (UINT16)ulMaxPacketSize;

    m_pAudioPlayer->CreateAudioStream( &m_pAudioStream );
    m_pAudioStream->Init(&m_audioFmt, pHeader);

    /* Add IRMADryNotification notification to get notifications if the 
     * audio stream runs dry.
     */
    m_pAudioStream->AddDryNotification(this);

    // set up initial time
    m_ulCurrentTime = 0;

    if (pOpaqueData)	pOpaqueData->Release();
    if (pStreamName)	pStreamName->Release();
    if (pMimeType)	pMimeType->Release();

    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnBegin                                   ref:  rmarendr.h
 *
 *  This routine is called by the RMA core to inform the Renderer that 
 *  playback has just begun or has been resumed. The stream's time value just
 *  after resuming the playback is provided.
 */
STDMETHODIMP
CPCMRenderer::OnBegin(UINT32 ulTime)
{
    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::GetDisplayType                             ref:  rmarendr.h
 *
 *  This routine returns the preferred display type of the Renderer. Any
 *  other additional information necessary for the display can also be
 *  provided. This method is called by the RMA core when the plug-in is being
 *  initialized.
 */
STDMETHODIMP
CPCMRenderer::GetDisplayType
(
    REF(RMA_DISPLAY_TYPE)   ulFlags,
    REF(IRMABuffer*)	    pBuffer
)
{
    //  Our renderer needs no visible display.
    ulFlags = RMA_DISPLAY_NONE;

    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnPacket                                   ref:  rmarendr.h
 *
 *  This routine is passed the packet object streamed by the associated
 *  file format plug-in. It is called by the RMA core when a packet is due
 *  to be delivered. In most cases, the actual rendering of the packet is
 *  done in this method. The stream's time offset with respect to the master
 *  timeline is provided.
 */
STDMETHODIMP CPCMRenderer::OnPacket(IRMAPacket* pPacket, LONG32 lTimeOffset)
{
    /* Ignore pre-seek packets. Some renderers may want to use them though */
    if (m_bInSeekMode)
    {
	return PNR_OK;
    }

    // did we get a valid packet?
    if (pPacket->IsLost())
    {
	m_bIsGapInStreaming = TRUE;
	return PNR_OK;
    }

    IRMABuffer* pBuffer = 0;
    pBuffer = pPacket->GetBuffer();

    /* subtract the offset from the packet time before sending to Audio services */
    LONG32 lRealTime = pPacket->GetTime(); 
    if (lTimeOffset > lRealTime)
    {
    	lRealTime = 0;
    }
    else
    {
    	lRealTime -= lTimeOffset;
    }

    IRMABuffer* pNewBuffer = 0;
    /* hmmm.. we will modify data, so need to make a copy since
     * we are not allowed to modify an existing COM object since it
     * might be in use by other parts of the system 
     * (e.g. one more PCM renderer)
     */
    if (m_bSwapSampleBytes || m_bZeroOffsetPCM)
    {
	 if (PNR_OK != 
	     m_pCommonClassFactory->CreateInstance(CLSID_IRMABuffer, (void**) &pNewBuffer))
	 {
	     return PNR_OUTOFMEMORY;
	 }

	 pNewBuffer->Set(pBuffer->GetBuffer(), pBuffer->GetSize());
    }
    else
    {
	pNewBuffer = pBuffer;
	pNewBuffer->AddRef();
    }

    // byte swap the samples if necessary
    if (m_bSwapSampleBytes)
    {
	UCHAR *pSampleBuf;
	UINT32 len;
	pNewBuffer->Get(pSampleBuf, len);
    	SwapWordBytes((UINT16 *)pSampleBuf, len);
    }

    // convert to two's complement if necessary (only used for 8-bit samples)
    if (m_bZeroOffsetPCM)
    {
	UCHAR *pSampleBuf;
	UINT32 len;
	pNewBuffer->Get(pSampleBuf, len);

	for (UINT32 i = 0; i < len; i++)
	{
	    pSampleBuf[i] -= 128;
	}
    }

    /* Store these packets so that they can be written later
     * on OnDryNotification
     */
    RMAAudioData*  pAudioData = new RMAAudioData;
    pAudioData->pData = pNewBuffer;
    pAudioData->ulAudioTime = lRealTime; //pPacket->GetTime(); 

    if (m_bIsGapInStreaming)
    {
	pAudioData->uAudioStreamType	= TIMED_AUDIO;
	m_bIsGapInStreaming		= FALSE;
    }
    else
    {
	pAudioData->uAudioStreamType	= STREAMING_AUDIO;
    }

    m_PacketQueue.Add((void*) pAudioData);

    m_ulCurrentTime  = pAudioData->ulAudioTime;

    pBuffer->Release();
    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnTimeSync                                 ref:  rmarendr.h
 *
 *  This routine is called by the RMA core periodically passing the current
 *  playback time. The Renderer should use this time value to synchronize the
 *  playback of various streams.
 */
STDMETHODIMP
CPCMRenderer::OnTimeSync(UINT32 ulTime)
{
    m_ulLastTime = ulTime;
    
    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnPreSeek                                  ref:  rmarendr.h
 *
 *  This routine is called by the RMA core before a seek is about to occur.
 *  The stream's time value just before the seek, and the time value just
 *  after the seek are provided.
 */
STDMETHODIMP
CPCMRenderer::OnPreSeek(UINT32 ulOldTime, UINT32 ulNewTime)
{
    m_bIsGapInStreaming = TRUE;
    m_bInSeekMode	= TRUE;

    RMAAudioData*  pAudioData = (RMAAudioData*) m_PacketQueue.Remove();
    while(pAudioData)
    {
	pAudioData->pData->Release();
	delete pAudioData;
        pAudioData = (RMAAudioData*) m_PacketQueue.Remove();
    }

    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnPostSeek                                 ref:  rmarendr.h
 *
 *  This routine is called by the RMA core just after a seek has occured.
 *  The stream's time value just before the seek, and the time value just
 *  after the seek are provided.
 */
STDMETHODIMP
CPCMRenderer::OnPostSeek(UINT32 ulOldTime, UINT32 ulNewTime)
{
    m_bInSeekMode   = FALSE;
    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnPause                                    ref:  rmarendr.h
 *
 *  This routine is called by the RMA core just after a pause has occured.
 *  The stream's time value just before pausing is provided.
 */
STDMETHODIMP
CPCMRenderer::OnPause(UINT32 ulTime)
{
    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnBuffering                                ref:  rmarendr.h
 *
 *  This routine is called by the RMA core to inform the Renderer that
 *  buffering of data is occuring. The reason for buffering (e.g. start-up
 *  of stream, seek has occured, network congestion, etc.), as well as the
 *  percentage of the buffering process complete are provided.
 */
STDMETHODIMP
CPCMRenderer::OnBuffering(UINT32 ulFlags, UINT16 unPercentComplete)
{
    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnEndofPackets                             ref:  rmarendr.h
 *
 *  Called by client engine to inform the renderer that all the
 *  packets have been delivered. However, if the user seeks before
 *  EndStream() is called, renderer may start getting packets again
 *  and the client engine will eventually call this function again.
 */
STDMETHODIMP
CPCMRenderer::OnEndofPackets(void)
{
    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::EndStream                                  ref:  rmarendr.h
 *
 *  This routine is called by the RMA core when the rendering of the stream
 *  has completed. Deallocation of any resources should be done here.
 */
STDMETHODIMP
CPCMRenderer::EndStream()
{
    // We're done with these...
    if (m_pHeader) m_pHeader->Release();	    m_pHeader	   = NULL;
    if (m_pLastPacket) m_pLastPacket->Release();    m_pLastPacket  = NULL;
    if (m_pContext) m_pContext->Release();	    m_pContext	   = NULL;
    if (m_pStream) m_pStream->Release();	    m_pStream	   = NULL;
    if (m_pPlayer) m_pPlayer->Release();	    m_pPlayer	   = NULL;
    if (m_pAudioPlayer) m_pAudioPlayer->Release();  m_pAudioPlayer = NULL;
    if (m_pAudioStream) m_pAudioStream->Release();  m_pAudioStream = NULL;
    if (m_pCommonClassFactory) m_pCommonClassFactory->Release(); m_pCommonClassFactory = NULL;

    RMAAudioData*  pAudioData = (RMAAudioData*) m_PacketQueue.Remove();
    while(pAudioData)
    {
	pAudioData->pData->Release();
	delete pAudioData;
        pAudioData = (RMAAudioData*) m_PacketQueue.Remove();
    }

    return PNR_OK;
}


// IRMADryNotification Interface Methods

/****************************************************************************
 *  IRMADryNotification::OnDryNotification                   ref:  rmaausvc.h
 *
 *  This function is called when it is time to write to audio device 
 *  and there is not enough data in the audio stream. The renderer can
 *  then decide to add more data to the audio stream. This should be 
 *  done synchronously within the call to this function.
 *  It is OK to not write any data. Silence will be played instead.
 */
STDMETHODIMP
CPCMRenderer::OnDryNotification
(
    UINT32 /*IN*/ ulCurrentStreamTime,
    UINT32 /*IN*/ ulMinimumDurationRequired
)
{
    /* Renderer will get dry notifications from Audio Services here */

    UINT32 ulNumMsWritten = 0;
    RMAAudioData*  pAudioData = (RMAAudioData*) m_PacketQueue.Remove();
    while(pAudioData)
    {
	m_pAudioStream->Write( pAudioData );
	ulNumMsWritten	+= CalcMs(pAudioData->pData->GetSize());

	pAudioData->pData->Release();
	delete pAudioData;

	/* Written enough data? */
	if (ulNumMsWritten >= ulMinimumDurationRequired)
	{
	    break;
	}

        pAudioData = (RMAAudioData*) m_PacketQueue.Remove();
    }

    return PNR_OK;
}


/****************************************************************************
 *  CPCMRenderer::CalcMs                                     ref:  pcmrendr.h
 *
 *  Calculate the duration in millisecs for this number of bytes.
 */
UINT32
CPCMRenderer::CalcMs
(
    UINT32	ulNumBytes
)
{
    return ( (UINT32) (( 1000.0 
		/ (m_audioFmt.uChannels * ((m_audioFmt.uBitsPerSample==8)?1:2) 
		*  m_audioFmt.ulSamplesPerSec) ) 
		*  ulNumBytes) );
}


// IUnknown COM Interface Methods

/****************************************************************************
 *  IUnknown::AddRef                                            ref:  pncom.h
 *
 *  This routine increases the object reference count in a thread safe
 *  manner. The reference count is used to manage the lifetime of an object.
 *  This method must be explicitly called by the user whenever a new
 *  reference to an object is used.
 */
STDMETHODIMP_(UINT32) CPCMRenderer::AddRef()
{
    return InterlockedIncrement(&m_lRefCount);
}


/****************************************************************************
 *  IUnknown::Release                                           ref:  pncom.h
 *
 *  This routine decreases the object reference count in a thread safe
 *  manner, and deletes the object if no more references to it exist. It must
 *  be called explicitly by the user whenever an object is no longer needed.
 */
STDMETHODIMP_(UINT32) CPCMRenderer::Release()
{
    if (InterlockedDecrement(&m_lRefCount) > 0)
    {
        return m_lRefCount;
    }

    delete this;
    return 0;
}


/****************************************************************************
 *  IUnknown::QueryInterface                                    ref:  pncom.h
 *
 *  This routine indicates which interfaces this object supports. If a given
 *  interface is supported, the object's reference count is incremented, and
 *  a reference to that interface is returned. Otherwise a NULL object and
 *  error code are returned. This method is called by other objects to
 *  discover the functionality of this object.
 */
STDMETHODIMP CPCMRenderer::QueryInterface(REFIID riid, void** ppvObj)
{
    if (IsEqualIID(riid, IID_IUnknown))
    {
	AddRef();
	*ppvObj = (IUnknown*)(IRMAPlugin*)this;
	return PNR_OK;
    }
    else if (IsEqualIID(riid, IID_IRMAPlugin))
    {
	AddRef();
	*ppvObj = (IRMAPlugin*)this;
	return PNR_OK;
    }
    else if (IsEqualIID(riid, IID_IRMARenderer))
    {
	AddRef();
	*ppvObj = (IRMARenderer*)this;
	return PNR_OK;
    }
    else if (IsEqualIID(riid, IID_IRMADryNotification))
    {
	AddRef();
	*ppvObj = (IRMADryNotification*)this;
	return PNR_OK;
    }

    *ppvObj = NULL;
    return PNR_NOINTERFACE;
}
