/****************************************************************************
 * 
 *  $Id: exauddev.cpp,v 1.2 1999/01/29 20:43:03 nhart Exp $
 *
 *  Copyright (C) 1995,1996,1997 RealNetworks, Inc.
 *  All rights reserved.
 *
 *  http://www.real.com/devzone
 *
 *  This program contains proprietary information of RealNetworks, Inc.,
 *  and is licensed subject to restrictions on use and distribution.
 * 
 *  exauddev.cpp
 *
 */

/****************************************************************************
 * Includes
 */
#include <ctype.h>

#include "pntypes.h"
#include "pncom.h"
#include "rmaausvc.h"
#include "rmapckts.h"
#include "rmaengin.h"
#include "rmacore.h"
#include "fivemque.h"
#include "os.h"
#include "exauddev.h"




/****************************************************************************
 *  FakeAudioServices::FakeAudioServices
 *
 *  The purpose of this class is to demonstrate how the IRMAAudioDevice
 *  might interact with the audio services of an OS.  This class merely
 *  discards any audio data written to it and prints a status to STDOUT.
 */

FakeAudioServices::FakeAudioServices()
    : m_ulAudioTime(0)
{
}


/****************************************************************************
 *  FakeAudioServices::Init
 *
 */

void
FakeAudioServices::Init()
{
    m_ulAudioTime = 0;
}


/****************************************************************************
 *  FakeAudioServices::Close
 *
 */

void
FakeAudioServices::Close()
{
    m_ulAudioTime = 0;
}


/****************************************************************************
 *  FakeAudioServices::AudioOut
 *
 *  Dump status to STDOUT and update our internal clock.  This clock is
 *  used to inform applications what the last timestamp of any audio it
 *  processed is.
 */

void
FakeAudioServices::AudioOut(void* pData, int nDataLen, int nMSecs)
{
    printf("audio out at %d ms\n", m_ulAudioTime);
    m_ulAudioTime = nMSecs;
}


/****************************************************************************
 *  FakeAudioServices::GetAudioClock
 *
 *  Return our audio service's clock-- a measurement of how many milliseconds
 *  the audio services have played so far.
 */

ULONG32
FakeAudioServices::GetAudioClock()
{
    // This value indicates the time of the last bit of audio played
    // by the physical audio device.
    return m_ulAudioTime;
}




/****************************************************************************
 *  ExampleAudioDevice::ExampleAudioDevice
 *
 */
ExampleAudioDevice::ExampleAudioDevice(IRMAPlayer* pPlayer)
    : m_lRefCount(0)
    , m_pResponse(NULL)
    , m_uVolume(0)
    , m_bPaused(FALSE)
    , m_bInitialized(FALSE)
    , m_pAudioServices(NULL)
    , m_pPlayer(pPlayer)
    , m_hCallback(NULL)
    , m_pQueue(NULL)
    , m_bFirstWrite(FALSE)
    , m_ulGranularity(100) // default to 100ms until we know the real value
{
    m_pAudioServices = new FakeAudioServices;
    m_pQueue = new FiveMinuteQueue;

    // We need the player so we can use its scheduler
    if (m_pPlayer)
    {
	m_pPlayer->AddRef();
    }
}


/****************************************************************************
 *  ExampleAudioDevice::~ExampleAudioDevice
 *
 */
ExampleAudioDevice::~ExampleAudioDevice()
{
    // In case someone didn't close us...
    if (m_pResponse)
    {
	m_pResponse->Release();
	m_pResponse = NULL;
    }

    if (m_pAudioServices)
    {
	delete m_pAudioServices;
	m_pAudioServices = NULL;
    }

    if (m_pPlayer)
    {
	m_pPlayer->Release();
	m_pPlayer = NULL;
    }
}


// IUnknown COM Interface Methods

/****************************************************************************
 *  IUnknown::AddRef                                            ref:  pncom.h
 *
 *  This routine increases the object reference count in a thread safe
 *  manner. The reference count is used to manage the lifetime of an object.
 *  This method must be explicitly called by the user whenever a new
 *  reference to an object is used.
 */
STDMETHODIMP_(UINT32)
ExampleAudioDevice::AddRef()
{
    return InterlockedIncrement(&m_lRefCount);
}


/****************************************************************************
 *  IUnknown::Release                                           ref:  pncom.h
 *
 *  This routine decreases the object reference count in a thread safe
 *  manner, and deletes the object if no more references to it exist. It must
 *  be called explicitly by the user whenever an object is no longer needed.
 */
STDMETHODIMP_(UINT32)
ExampleAudioDevice::Release()
{
    if (InterlockedDecrement(&m_lRefCount) > 0)
    {
        return m_lRefCount;
    }

    delete this;
    return 0;
}


/****************************************************************************
 *  IUnknown::QueryInterface                                    ref:  pncom.h
 *
 *  This routine indicates which interfaces this object supports. If a given
 *  interface is supported, the object's reference count is incremented, and
 *  a reference to that interface is returned. Otherwise a NULL object and
 *  error code are returned. This method is called by other objects to
 *  discover the functionality of this object.
 */
STDMETHODIMP
ExampleAudioDevice::QueryInterface(REFIID riid, void**ppvObj)
{
    if(IsEqualIID(riid, IID_IUnknown))
    {
        AddRef();
        *ppvObj = (IUnknown*)(IRMAAudioDevice*)this;
        return PNR_OK;
    }
    else if(IsEqualIID(riid, IID_IRMAAudioDevice))
    {
        AddRef();
        *ppvObj = (IRMAAudioDevice*)this;
        return PNR_OK;
    }
    else if(IsEqualIID(riid, IID_IRMACallback))
    {
        AddRef();
        *ppvObj = (IRMACallback*)this;
        return PNR_OK;
    }
    *ppvObj = NULL;
    return PNR_NOINTERFACE;
}


/****************************************************************************
 *  IRMAAudioDevice::Open
 *
 *  This routine initializes our audio device and prepares it for writing.
 *  We must store the IRMAAudioDeviceResponse to notify the player of
 *  time syncs.
 */

STDMETHODIMP 
ExampleAudioDevice::Open(const RMAAudioFormat*    /*IN*/ pFormat,
			IRMAAudioDeviceResponse* /*IN*/ pResponse)
{
    if (m_bInitialized || m_pResponse || !m_pPlayer)
    {
	return PNR_UNEXPECTED;
    }

    if (!pResponse || !pFormat)
    {
	return PNR_INVALID_PARAMETER;
    }

    if (!m_pAudioServices || !m_pQueue)
    {
	return PNR_OUTOFMEMORY;
    }

    m_pResponse = pResponse;
    m_pResponse->AddRef();

    memcpy(&m_format, pFormat, sizeof(RMAAudioFormat));

    // Initialize our OS's audio services
    m_pAudioServices->Init();

    m_bFirstWrite = TRUE;
    m_bInitialized = TRUE;

    return PNR_OK;
}


/****************************************************************************
 *  IRMAAudioDevice::Close
 *
 *  This routine frees up the resources in use by our audio device.
 */

STDMETHODIMP 
ExampleAudioDevice::Close(const BOOL  /*IN*/ bFlush)
{
    if (!m_bInitialized || !m_pAudioServices || !m_pPlayer)
    {
	return PNR_UNEXPECTED;
    }

    if (m_pResponse)
    {
	m_pResponse->Release();
	m_pResponse = NULL;
    }

    if (bFlush)
    {
	Drain();
    }

    // Close our OS's audio services
    m_pAudioServices->Close();

    m_bInitialized = FALSE;

    return PNR_OK;
}


/****************************************************************************
 *  IRMAAudioDevice::Resume
 *
 *  Put the audio device into an active state.  It must be resumed to play
 *  audio.  In our implementation we schedule a callback with the player
 *  engine because in this sample we want to buffer the audio data and push
 *  it to our faked OS audio services.
 *  
 */

STDMETHODIMP 
ExampleAudioDevice::Resume()
{
    IRMAScheduler* pScheduler = NULL;

    if (!m_bInitialized || !m_bPaused)
    {
	return PNR_UNEXPECTED;
    }

    // Use the scheduler to schedule a callback for processing our buffered
    // audio data.
    ScheduleCallback();
    
    if (m_hCallback)
    {
	m_bPaused = FALSE;
	return PNR_OK;
    }
    else
    {
	// something catastrophic happened and we weren't able to schedule a
	// callback with the player
	return PNR_FAIL;
    }
}


/****************************************************************************
 *  IRMAAudioDevice::Pause
 *
 *  Pause the audio device and put it into an inactive state.  This state
 *  should prevent any more audio from being played until Resume() is called.
 *  
 */

STDMETHODIMP 
ExampleAudioDevice::Pause()
{
    IRMAScheduler* pScheduler = NULL;

    if (!m_bInitialized || m_bPaused)
    {
	return PNR_UNEXPECTED;
    }

    RemoveCallback();
    
    m_bPaused = TRUE;
    return PNR_OK;
}


/****************************************************************************
 *  IRMAAudioDevice::Write
 *
 *  This function receives a packet of raw audio data and may do whatever it
 *  wishes with it.  You could immediately write it to your OS's audio device
 *  or dump it to a file or draw a visual representation of it.  In our sample
 *  we buffer the data in a queue and actually process it on a scheduled
 *  callback.  (On some multithreaded OSes one might wish to run a separate
 *  thread to process this data as it is buffered.)
 *  
 */

STDMETHODIMP 
ExampleAudioDevice::Write(const RMAAudioData* /*IN*/ pAudioData)
{
    if (!pAudioData)
    {
	return PNR_INVALID_PARAMETER;
    }

    if (!m_bInitialized)
    {
	return PNR_UNEXPECTED;
    }

    if (!m_pResponse)
    {
	return PNR_FAIL;
    }

    // Calculate the granularity to use for our scheduled callbacks
    if (m_bFirstWrite)
    {
	m_bFirstWrite = FALSE;

	ULONG32 ulBitsPerSec = m_format.uChannels * 
				m_format.uBitsPerSample * 
				m_format.ulSamplesPerSec;

	// Calculate the interval between blocks.  This is equal to
	// (1000 ms * 8 bits/byte * max bytes/block) / num bits/sec
	m_ulGranularity = (1000 * 8 * m_format.uMaxBlockSize) / ulBitsPerSec;
    }

    // Buffer the audio data and let our scheduler feed it to our OS's audio
    // services.
    RMAAudioData* pMyData = new RMAAudioData;
    if (pMyData)
    {
	// Make our own copy of the RMAAudioData struct
	memcpy(pMyData, pAudioData, sizeof(RMAAudioData));

	// keep a ref on the IRMABuffer member until we process it
	pMyData->pData->AddRef(); 

	// queue my audio data
	m_pQueue->Add((void*)pMyData);

	// Make sure we have a callback scheduled to process the data
	ScheduleCallback();
    }
    return PNR_OK;
}


/****************************************************************************
 *  IRMAAudioDevice::InitVolume
 *
 *  When this function is called the audio level of your audio device should
 *  be initialized.  In this sample we ignore it because we are just dumping
 *  audio timestamps to STDOUT, nothing more.
 *  
 */

STDMETHODIMP_(BOOL)
ExampleAudioDevice::InitVolume(const UINT16	/*IN*/ uMinVolume,
			       const UINT16	/*IN*/ uMaxVolume)
{
    return TRUE; // this audio device doesn't care about the volume
}


/****************************************************************************
 *  IRMAAudioDevice::SetVolume
 *
 *  This function should cause your audio device to change the volume level.
 *  
 */

STDMETHODIMP 
ExampleAudioDevice::SetVolume(const UINT16    /*IN*/ uVolume)
{
    m_uVolume = uVolume;
    return PNR_OK;
}


/****************************************************************************
 *  IRMAAudioDevice::GetVolume
 *
 *  This function should return the current volume level.
 *  
 */

STDMETHODIMP_(UINT16) 
ExampleAudioDevice::GetVolume()
{
    return m_uVolume;
}


/****************************************************************************
 *  IRMAAudioDevice::Reset
 *
 *  This function should reinitialize your audio device.
 *  
 */

STDMETHODIMP 
ExampleAudioDevice::Reset()
{
    m_bPaused = FALSE;

    if (m_pAudioServices)
    {
	m_pAudioServices->Close();
	m_pAudioServices->Init();
    }

    RemoveCallback();
    m_bFirstWrite = TRUE;

    return PNR_OK;
}


/****************************************************************************
 *  IRMAAudioDevice::Drain
 *
 *  This function should cause the audio device to discard any audio data it
 *  may have pending.  When this function is called it indicates that the
 *  audio data that has been written to the audio device is no longer needed.
 *  For instance, since the core generally buffers a second or two of audio
 *  to the audio device if a user stops a presentation mid-stream there will
 *  be one or two seconds of audio data waiting in the audio device's buffers
 *  that should not be played.  (Otherwise the user would hear one or two
 *  seconds of audio after he/she stopped playback.)
 *  
 */

STDMETHODIMP 
ExampleAudioDevice::Drain()
{
    // Flush pending buffers and free the memory
    RMAAudioData* pMyData = (RMAAudioData*)m_pQueue->Remove();
    while (pMyData)
    {
	pMyData->pData->Release();
	delete pMyData;

	pMyData = (RMAAudioData*)m_pQueue->Remove();
    }

    RemoveCallback();

    return PNR_OK;
}


/****************************************************************************
 *  IRMAAudioDevice::CheckFormat
 *
 *  This function is used to determine if the audio device supports a
 *  requested audio format.  Our implementation accepts any format because
 *  we don't actually use the audio data-- we just look at the timestamps.
 *  
 */

STDMETHODIMP 
ExampleAudioDevice::CheckFormat(const RMAAudioFormat* /*IN*/ pFormat)
{
    return PNR_OK; // we accept any format
}


/****************************************************************************
 *  IRMAAudioDevice::GetCurrentAudioTime
 *
 *  This function is critical to optimal playback.  It *MUST* return the
 *  most accurate value as possible, otherwise the audio could get out of
 *  synch with the video, buffering problems could occur or the entire
 *  presentation could slowly grind to a slow crawl.  Whatever value is
 *  returned here is what the player engine uses to determine what the
 *  user is currently hearing.
 *
 *  We do not want to simply return the timestamp of the last audio packet 
 *  passed to IRMAAudioDevice::Write().  That is merely the time of the
 *  last packet buffered.  What we really want to do is query the OS's
 *  audio services and find out how far along in the audio presentation it is.
 *
 *  Our sample uses a faked audio services that acts like a simplistic
 *  audio API of an imaginary OS.  We ask it what time it is according to
 *  its clock and return that value.
 */

STDMETHODIMP 
ExampleAudioDevice::GetCurrentAudioTime(REF(ULONG32) /*OUT*/ ulCurrentTime)
{
    if (m_bInitialized && m_pAudioServices)
    {
	ulCurrentTime = m_pAudioServices->GetAudioClock();
    }
    else
    {
	ulCurrentTime = 0;
    }

    return PNR_OK;
}


/****************************************************************************
 *  IRMACallback::Func
 *
 *  This is the callback called by the scheduler.  Our sample uses it to
 *  process the buffered audio data and push it downstream to our fake
 *  audio services.
 */

STDMETHODIMP 
ExampleAudioDevice::Func()
{
    // We just got our callback, so reinitialize our handle to null
    m_hCallback = NULL;

    RMAAudioData* pMyData = (RMAAudioData*)m_pQueue->Remove();
    if (pMyData)
    {
	IRMABuffer* pBuf = pMyData->pData;

	// write this audio data to our OS's audio services
	m_pAudioServices->AudioOut(pBuf->GetBuffer(), 
				    pBuf->GetSize(), 
				    pMyData->ulAudioTime);

	// Let the player engine know we just processed some audio and
	// give it the audio service's current clock time
	if (m_pResponse)
	{
	    m_pResponse->OnTimeSync(m_pAudioServices->GetAudioClock());
	}

	// release our ref on the IRMABuffer and free our copy of the 
	// RMAAudioData struct
	pBuf->Release();
	delete pMyData;

	// Schedule another callback to handle more audio data
	ScheduleCallback();
    }

    return PNR_OK;
}


/****************************************************************************
 *  ScheduleCallback
 *
 *  This function is used internally to setup a callback with the player
 *  engine.
 */

void
ExampleAudioDevice::ScheduleCallback()
{
    if (m_hCallback)
    {
	// we already have a callback scheduled
	return;
    }

    IRMAScheduler* pScheduler = NULL;

    // Use the scheduler to schedule a callback for processing our buffered
    // audio data.
    if (m_pPlayer->QueryInterface(IID_IRMAScheduler, (void**)&pScheduler) ==
	PNR_OK)
    {
	m_hCallback = pScheduler->RelativeEnter((IRMACallback*)this, 
						m_ulGranularity);

	pScheduler->Release();
	pScheduler = NULL;
    }
}


/****************************************************************************
 *  RemoveCallback
 *
 *  This function is used internally to remove a callback with the player
 *  engine.
 */

void
ExampleAudioDevice::RemoveCallback()
{
    if (!m_hCallback)
    {
	// we don't have a callback scheduled
	return;
    }

    IRMAScheduler* pScheduler = NULL;

    // Use the scheduler to remove a scheduled callback and stop processing
    // our buffered audio data.
    if (m_pPlayer->QueryInterface(IID_IRMAScheduler, (void**)&pScheduler) ==
	PNR_OK)
    {
	pScheduler->Remove(m_hCallback);

	m_hCallback = NULL;

	pScheduler->Release();
	pScheduler = NULL;
    }
}
