/****************************************************************************
 * 
 * $Id: exencaud.cpp,v 1.2 1999/10/05 19:17:10 ehodge Exp $
 *
 *  Copyright (C) 1999 RealNetworks, Inc.
 *  All rights reserved.
 *
 *  http://www.real.com/devzone
 *
 *  This program contains proprietary information of RealNetworks, Inc.,
 *  and is licensed subject to restrictions on use and distribution.
 * 
 *  exencaud.cpp
 *
 *  ExAudio encoder class for the SuperEncoder library.
 *
 */


/****************************************************************************
 * Defines
 */
#define INITGUID


/****************************************************************************
 * Includes
 */
#include <stdio.h>

#if defined (_WINDOWS)
#include <windows.h>
#elif defined (_UNIX)
#include "unix/os.h"
#elif defined (_MACINTOSH)
#include "os.h"
#endif

#include "pntypes.h"
#include "pncom.h"
#include "rmapckts.h"
#include "rmacomm.h"
#include "rmafiles.h"
#include "rmaformt.h"
#include "rmaengin.h"
#include "rmaencod.h"
#include "rmaerror.h"

#include "exencaud.h"

typedef IUnknown* (PNEXPORT_PTR FPCREATE_CONTEXT) ();
typedef PN_RESULT (PNEXPORT_PTR FPRMSETDLLACCESSPATH) (const char*);

/* These defines are for RTP payload type LIN-16 1 channel.
 * See http://194.52.182.96/rfc/rfc1890.html, section 7, for other
 * RTP Payload types' values. */
#define RTP_PAYLOAD_TYPE    11 /* PCMU */
#define SAMPLES_PER_SEC	    44100 /* samples/sec for the above RTP payload. */
#define AVG_PKT_SIZE	    441 /* use 320 for all RTP payload types except:
				 *  GSM: use 455, 
				 *  DVI4_8: use 160
				 */
#define PKT_DUR	    (1000.0/((double)SAMPLES_PER_SEC/(double)AVG_PKT_SIZE))

/****************************************************************************
 *  ExAudioEncoder::ExAudioEncoder                           ref:  exencaud.h
 *
 *  Constructor. Address and port are passed in from main().
 */
ExAudioEncoder::ExAudioEncoder(char* address, UINT16 port,
			       char* username, char* password,
			       ULONG32 ulRTPPayloadType,
			       ULONG32 ulNumPacketsToSend) :
    m_lRefCount(0),
    m_pContext(NULL),
    m_pRequest(NULL),
    m_pEncoderResponse(NULL),
    m_pEncoderResponseCompletion(NULL),
    m_pScheduler(NULL),
    m_EncoderDone(FALSE),
    m_EncoderInitialized(FALSE),
    m_bPacketsStarted(FALSE),
    m_pAddress(address),
    m_port(port),
    m_pUsername(username),
    m_pPassword(password),
    m_ulRTPPayloadType(ulRTPPayloadType),
    m_ulPacketsToSend(ulNumPacketsToSend),
    m_ulPriorPacketTime(0),
    m_ulPacketsSent(0)
{
}


/****************************************************************************
 *  ExAudioEncoder::~ExAudioEncoder                          ref:  exencaud.h
 *
 *  Destructor.
 */
ExAudioEncoder::~ExAudioEncoder()
{
    Cleanup();
}


/****************************************************************************
 *  ExAudioEncoder::Encode                                   ref:  exencaud.h
 *
 *  The main loop for this encoder.  main creates an ExAudioEncoder, then
 *  calls this.
 */
void
ExAudioEncoder::Encode(const char* pFilename)
{
    BOOL done = FALSE;
    UINT32 lasttime = 0;

    /* First let's get ourselves a context, which will give us access to
     * IRMACommonClassFactory in particular, and several other interfaces
     * this example does not use.
     */

#if defined (_WINDOWS) || (_WIN32)
    HINSTANCE hDll = LoadLibrary("encn3260.dll");
#elif defined (_UNIX)
    HINSTANCE hDll = LoadLibrary("encnet.so.6.0");
#elif defined (_MACINTOSH)
    HINSTANCE hDll = LoadLibrary("encnet60.dll");
#else
#error "No LoadLibrary for this platform"
#endif

    if (!hDll)
    {
	fprintf(stdout, "Unable to LoadLibrary\n");
	return;
    }
	
    IUnknown* (PNEXPORT_PTR fpCreateContext) ();
    FPRMSETDLLACCESSPATH fpSetDLLAccessPath;
 
    fpCreateContext =
	(FPCREATE_CONTEXT)GetProcAddress(hDll, "CreateContext");
    fpSetDLLAccessPath = 
	(FPRMSETDLLACCESSPATH)GetProcAddress(hDll, "SetDLLAccessPath");

    if (NULL != fpSetDLLAccessPath)
    {
	/*Create a null delimited, double-null terminated string containing
	 * the paths to the encnet library (DT_Common) and the sdpplin library
	 * (DT_Plugins)... */
	char pPaths[256];
	UINT32 ulPathLen = 0;
	char pNextPath[256];
#if defined(_MACINTOSH)
	const char* pCurrentDir = ".:";
#elif defined(_UNIX)
	const char* pCurrentDir = "./";
#else
	const char* pCurrentDir = ".\\";
#endif

	memset(pPaths, 0, 256);

	sprintf(pNextPath, "DT_Common=%s", pCurrentDir);
	memcpy(pPaths, pNextPath, strlen(pNextPath) + 1);
	ulPathLen = strlen(pNextPath) + 1;

	sprintf(pNextPath, "DT_Plugins=%s", pCurrentDir);
	memcpy(pPaths + ulPathLen, pNextPath, strlen(pNextPath) + 1);
	ulPathLen = strlen(pNextPath) + 1;

	fpSetDLLAccessPath((char*)pPaths);
    }

    if (!fpCreateContext)
    {
	fprintf(stdout, "Unable to GetProcAddress\n");
	return;
    }
	
    m_pContext = fpCreateContext();

    /* Now that we got a context object, get a CommonClassFactory object
     */
    if(PNR_OK != m_pContext->QueryInterface(IID_IRMACommonClassFactory,
					  (void**)&m_pCommonClassFactory))
    {
	fprintf(stdout, "No common class factory\n");
	return;
    }

    /* Also, with the context object we can now get a Scheduler object
     * which we will use to schedule callbacks for sending packets:
     */
    if(PNR_OK != m_pContext->QueryInterface(IID_IRMAScheduler,
					  (void**)&m_pScheduler))
    {
	fprintf(stdout, "No scheduler available\n");
	return;
    }

    /*
     * Create and initialize the IRMARequest object
     */
    if(PNR_OK != SetupRequest(pFilename))
    {
	return;
    }

    /* Now we need an EncoderResponse object.  Once we have it, we will
     * use it for all further communication with the library.
     */
    if(PNR_OK != m_pContext->QueryInterface(IID_IRMAEncoderResponse,
	(void**)&m_pEncoderResponse))
    {
	fprintf(stdout, "No encoder response\n");
	return;
    }

    if(PNR_OK != m_pContext->QueryInterface(IID_IRMAEncoderResponseCompletion,
	(void**)&m_pEncoderResponseCompletion))
    {
	fprintf(stdout, "No encoder response completion\n");
	return;
    }

    /* We've created the EncoderResponse object, now give it the info
     * it needs to operate: the server and port to connect to, the filename
     * we are encoding, and a pointer back to this object
     */
    m_pEncoderResponse->InitEncoderResponse(m_pAddress,
					    m_port,
					    m_pRequest,
					    m_pUsername,
					    m_pPassword,
					    (IRMAEncoder*)this);

    fprintf(stdout, "Entering mainloop\n");

    BOOL bIsFirstTimeHere = TRUE;

    /* The encoder will remain in this loop until the EncoderResponse
     * object tells us to stop
     */
    while(!m_EncoderDone)
    {
	/* If we haven't gotten the InitEncoderResponseDone callback yet,
	 * then don't do anything but Process().
	 */
	if(m_EncoderInitialized)
	{
	    /* Jump start packet sending by calling SendAPacket() which
	     * schedules a callback for the next packet time, and then
	     * SendPacketCallback::Func() in turn calls SendAPacket(),
	     * and so on... */
	    if(m_bPacketsStarted  &&  bIsFirstTimeHere)
	    {
		bIsFirstTimeHere = FALSE;
		SendAPacket();
	    }
	}
	/* Every user of the superencoder library must call 
	 * IRMAEncoderResponse::Process() regularly.
	 */
	if(PNR_OK != m_pEncoderResponse->Process())
	{
	    m_EncoderDone = TRUE;
	}
    }
}



/****************************************************************************
 *  ExAudioEncoder::SendAPacket                              ref:  exencaud.h
 *
 *  Send a packet!
 */
void
ExAudioEncoder::SendAPacket()
{
    IRMAPacket* pPacket;
    IRMABuffer* pData;
    RMATimeval  now;
    SendPacketCallback* pCallback;

    /*
     * If the stream has been stopped, just exit
     */
    if (!m_bPacketsStarted  ||  m_EncoderDone)
    {
	return;
    }

    /* First create an IRMAPacket to contain our data */
    if(m_pCommonClassFactory  &&
	PNR_OK != m_pCommonClassFactory->CreateInstance(IID_IRMAPacket,
						     (void**)&pPacket))
    {
	return;
    }

    /* And a buffer for the data itself */
    if(PNR_OK != m_pCommonClassFactory->CreateInstance(IID_IRMABuffer,
						     (void**)&pData))
    {
	return;
    }

    now = m_pScheduler->GetCurrentSchedulerTime();

    pData->SetSize(AVG_PKT_SIZE);

    UINT32 timenow;
    timenow = m_pEncoderResponse->GetTime();

    /*We need to make sure we send each packet with a packet time that's
     * equal to the start time of the prior packet plus the duration of
     * the prior packet:
     */
    ULONG32 ulPacketTime = timenow;
    ULONG32 ulPacketDuration = ULONG32(1000.0 /
	    ((double)g_pPayloadSamplesPerSec[m_ulRTPPayloadType] /
	    (double)g_pPayloadPacketSize[m_ulRTPPayloadType] ));

    if(m_ulPacketsSent > 0L) /* Only do this if not the first packet */
    {
	ulPacketTime = m_ulPriorPacketTime + ulPacketDuration;
    }
    m_ulPriorPacketTime = ulPacketTime;

    /* Fill in the raw data with a some basic waveform audio:
     */
    char* pFoo = (char*)pData->GetBuffer();
    if((ulPacketTime/1000)&1) /*Odd seconds are blank, even seconds beep:*/
    {
	memset(pFoo, 0, AVG_PKT_SIZE);
    }
    else
    {
	const UINT16 sinwav[16] = {
	      0,   8,  32,  64, 128, 192, 224, 248,
	    255, 248, 224, 192, 128,  64,  32,   8 };
	/* Make a crude sinusoidal wave: */
	for(UINT16 uI=0; uI<AVG_PKT_SIZE; uI++)
	{
	    pFoo[uI] = char(sinwav[uI%16]);
	}
    }

    /* Set all the info in the IRMAPacket */
    pPacket->Set(pData,
		 ulPacketTime,
		 0,
		 RMA_ASM_SWITCH_ON,
		 0);

    fprintf(stdout, "Packet: %lu sent at %lu [timenow=%lu]\n",
	    m_ulPacketsSent, ulPacketTime, timenow);

    /* Give the packet to the EncoderResponse */
    m_pEncoderResponse->PacketReady(PNR_OK, pPacket);

    /* Now we're done with these objects, release them to free the
     * memory
     */
    pPacket->Release();
    pData->Release();

    /*
     * If we have sent as many packets as the user wanted, then
     * tell the encoder response that we are done.  We do this
     * by calling StreamDone on all streams.
     */
    if(m_ulPacketsToSend)
    {
	m_ulPacketsSent++;
	if(m_ulPacketsSent == m_ulPacketsToSend)
	{
	    m_pEncoderResponse->StreamDone(0);
	    /*
	     * Don't send out any more packets, just call process.
	     */
	    m_EncoderInitialized = FALSE;
	}
    }

    /* Calculate the time for the next packet */
    ULONG32 ulPacketDurationInMicrosec = ULONG32(1000000.0 /
	    ((double)g_pPayloadSamplesPerSec[m_ulRTPPayloadType] /
	    (double)g_pPayloadPacketSize[m_ulRTPPayloadType] ));
    m_nextPacketTime.tv_usec += ulPacketDurationInMicrosec;
    m_nextPacketTime.tv_sec += m_nextPacketTime.tv_usec / 1000000;
    m_nextPacketTime.tv_usec = m_nextPacketTime.tv_usec % 1000000;

    /* And create a callback object for the scheduler to call at the
     * time we want to send the next packet
     */
    pCallback = new SendPacketCallback(this);
    pCallback->AddRef();
    m_pScheduler->AbsoluteEnter(pCallback, m_nextPacketTime);
    pCallback->Release();
}


/****************************************************************************
 *  ExAudioEncoder::SetupRequest                             ref:  exencaud.h
 *
 *  Sets up the request information for the IRMARequest object
 */
STDMETHODIMP
ExAudioEncoder::SetupRequest(const char* pFilename)
{
    PN_RESULT result = PNR_OK;

    if(PNR_OK != m_pCommonClassFactory->CreateInstance(CLSID_IRMARequest,
                                                       (void**)&m_pRequest))
    {
	fprintf(stdout, "Unable to create request\n");
	return PNR_FAIL;
    }

    if (PNR_OK != m_pRequest->SetURL(pFilename))
    {
	fprintf(stdout, "Unable to set URL\n");
	return PNR_FAIL;
    }

    IRMABuffer* pBuffer  = NULL;
    IRMAValues* pHeaders = NULL;

    result = m_pCommonClassFactory->CreateInstance(CLSID_IRMABuffer,
                                                   (void**)&pBuffer);

    if (PNR_OK != result)
    {
	fprintf(stdout, "Unable to create Buffer\n");
	goto setupDone;
    }

    /*
     * Must include the '\0' in the buffer string
     */

    result = pBuffer->Set((BYTE*)"buzz; buzz=zzub", 16);

    if(PNR_OK != result)
    {
	fprintf(stdout, "Unable to set Buffer\n");
	goto setupDone;
    }

    result = m_pCommonClassFactory->CreateInstance(CLSID_IRMAValues,
                                                   (void**)&pHeaders);

    if(PNR_OK != result)
    {
	fprintf(stdout, "Unable to create Header\n");
	goto setupDone;
    }

    /*
     * Setting RFC822 header...foo:buzz; buzz=zzub
     */

    result = pHeaders->SetPropertyCString("foo", pBuffer);

    if(PNR_OK != result)
    {
	fprintf(stdout, "Unable to set Property\n");
	goto setupDone;
    }

    if(m_pRequest->SetRequestHeaders(pHeaders))
    {
	fprintf(stdout, "Unable to set Headers\n");
    }

setupDone:

    if (pBuffer)
    {
	pBuffer->Release();
    }

    if (pHeaders)
    {
	pHeaders->Release();
    }

    return result;
}


/****************************************************************************
 *  ExAudioEncoder::SetupRequest                             ref:  exencaud.h
 *
 *  Handle the response information contained in the IRMARequest object
 */
STDMETHODIMP
ExAudioEncoder::HandleResponse()
{
    IRMAValues* pHeaders;

    if (PNR_OK == m_pRequest->GetResponseHeaders(pHeaders))
    {
	const char* name;
	IRMABuffer* value;
	PN_RESULT result;

	fprintf(stdout, "\n\nHEADERS...\n");

	result = pHeaders->GetFirstPropertyCString(name, value);

	while (PNR_OK == result)
	{
	    fprintf(stdout, "\t%s: %s\n", name, (char*)value->GetBuffer());
	    value->Release();
	    result = pHeaders->GetNextPropertyCString(name, value);
	}

	fprintf(stdout, "\n");

	pHeaders->Release();
    }

    return PNR_OK;
}

/* IRMAEncoderCompletion methods: */

/****************************************************************************
 *  IRMAEncoderCompletion::EncoderDone			    ref: rmaencod.h
 *
 *  Called by the encoder response when it gets a response from the server
 *  that the connection has been shut down.
 */

STDMETHODIMP
ExAudioEncoder::EncoderDone(PN_RESULT result)
{
    m_EncoderDone = TRUE;
    fprintf(stdout, "Got the EncoderDone.\n");
    Cleanup();
    return PNR_OK;
}


/* IRMAEncoder Interface Methods: */

/****************************************************************************
 *  IRMAEncoder::InitEncoderResponseDone                     ref:  rmaencod.h
 *
 *  InitEncoderResponseDone is called when InitEncoderResponse has
 *  completed.  It may return PNR_OK or an error result.  An error
 *  should be considered critical, and the encoder should exit (or
 *  prompt for new parameters and start over or whatever)
 */
STDMETHODIMP
ExAudioEncoder::InitEncoderResponseDone(PN_RESULT status)
{
    fprintf(stdout, "InitEncoderResponseDone(%ld)\n", status);
    if(PNR_OK == status)
    {
	m_EncoderInitialized = TRUE;
    }
    else
    {
	/* Failed, might as well exit.
	 */
	m_EncoderDone = TRUE;
    }

    return PNR_OK;
}


/****************************************************************************
 *  IRMAEncoder::GetFileHeader                               ref:  rmaencod.h
 *
 *  GetFileHeader is called to retrieve the header for the entire
 *  presentation.  Most importantly, it should contain the number
 *  of streams present.  This is very similar to the GetFileHeader
 *  method of IRMAFileFormat.
 */
STDMETHODIMP
ExAudioEncoder::GetFileHeader()
{
    IRMAValues* pHeader;

    /* Create an IRMAValues object to hold the header */
    if(PNR_OK != m_pCommonClassFactory->CreateInstance(CLSID_IRMAValues,
						     (void**)&pHeader))
    {
	return PNR_FAIL;
    }

    /* The only property in a file header is the number of streams the
     * encoder is encoding.  In this case, always 1.
     */
    pHeader->SetPropertyULONG32("StreamCount", 1);

    /* Give the header to the IRMAEncoderResponse */
    m_pEncoderResponse->FileHeaderReady(PNR_OK, pHeader);

    /* And clean up our reference on the header */
    pHeader->Release();
    return PNR_OK;
}


/****************************************************************************
 *  IRMAEncoder::GetStreamHeader                             ref:  rmaencod.h
 *
 *  GetStreamHeader is called to retrieve a header for each stream
 *  in the presentation.  It is similar to GetStreamHeader in
 *  IRMAFileFormat.
 */
STDMETHODIMP
ExAudioEncoder::GetStreamHeader(UINT16 stream)
{
    IRMAValues* pHeader;
    IRMABuffer* pOpaqueData;
    IRMABuffer* pStreamName;
    IRMABuffer* pMimeType;

    /* Create an IRMAValues object to hold the header */
    if(PNR_OK != m_pCommonClassFactory->CreateInstance(CLSID_IRMAValues,
						     (void**)&pHeader))
    {
	return PNR_FAIL;
    }

    /* A buffer for the opaque (type specific) data */
    if(PNR_OK != m_pCommonClassFactory->CreateInstance(CLSID_IRMABuffer,
						     (void**)&pOpaqueData))
    {
	return PNR_FAIL;
    }

    if(PNR_OK != m_pCommonClassFactory->CreateInstance(CLSID_IRMABuffer,
						       (void**)&pStreamName))
	return PNR_FAIL;

    if(PNR_OK != m_pCommonClassFactory->CreateInstance(CLSID_IRMABuffer,
						       (void**)&pMimeType))
	return PNR_FAIL;

    /* All these properties should be filled in for stream headers. */
    pHeader->SetPropertyULONG32("StreamNumber",  0);
    pHeader->SetPropertyULONG32("Duration", 0xffffffff); /* "infinity" */
    pHeader->SetPropertyULONG32("RTPPayloadType", m_ulRTPPayloadType);

    /*
     * Now set the ASM rule book
     */


    /* Give the header to the EncoderResponse */
    m_pEncoderResponse->StreamHeaderReady(PNR_OK, pHeader);

    /* And clean up the objects we created */
    pOpaqueData->Release();
    pHeader->Release();
    pStreamName->Release();
    pMimeType->Release();

    return PNR_OK;
}


/****************************************************************************
 *  IRMAEncoder::StartPackets                                ref:  rmaencod.h
 *
 *  StartPackets is called to inform the encoder that it should start
 *  sending packets for a particular stream.
 */
STDMETHODIMP
ExAudioEncoder::StartPackets(UINT16 stream)
{
    SendPacketCallback* cb;
    /*
     * The response headers should be available by now
     */
    HandleResponse();

    m_bPacketsStarted = TRUE;

    m_startTime = m_pScheduler->GetCurrentSchedulerTime();
    m_nextPacketTime = m_startTime;

    /* Create a callback object for the first packet: */
    cb = new SendPacketCallback(this);
    cb->AddRef();
    /* Schedule the first packet callback for right now. */
    m_pScheduler->AbsoluteEnter(cb, m_startTime);
    cb->Release();

    return PNR_OK;
}


/****************************************************************************
 *  IRMAEncoder::StopPackets                                 ref:  rmaencod.h
 *
 *  StopPackets is called when the encoder should stop sending packets
 *  for a stream.  This will not be called often, as it is generally
 *  up to the encoder application to determine when a stream is done.
 */
STDMETHODIMP
ExAudioEncoder::StopPackets(UINT16 stream)
{
    m_bPacketsStarted = FALSE;

    return PNR_OK;
}


/* IRMAErrorSink Interface Methods: */

/****************************************************************************
 *  IRMAEncoder::ErrorOccurred                               ref:  rmaerror.h
 *
 *  ErrorOccurred is called when an error occurs!
 */
STDMETHODIMP
ExAudioEncoder::ErrorOccurred(const UINT8  unSeverity,
			      const UINT32 ulRMACode,
			      const UINT32 ulUserCode,
			      const char*  pUserString,
			      const char*  pMoreInfoURL)
{
    fprintf(stdout, "*** An error occurred:\n*** %s\n*** See %s for details\n",
	    pUserString, pMoreInfoURL);
    Cleanup();
    m_EncoderDone = TRUE;
    
    return PNR_OK;
}


/* IUnknown COM Interface Methods: */
 
/****************************************************************************
 *  IUnknown::QueryInterface                                    ref:  pncom.h
 *
 *  This routine indicates which interfaces this object supports. If a given
 *  interface is supported, the object's reference count is incremented, and
 *  a reference to that interface is returned. Otherwise a NULL object and
 *  error code are returned. This method is called by other objects to
 *  discover the functionality of this object.
 */
STDMETHODIMP
ExAudioEncoder::QueryInterface(REFIID riid, void** ppvObj)
{
    if (IsEqualIID(riid, IID_IUnknown))
    {
	AddRef();
	*ppvObj = (IUnknown*)(IRMAEncoder*)this;
	return PNR_OK;
    }
    else if(IsEqualIID(riid, IID_IRMAEncoder))
    {
	AddRef();
	*ppvObj = (IRMAEncoder*)this;
	return PNR_OK;
    }
    else if(IsEqualIID(riid, IID_IRMAErrorSink))
    {
	AddRef();
	*ppvObj = (IRMAErrorSink*)this;
	return PNR_OK;
    }
    else if(IsEqualIID(riid, IID_IRMAEncoderCompletion))
    {
	AddRef();
	*ppvObj = (IRMAEncoderCompletion*)this;
	return PNR_OK;
    }
    *ppvObj = NULL;
    return PNR_NOINTERFACE;
}

/****************************************************************************
 *  IUnknown::AddRef                                            ref:  pncom.h
 *
 *  This routine increases the object reference count in a thread safe
 *  manner. The reference count is used to manage the lifetime of an object.
 *  This method must be explicitly called by the user whenever a new
 *  reference to an object is used.
 */
STDMETHODIMP_(ULONG32)
ExAudioEncoder::AddRef()
{
    return InterlockedIncrement(&m_lRefCount);
}

/****************************************************************************
 *  IUnknown::Release                                           ref:  pncom.h
 *
 *  This routine decreases the object reference count in a thread safe
 *  manner, and deletes the object if no more references to it exist. It must
 *  be called explicitly by the user whenever an object is no longer needed.
 */
STDMETHODIMP_(ULONG32)
ExAudioEncoder::Release()
{
    if (InterlockedDecrement(&m_lRefCount) > 0)
    {
	return m_lRefCount;
    }

    delete this;
    return 0;
}

void
ExAudioEncoder::Cleanup()
{
    if (m_pCommonClassFactory)
    {
	m_pCommonClassFactory->Release();
	m_pCommonClassFactory = 0;
    }

    if (m_pScheduler)
    {
	m_pScheduler->Release();
	m_pScheduler = 0;
    }

    if (m_pContext)
    {
	m_pContext->Release();
	m_pContext = 0;
    }

    if (m_pEncoderResponse)
    {
	/*
	 * Must allow EncoderResponse to free up protocol before calling
	 * Release()
	 */

	if (m_pEncoderResponseCompletion)
	{
	    m_pEncoderResponseCompletion->EncoderResponseDone();
	    m_pEncoderResponseCompletion->Release();
	    m_pEncoderResponseCompletion = 0;
	}

	m_pEncoderResponse->Release();
	m_pEncoderResponse = 0;
    }

    if (m_pRequest)
    {
	m_pRequest->Release();
	m_pRequest = 0;
    }
}



/* SendPacketCallback Class: */

/****************************************************************************
 *  SendPacketCallback::SendPacketCallback                   ref:  exencaud.h
 *
 */
SendPacketCallback::SendPacketCallback(ExAudioEncoder* obj)
{
    m_lRefCount = 0;
    m_pObj = obj;
    m_pObj->AddRef();
}


/****************************************************************************
 *  SendPacketCallback::~SendPacketCallback                  ref:  exencaud.h
 *
 */
SendPacketCallback::~SendPacketCallback()
{
    m_pObj->Release();
}

/****************************************************************************
 *  IRMACallback::Func                                       ref:  rmaengin.h
 *
 *  Called by the scheduler object when it's time to send a new packet.
 */
STDMETHODIMP
SendPacketCallback::Func()
{
    /* The real work is done by the ExAudioEncoder that created
     * this callback.
     */
    m_pObj->SendAPacket();
    return PNR_OK;
}


/****************************************************************************
 *  IUnknown::AddRef                                            ref:  pncom.h
 *
 */
STDMETHODIMP_(ULONG32) SendPacketCallback::AddRef()
{
    return InterlockedIncrement(&m_lRefCount);
}


/****************************************************************************
 *  IUnknown::Release                                           ref:  pncom.h
 *
 */
STDMETHODIMP_(ULONG32) SendPacketCallback::Release()
{
    if (InterlockedDecrement(&m_lRefCount) > 0)
    {
        return m_lRefCount;
    }

    delete this;
    return 0;
}


/****************************************************************************
 *  IUnknown::QueryInterface                                    ref:  pncom.h
 *
 */
STDMETHODIMP SendPacketCallback::QueryInterface(
    REFIID riid, void** ppvObj)
{
    if (IsEqualIID(riid, IID_IUnknown))
    {
	AddRef();
	*ppvObj = (IUnknown*)(IRMABroadcastFormatObject*)this;
	return PNR_OK;
    } 
    else if(IsEqualIID(riid, IID_IRMACallback))
    {
	AddRef();
	*ppvObj = (IRMABroadcastFormatObject*)this;
	return PNR_OK;
    }

    *ppvObj = NULL;
    return PNR_NOINTERFACE;
}
