/****************************************************************************
 * 
 *  $Id: exlogpln.cpp,v 1.12 1997/10/17 20:52:50 christof Exp $
 *
 *  Copyright (C) 1995,1996,1997 RealNetworks, Inc.
 *  All rights reserved.
 *  
 *  http://www.real.com/devzone
 *
 *  This program contains proprietary information of RealNetworks, Inc.,
 *  and is licensed subject to restrictions on use and distribution.
 *
 *  exlogpln.cpp
 *
 *  Example Logger plug-in for RealMedia Architecture (RMA).
 *
 */

/****************************************************************************
 * 
 *  EXAMPLE LOGGER PLUGIN for RMA Server
 *
 *  This is a very simple logger plugin, that it receives the log information
 *  (i.e. Warnings, Errors) from the server and saves them into a file.
 */


/****************************************************************************
 * Includes
 */
#include <stdio.h>

#define INITGUID    1
#include "pntypes.h"
#include "pncom.h"
#include "rmacomm.h"
#include "rmaplugn.h"
#include "rmaerror.h"
#include "exlogpln.ver"
#undef INITGUID
#include "exlogpln.h"


/****************************************************************************
 * Constants
 */
#define MAX_DISPLAY_NAME   256
#define LOGFILE	          "exlog.log"    /* sample log file name */


/****************************************************************************
 *  RMACreateInstance                                        ref:  rmaplugn.h
 *
 *  This routine creates a new instance of the CExampleRenderer class.
 *  It is called when the RMA core application is launched, and whenever
 *  a stream associated with this plug-in needs to be rendered.
 */
STDAPI
RMACreateInstance(IUnknown** ppLogger)
{
    *ppLogger = (IUnknown*)(IRMAPlugin*)new Logger();
    if (*ppLogger)
    {
	(*ppLogger)->AddRef();
	return PNR_OK;
    }

    return PNR_OUTOFMEMORY;
}


/****************************************************************************
 *  Logger::Logger                                           ref:  exlogpln.h
 *
 * Constructor
 */
Logger::Logger()
{ 
    m_lRefCount = 0;
    m_hLogFile = NULL;
}


/****************************************************************************
 *  IRMAPlugin::GetPluginInfo                                ref:  rmaplugn.h
 *
 *  This routine returns descriptive information about the plug-in, most
 *  of which is used in the About box of the user interface. It is called
 *  when the RMA core application is launched.
 */
STDMETHODIMP 
Logger::GetPluginInfo
    (
     REF(BOOL)       /*OUT*/ bLoadMultiple,
     REF(const char*)/*OUT*/ pDescription,
     REF(const char*)/*OUT*/ pCopyright,
     REF(const char*)/*OUT*/ pMoreInfoURL,
     REF(UINT32)     /*OUT*/ ulVersionNumber
    )
{   
    bLoadMultiple = FALSE;   // Must be true for file formats.

    pDescription    = "RealNetworks Basic Log Module";
    pCopyright	    = "(c) 1995,1996,1997 RealNetworks, All rights reserved.";
    pMoreInfoURL    = "http://www.real.com";
    ulVersionNumber = TARVER_ULONG32_VERSION;

    return PNR_OK;
}


/****************************************************************************
 *  IRMAPlugin::InitPlugin                                   ref:  rmaplugn.h
 *
 *  This routine performs initialization steps such as determining if
 *  required interfaces are available. It is called when the RMA core 
 *  application is launched, and whenever a stream associated with this
 *  plug-in needs to be rendered.
 */
STDMETHODIMP 
Logger::InitPlugin(IUnknown* /*IN*/ pContext)
{
    PN_RESULT hr				= PNR_OK;
    IRMAErrorSink*	    pErrorSink		= NULL;
    IRMAErrorSinkControl* pErrorSinkControl	= NULL;

    m_hLogFile = fopen(LOGFILE, "w");

    if (!m_hLogFile)
    {
	hr = PNR_UNEXPECTED;
    }
    
    /* tell our context that we wish to receive error messages */
    pContext->QueryInterface(  IID_IRMAErrorSinkControl, 
					(void**) &pErrorSinkControl);
    if ( pErrorSinkControl )
    {	
	this->QueryInterface(  IID_IRMAErrorSink, 
					(void**) &pErrorSink);
	if (pErrorSink)
	{
	    /* add ourselves to the context's error sink list
	     * we are interested in all messages whose severity
	     * ranges from PNLOG_EMERG to PNLOG_INFO.
	     */
	    pErrorSinkControl->AddErrorSink(pErrorSink, PNLOG_EMERG, PNLOG_INFO);
	}
    }	

    return hr;
}


/****************************************************************************
 *  Logger::~Logger                                          ref:  exlogpln.h
 *
 *  Destructor
 */
Logger::~Logger()
{
    if (m_hLogFile)
    {
	fclose(m_hLogFile);
	m_hLogFile = NULL;
    }
}


/****************************************************************************
 *  IRMAErrorSink::ErrorOccurred                             ref:  rmaerror.h
 *
 */
STDMETHODIMP
Logger::ErrorOccurred(
			const UINT8     ulSeverity,
			const UINT32    ulRMACode,
			const UINT32    ulUserCode,
			const char*     pUserString,
			const char*     pMoreInfoURL
		    )
{
    char szLogInfo[MAX_DISPLAY_NAME] = {0};

    switch (ulSeverity)
    {
    case PNLOG_EMERG:
	sprintf(szLogInfo, "Emergency: %s\n", pUserString);
	break;
    case PNLOG_ALERT:
	sprintf(szLogInfo, "Alert: %s\n", pUserString);
	break;
    case PNLOG_CRIT:
	sprintf(szLogInfo, "Critical: %s\n", pUserString);
	break;
    case PNLOG_ERR:
	sprintf(szLogInfo, "Error: %s\n", pUserString);
	break;
    case PNLOG_WARNING:
	sprintf(szLogInfo, "Warning: %s\n", pUserString);
	break;
    case PNLOG_NOTICE:
	sprintf(szLogInfo, "Notice: %s\n", pUserString);
	break;
    case PNLOG_INFO:
	sprintf(szLogInfo, "Information: %s\n", pUserString);
	break;
    case PNLOG_DEBUG:
	sprintf(szLogInfo, "Debug: %s\n", pUserString);
	break;
    default:
	break;
    }
    fprintf(m_hLogFile, szLogInfo);
    fflush(m_hLogFile);

    return PNR_OK;
}


/****************************************************************************
 *  IUnknown::AddRef                                            ref:  pncom.h
 *
 *  This routine increases the object reference count in a thread safe
 *  manner. The reference count is used to manage the lifetime of an object.
 *  This method must be explicitly called by the user whenever a new
 *  reference to an object is used.
 */
STDMETHODIMP_(UINT32)
Logger::AddRef()
{
    return InterlockedIncrement(&m_lRefCount);
}   


/****************************************************************************
 *  IUnknown::Release                                           ref:  pncom.h
 *
 *  This routine decreases the object reference count in a thread safe
 *  manner, and deletes the object if no more references to it exist. It must
 *  be called explicitly by the user whenever an object is no longer needed.
 */
STDMETHODIMP_(UINT32)
Logger::Release()
{
    if (InterlockedDecrement(&m_lRefCount) > 0)
    {
        return m_lRefCount;
    }
    
    delete this;
    return 0;
}   


/****************************************************************************
 *  IUnknown::QueryInterface                                    ref:  pncom.h
 *
 *  This routine indicates which interfaces this object supports. If a given
 *  interface is supported, the object's reference count is incremented, and
 *  a reference to that interface is returned. Otherwise a NULL object and
 *  error code are returned. This method is called by other objects to
 *  discover the functionality of this object.
 */
STDMETHODIMP
Logger::QueryInterface(REFIID riid, void** ppvObj)
{

    if (IsEqualIID(riid, IID_IRMAErrorSink))
    {
        AddRef();
        *ppvObj = (IRMAErrorSink*)this;
        return PNR_OK;
    }
    else if (IsEqualIID(riid, IID_IRMAPlugin))
    {
        AddRef();
        *ppvObj = (IRMAPlugin*)this;
        return PNR_OK;
    }
    else if (IsEqualIID(riid, IID_IUnknown))
    {
        AddRef();
        *ppvObj = (IUnknown*)(IRMAPlugin*)this;
        return PNR_OK;
    }
    
    *ppvObj = NULL;
    return PNR_NOINTERFACE;
}   
