/****************************************************************************
 * 
 *  $Id: exmonpln.cpp,v 1.29 1998/12/09 22:27:12 darrens Exp $
 *
 *  Copyright (C) 1995,1996,1997 RealNetworks, Inc.
 *  All rights reserved.
 *  
 *  http://www.real.com/devzone
 *
 *  This program contains proprietary information of RealNetworks, Inc.,
 *  and is licensed subject to restrictions on use and distribution.
 *
 *  exmonpln.cpp
 *
 *  Example Monitor plug-in for RealMedia Architecture (RMA).
 *
 */

/****************************************************************************
 * 
 *  EXAMPLE MONITOR PLUGIN for RMA Server
 *
 *  This is a very simple monitor plugin, that it monitors the active 
 *  connections on the server and display it at the console screen.
 *
 *  We will set watch on the following properties:
 *  1   server			topmost property level of server object
 *  2   server.clientcount	sub-property associated with the server 
 *                             	object(active connections served by the server)
 *  3   client			topmost property level of client object
 *  4..n client.n.addr		the IP address of the client that has just
 *				just connected (where the 'n' indicates the
 *				client number, i.e. it is the n th. player
 *				that has connected since the server was
 *				started)
 *  
 */

/****************************************************************************
 * Defines
 */
#define INITGUID
#ifdef _DEBUG
    #undef PN_THIS_FILE		
    static char PN_THIS_FILE[] = __FILE__;
#endif


/****************************************************************************
 * Includes
 */
#include <stdio.h>
#include <stdlib.h>

#include "pncom.h"
#include "rmacomm.h"
#include "rmaengin.h"
#include "rmaplugn.h"
#include "rmamon.h"
#include "rmapckts.h"
#include "rmaerror.h"

#include "exmonpln.h"
#include "exmonpln.ver"


/****************************************************************************
 *  RMACreateInstance                                        ref:  rmaplugn.h
 *
 *  This routine creates a new instance of the CExampleRenderer class.
 *  It is called when the RMA core application is launched, and whenever
 *  a stream associated with this plug-in needs to be rendered.
 */
STDAPI
RMACreateInstance(IUnknown** /*OUT*/ ppMonitor)
{
    *ppMonitor = (IUnknown*)(IRMAPlugin*)new CExampleMonitor;
    if (*ppMonitor)
    {
	(*ppMonitor)->AddRef();
	return PNR_OK;
    }

    return PNR_OUTOFMEMORY;
}


/****************************************************************************
 *  CExampleMonitor static variables                         ref:  exmonpln.h
 *
 *  These variables are passed to the RMA core to provide information about
 *  this plug-in. They are required to be static in order to remain valid
 *  for the lifetime of the plug-in.
 */
const char* CExampleMonitor::zm_pDescription    = "Example Monitor Plug-in";
const char* CExampleMonitor::zm_pCopyright      = "(c) 1995,1996,1997, 1998 RealNetworks, All rights reserved.";
const char* CExampleMonitor::zm_pMoreInfoURL    = "http://www.real.com";


/****************************************************************************
 *  IRMAPlugin::GetPluginInfo                                ref:  rmaplugn.h
 *
 *  This routine returns descriptive information about the plug-in, most
 *  of which is used in the About box of the user interface. It is called
 *  when the RMA core application is launched.
 */
STDMETHODIMP 
CExampleMonitor::GetPluginInfo
(
    REF(BOOL)        /*OUT*/ bLoadMultiple,
    REF(const char*) /*OUT*/ pDescription,
    REF(const char*) /*OUT*/ pCopyright,
    REF(const char*) /*OUT*/ pMoreInfoURL,
    REF(UINT32)      /*OUT*/ ulVersionNumber
)
{
    bLoadMultiple   = FALSE;
    pDescription    = zm_pDescription;
    pCopyright	    = zm_pCopyright;
    pMoreInfoURL    = zm_pMoreInfoURL;
    ulVersionNumber = TARVER_ULONG32_VERSION;

    return PNR_OK;
}


/****************************************************************************
 *  IRMAPlugin::InitPlugin                                   ref:  rmaplugn.h
 *
 *  This routine performs initialization steps such as determining if
 *  required interfaces are available. It is called when the RMA core 
 *  application is launched, and whenever a stream associated with this
 *  plug-in needs to be rendered.
 */
STDMETHODIMP 
CExampleMonitor::InitPlugin(IUnknown* /*IN*/ pContext)
{   
    PN_RESULT hr = PNR_OK;
    IRMAPropWatch* pPropWatch = NULL;
    IRMACommonClassFactory* iccf = NULL;

    if (!pContext)
	return PNR_UNEXPECTED;

    m_pContext = pContext;
  
    // get the logger interface
    hr = pContext->QueryInterface(IID_IRMAErrorMessages, (void **)&m_pLogger);
    if (PNR_OK != hr)
    {
	hr = PNR_UNEXPECTED;
	goto cleanup;
    }
 
    hr = pContext->QueryInterface(IID_IRMACommonClassFactory, (void**)&iccf);
    if (PNR_OK != hr)
    {
	hr = PNR_UNEXPECTED;
	goto cleanup;
    }

    hr = pContext->QueryInterface(IID_IRMAPNRegistry, (void**)&m_pRegistry);
    if (PNR_OK != hr)
    {
	hr = PNR_UNEXPECTED;
	goto cleanup;
    }
	    
    // get the interface factory of the registry    
    hr = m_pRegistry->CreatePropWatch(pPropWatch);
    if (PNR_OK != hr)
    {
	hr = PNR_UNEXPECTED;
	goto cleanup;
    }

    // get the IRMAPropWatch interface
    hr = pPropWatch->QueryInterface(IID_IRMAPropWatch, 
				  (void **)&m_pPropWatch);
    if (PNR_OK != hr)
    {
	hr = PNR_UNEXPECTED;
	goto cleanup;
    }

    m_pPropWatch->Init((IRMAPropWatchResponse *)this);

    // initialize the counters and console display 
    m_nPlayers = 0;

    // initialize the properties we are interested in
    strcpy(m_ServerPropList[0].pName, "server");
    strcpy(m_ServerPropList[1].pName, "server.clientcount");
    
    m_ServerPropList[0].id = m_ServerPropList[1].id = 0;

    strcpy(m_ClientPropList[0].pName, "client");
  
    m_ClientPropList[0].id = 0;

    // get the IRMAScheduler interface to poll the data from the registry
    hr = pContext->QueryInterface(IID_IRMAScheduler, (void**)&m_pScheduler);
    if (PNR_OK != hr)
    {
	hr = PNR_UNEXPECTED;
	goto cleanup;
    }

    // initialize the connection list
    m_pConnList = new ConnList();

    // set the watch on player count
    SetWatches();
   
cleanup:

    return hr;
}


/****************************************************************************
 *  CExampleMonitor::UpdateDisplay                           ref:  exmonpln.h
 *
 *  Help functions to prepare/display the active connections at the console
 *  screen.
 */
STDMETHODIMP
CExampleMonitor::UpdateDisplay()
{
    int		i = 0;
    PN_RESULT	hr = PNR_OK;
    ConnNode*	pNode = NULL;
    FILE*	pConsole = stdout;

    // reset the cursor to the connection list head
    m_pConnList->ResetCursor();

    fprintf(pConsole, "\n\n\n\n");
    fprintf(pConsole, "Connection List:\n");

    pNode = m_pConnList->GetNextNode();
    while (pNode)
    {
	fprintf(pConsole, "%i\tIP Address:%s\n", ++i, pNode->m_pszIPAddr);

	pNode = m_pConnList->GetNextNode();
    }

    fprintf(pConsole, "\nActive Connections: %ld\n\n", m_nPlayers);

    fflush(pConsole);

    return hr;
}
  


// IRMAPropWatchResponse Interface Methods

/****************************************************************************
 *  CExampleMonitor::Close                                   ref:  exmonpln.h
 *
 *  Just a clean-up routine in case the Monitor decides to quit.
 */
STDMETHODIMP
CExampleMonitor::Close()
{
    /*
     * delete the plugin, server communication module
     * deletion of the log database is still an issue of persistence
     */
    
    if (m_pContext) 
    {
	m_pContext->Release();
	m_pContext = NULL;
    }

    if (m_pRegistry) 
    {
	m_pRegistry->Release();
	m_pRegistry = NULL;
    }

    if (m_pPropWatch)
    {
	m_pPropWatch->Release();
	m_pPropWatch = NULL;
    }

    if (m_pScheduler)
    {
	m_pScheduler->Release();
	m_pScheduler = NULL;
    }

    if (m_pLogger)
    {
	m_pLogger->Release();
	m_pLogger = NULL;
    }

    if (m_pConnList)
    {
        delete m_pConnList;
        m_pConnList = NULL;
    }

    return PNR_OK;
}


/****************************************************************************
 *  CExampleMonitor::SetWatches                              ref:  exmonpln.h
 *
 *  This method gets called after the InitPlugin() function is called and
 *  the RealMedia server's registry has been created and initialized.
 *  We just set a Watch at the top level of the registry to let us know
 *  if someone adds a new Property at that level. Then we set Watches
 *  on "Server" COMPOSITE Property which will notify us when new
 *  Properties gets added under "Server". Following that we set a watch on
 *  "Server.ClientCount" so that we get notified whenever it changes
 *  (i.e. as and when players connect/disconnect).
 */
STDMETHODIMP
CExampleMonitor::SetWatches()
{
    PN_RESULT hr = PNR_OK;

    // watch for any new Properties at the highest level in database
    m_pPropWatch->SetWatchOnRoot();

    // get the property ID of "server" (parent ID of "server.clientcount")
    m_ServerPropList[0].id = m_pRegistry->GetId(m_ServerPropList[0].pName);
    
    if (0 == m_ServerPropList[0].id)
    {
	hr = PNR_UNEXPECTED;
	goto cleanup;
    }
    
    // get the property ID of "server.clientcount"
    m_ServerPropList[1].id 
	= m_pPropWatch->SetWatchByName(m_ServerPropList[1].pName);

    if (0 == m_ServerPropList[1].id)
    {
	hr = PNR_UNEXPECTED;
	goto cleanup;
    }

    // get the property ID of "client" (parent ID of "client.*")
    m_ClientPropList[0].id = m_pRegistry->GetId(m_ClientPropList[0].pName);

    if (0 == m_ClientPropList[0].id)
    {
	hr = PNR_UNEXPECTED;
	goto cleanup;
    }

    // set watch on the "client" to receive notification on new connection
    if (!m_pPropWatch->SetWatchById(m_ClientPropList[0].id))
    {
	hr = PNR_UNEXPECTED;
	goto cleanup;
    }

cleanup:

    return hr;
}


/****************************************************************************
 *  IRMAPropWatchResponse::AddedProp                           ref:  rmamon.h
 *
 *  We identify the Property that got added and the COMPOSITE under
 *  which it got added. then retreive the ID for the newly added Property,
 *  and if it "Server.PlayerCount" we retreive its value. We also try and
 *  retrieve the Client's address when a Client.n gets added.
 */
STDMETHODIMP
CExampleMonitor::AddedProp(const UINT32 id, 
			   const RMAPropType propType,
			   const UINT32 ulParentID)
{
    PN_RESULT   hr = PNR_OK;
    ConnNode*   pNode = NULL;
    char        szConnProp[MAX_DISPLAY_NAME] = {0};
    char	szLogInfo[MAX_DISPLAY_NAME] = {0};
    char* 	pszNewConnName = new CHAR[MAX_DISPLAY_NAME];
    IRMABuffer* pszConnName = 0;

    if (!pszNewConnName)
    {
	return PNR_OUTOFMEMORY;
    }

    memset(pszNewConnName, 0, MAX_DISPLAY_NAME);
  
    // we only care about the property we are watching at("server.clientcount")
    if (ulParentID == m_ServerPropList[0].id)    	// server
    {
	if (0 == m_ServerPropList[1].id)		// server.clientcount
	{
	    m_ServerPropList[1].id = m_pRegistry->GetId("server.clientcount");
	    // retrieve the property value
	    if (PNR_OK != m_pRegistry->GetIntById(id, m_nPlayers))
	    {
		hr = PNR_UNEXPECTED;
		goto cleanup;
	    }	
	}
    }
    else if (ulParentID == m_ClientPropList[0].id)		    // client.*
    {	     
	// add new connection to the connlist
        pNode = m_pConnList->AddNode(id);

        // get the new connection name (i.e. "client.x")
        if (PNR_OK != m_pRegistry->GetPropName(id, pszConnName))
	{
	    hr = PNR_UNEXPECTED;
	    goto cleanup;
	}

	//
	// log this!
	//
	sprintf(szLogInfo, "%s with ID(%ld) has been ADDED\n", 
	    pszConnName->GetBuffer(), id); 
	m_pLogger->Report(PNLOG_INFO, 0, 0, szLogInfo, 0);

	// 
	// retreive the client.n.addr Property if it has been added
	//
	if (m_pRegistry->GetNumPropsById(id) > 0)
	{
	    m_pRegistry->GetPropName(id, pszConnName);
	    sprintf(pszNewConnName, "%s.addr", pszConnName->GetBuffer());
	    UINT32 clientId 
		= m_pRegistry->GetId(pszNewConnName);
	    if (clientId)
	    {
		IRMABuffer* pValue = NULL;
		if (PNR_OK == m_pRegistry->GetStrById(clientId, pValue)
		    &&
		    pValue)
		{
		    pNode->m_pszIPAddr 
			= new CHAR[strlen((const char*)pValue->GetBuffer())+1];
		    strcpy(pNode->m_pszIPAddr, 
			   (const char*)pValue->GetBuffer());
		    pValue->Release();
		}
	    }
	}
	else
	{
#ifndef _WIN16
	    printf("awaiting addition of the Address\n");
#endif
	    // await addition of the Address
	    m_pPropWatch->SetWatchById(id);
	}
    }
    else if ((pNode = m_pConnList->FindNode(ulParentID)))       	// client.*
    {
	m_pRegistry->GetPropName(ulParentID, pszConnName);
	sprintf(pszNewConnName, "%s.addr", pszConnName->GetBuffer());
	UINT32 clientId 
	    = m_pRegistry->GetId(pszNewConnName);
	if (clientId)
	{
	    IRMABuffer* pValue = NULL;
	    if (PNR_OK == m_pRegistry->GetStrById(clientId, pValue) && pValue)
	    {
		pNode->m_pszIPAddr 
		    = new CHAR[strlen((const char*)pValue->GetBuffer())+1];
		strcpy(pNode->m_pszIPAddr, (const char*)pValue->GetBuffer());
		pValue->Release();

		// clear the watch on client.n because we got the address
		m_pPropWatch->ClearWatchById(ulParentID);
	    }
	}
    }

    // update the counter display
    UpdateDisplay();

cleanup:
    
    if (pszConnName)
    {
	pszConnName->Release();
	pszConnName = NULL;
    }

    if (pszNewConnName)
    {
	delete [] pszNewConnName;
	pszNewConnName = NULL;
    }

    return hr;
}


/****************************************************************************
 *  IRMAPropWatchResponse::ModifiedProp                        ref:  rmamon.h
 *
 *  Just keeping track of the number of players connecting/disconnecting
 *  at any time. we then retreive the current number of players and write
 *  out to a file.
  */
STDMETHODIMP
CExampleMonitor::ModifiedProp(const UINT32 id, 
			      const RMAPropType propType,
			      const UINT32 ulParentID)
{
    PN_RESULT hr = PNR_OK;
    char    szLogInfo[MAX_DISPLAY_NAME] = {0};

    if (id == m_ServerPropList[1].id)			// server.clientcount
    {
	// retrieve the property value(int)
	m_pRegistry->GetIntById(id, m_nPlayers);

	//
	// log this !
	//
	sprintf(szLogInfo, "%s has been MODIFIED\n", m_ServerPropList[1].pName); 
	m_pLogger->Report(PNLOG_INFO, 0, 0, szLogInfo, 0);
    }
   
    // update the counter display
    UpdateDisplay();
   
    return hr;
}


/****************************************************************************
 *  IRMAPropWatchResponse::DeletedProp                         ref:  rmamon.h
 *
 *  Gets called whenver a Property being watched gets deleted. doesn't
 *  do much now, except write the id key of the deleted Property into
 *  a log file and display it on the screen.
 */
STDMETHODIMP
CExampleMonitor::DeletedProp(const UINT32 id, const UINT32 ulParentID)
{
    PN_RESULT     hr = PNR_OK;
    ConnNode*   pNode = NULL;
    char	szLogInfo[MAX_DISPLAY_NAME] = {0};

    if (id == m_ServerPropList[1].id)			// server.clientcount
    {	
	m_ServerPropList[1].id = 0;

	//
	// log this!
	//
	sprintf(szLogInfo, "%s has been DELETED\n", m_ServerPropList[1].pName); 
	m_pLogger->Report(PNLOG_INFO, 0, 0, szLogInfo, 0);
    }
    else if (ulParentID == m_ClientPropList[0].id)		    // client
    {
	// 
	// log this!
	//
	sprintf(szLogInfo, "Client with ID(%ld) has been DELETED\n", id);
	m_pLogger->Report(PNLOG_INFO, 0, 0, szLogInfo, 0);

	if ((pNode = m_pConnList->FindNode(id)))                    // client.x
        {
            m_pConnList->RemoveNode(id);
        }
    }

    // update counter display
    UpdateDisplay();

    return hr;
}


// IUnknown COM Interface Methods

/****************************************************************************
 *  IUnknown::AddRef                                            ref:  pncom.h
 *
 *  This routine increases the object reference count in a thread safe
 *  manner. The reference count is used to manage the lifetime of an object.
 *  This method must be explicitly called by the user whenever a new
 *  reference to an object is used.
 */
STDMETHODIMP_(UINT32)
CExampleMonitor::AddRef()
{
    return InterlockedIncrement(&m_lRefCount);
}   

/****************************************************************************
 *  IUnknown::Release                                           ref:  pncom.h
 *
 *  This routine decreases the object reference count in a thread safe
 *  manner, and deletes the object if no more references to it exist. It must
 *  be called explicitly by the user whenever an object is no longer needed.
 */
STDMETHODIMP_(UINT32)
CExampleMonitor::Release()
{
    if (InterlockedDecrement(&m_lRefCount) > 0)
    {
        return m_lRefCount;
    }
    
    delete this;
    return 0;
}   


/****************************************************************************
 *  IUnknown::QueryInterface                                    ref:  pncom.h
 *
 *  This routine indicates which interfaces this object supports. If a given
 *  interface is supported, the object's reference count is incremented, and
 *  a reference to that interface is returned. Otherwise a NULL object and
 *  error code are returned. This method is called by other objects to
 *  discover the functionality of this object.
 */
STDMETHODIMP
CExampleMonitor::QueryInterface(REFIID riid, void** ppvObj)
{
    if (IsEqualIID(riid, IID_IRMAPlugin))
    {
        AddRef();
        *ppvObj = (IRMAPlugin*)this;
        return PNR_OK;
    }
    else if (IsEqualIID(riid, IID_IRMAPropWatchResponse))
    {
	AddRef();
	*ppvObj = (IRMAPropWatchResponse*)this;
	return PNR_OK;
    }
    else if (IsEqualIID(riid, IID_IUnknown))
    {
        AddRef();
        *ppvObj = (IUnknown*)(IRMAPlugin*)this;
        return PNR_OK;
    }
    
    *ppvObj = NULL;
    return PNR_NOINTERFACE;
} 


// Poll Class Methods


/****************************************************************************
 *  Poll::Poll                                               ref:  exmonpln.h
 *
 */
Poll::Poll(CExampleMonitor*   pMonitor,
	   IRMAPNRegistry*    pRegistry,
	   IRMAScheduler*     pScheduler,
	   char*	      pszConnName)
    : m_pMonitor(pMonitor), m_pRegistry(pRegistry), m_pScheduler(pScheduler),
      m_id(0), m_lRefCount(0)
{
    if (pMonitor) m_pMonitor->AddRef();
    if (pRegistry) m_pRegistry->AddRef();
    if (pScheduler) m_pScheduler->AddRef();

    // the additional 13 bytes are for the Property "UDPResendReqs"
    m_pszConnName = new char[strlen(pszConnName)+13+1];
    sprintf(m_pszConnName, "%s.UDPResendReqs", pszConnName);
}


/****************************************************************************
 *  Poll::~Poll                                              ref:  exmonpln.h
 *
 */
Poll::~Poll()
{
    if (m_pMonitor)
    {
	m_pMonitor->Release();
	m_pMonitor = NULL;
    }

    if (m_pRegistry)
    {
	m_pRegistry->Release();
	m_pRegistry = NULL;
    }

    if (m_pScheduler)
    {
	m_pScheduler->Release();
	m_pScheduler = NULL;
    }

    if (m_pszConnName)
    {
	delete m_pszConnName;
	m_pszConnName = NULL;
    }
}


/****************************************************************************
 *  Poll::Func                                               ref:  exmonpln.h
 *
 */
STDMETHODIMP
Poll::Func()
{
    INT32	lVal = 0;
    UINT32	id = 0;
    char	szPropName[MAX_DISPLAY_NAME] = {0};

    memset(szPropName, 0, MAX_DISPLAY_NAME);

    id = m_pRegistry->FindParentIdByName(m_pszConnName);
    /*
     * retrieve the ID for the client.n.UDPResendReqs
     * this will speed up all subsequent accesses to the RM Server's
     * Registry
     */
    if (!m_id)
	m_id = m_pRegistry->GetId(m_pszConnName);

    if (m_pMonitor->m_pConnList->FindNode(id))
    {
	if (m_id)
	    m_pRegistry->GetIntById(m_id, lVal);
	else
	    m_pRegistry->GetIntByName(m_pszConnName, lVal);

	m_pMonitor->m_pConnList->Cursor()->m_ulUDPResendReqs = lVal;

	m_pMonitor->UpdateDisplay();
    }
    
    m_pScheduler->RelativeEnter(this, 5000);

    return PNR_OK;
}


/****************************************************************************
 *  IUnknown::AddRef                                            ref:  pncom.h
 *
 */
STDMETHODIMP_(UINT32)
Poll::AddRef()
{
    return InterlockedIncrement(&m_lRefCount);
}


/****************************************************************************
 *  IUnknown::Release                                           ref:  pncom.h
 *
 */
STDMETHODIMP_(UINT32)
Poll::Release()
{
    if (InterlockedDecrement(&m_lRefCount) > 0)
    {
	return m_lRefCount;
    }

    delete this;
    return 0;
}


/****************************************************************************
 *  IUnknown::QueryInterface                                    ref:  pncom.h
 *
 */
STDMETHODIMP
Poll::QueryInterface(REFIID riid, void** ppvObj)
{
    if (IsEqualIID(riid, IID_IRMACallback))
    {
	AddRef();
	*ppvObj = (IRMACallback*)this;
	return PNR_OK;
    }
    else if (IsEqualIID(riid, IID_IUnknown))
    {
	AddRef();
	*ppvObj = (IUnknown*)(IRMACallback*)this;
	return PNR_OK;
    }

    *ppvObj = NULL;
    return PNR_NOINTERFACE;
}
