/****************************************************************************
 *
 *  $Id: exppvpln.cpp,v 1.1 1999/06/08 22:31:21 darrens Exp $
 *
 *  Copyright (C) 1998 RealNetworks, Inc.
 *  All rights reserved.
 *
 *  http://www.real.com/devzone
 *
 *  This program contains proprietary information of RealNetworks, Inc.,
 *  and is licensed subject to restrictions on use and distribution.
 *
 *  exppvpln.cpp
 *
 *  Example Allowance plug-in for the RealMedia Architecture (RMA).
 */

/****************************************************************************
 *  This is an example of a Pay Per View Allowance plug-in. The purpose of a
 *  Pay Per View Allowance server plug-in is to monitor player connections 
 *  to the server and use the authentication capabilities supported by the
 *  RealSystem to validate a user for specific pieces of content. A Pay Per
 *  View Allowance plug-in is just like any other Allowance plug-in in that
 *  it receives notifications for each player that connects to the server.
 *  When the plug-in determines that a player is attempting to access secure
 *  (i.e. password protected) content, it uses several RealServer interfaces
 *  to request the identity of the user, and once that user is verified,
 *  they are allowed to access the content.
 *
 *  In order for your plug-in to achieve this functionality, it must
 *  implement the IRMAPlayerConnectionAdviseSink interface which has 
 *  methods for receiving notifications of many events in the course of
 *  each player connection. The IRMAPlugin interface must also be 
 *  implemented to handle fundamental plug-in operations. Finally, the
 *  IRMAServerAuthResponse interface must be implemented to receive the
 *  response to an authentication request. This response will include a
 *  status code indicating whether or not the user was successfully
 *  authenticated.
 *
 *
 *  Example: exppvpln.cpp
 *
 *  This example demonstrates a Pay Per View Allowance Plug-in. It will 
 *  watch all incoming player connections, and will password protect
 *  all URLs that contain the word "protected" in them. When a player tries
 *  to access such a URL, this plug-in will trigger the authentication
 *  process, and, if the player is successfully authenticated, they will
 *  be allowed to play the content.
 *
 *  Note: This plug-in does NOT determine if a user-supplied password is
 *  valid! The job of password validation is handled by an authentication
 *  plug-in, and an example of that plug-in is included elsewhere in the
 *  RMA SDK. This plug-in is simply notified when the authentication
 *  process succeeds or fails.
 *
 *  To test this example plug-in, simply build the plug-in and place it in
 *  your RMA server plug-in directory. Create a "protected" subdirectory 
 *  within the default Content directory of your RMA server and place a 
 *  content file in that new directory. Restart the server and request the 
 *  protected file from your RealServer with a URL like the following:
 *
 *      rtsp://[server]:[port]/protected/g2video.rm
 *
 *  Note that you should replace [server] with your server name and [port]
 *  with the RTSP Port on which your RealServer is listening.
 *
 *  Because this URL contains the word "protected", this sample plug-in will
 *  trigger an authentication request, and the RealPlayer should display
 *  a username and password dialog. Since this plug-in tells the 
 *  authentication feature of the server to use passwords from the default
 *  Administrators Realm, you must enter a valid RealServer administrator 
 *  password in order to successfully view the protected file.
 *
 *  The RMA server calls the various interface routines at well defined times.
 *  The sequence of the major events is as follows:
 *
 *  Upon RMA server startup:
 *  - RMACreateInstance() creates new instance of CMyExampleRenderer class
 *  - GetPlugInInfo()     returns descriptive information about the plug-in
 *
 *  When a player first connects to the server:
 *  - OnConnection()	    called when a player connects
 *  - SetPlayerController() called to provide a player controlling interface
 *  - SetRegistryID()	    called to provide the location of player info.
 *  - OnURL()		    called when the player requests a URL
 *
 *  During the player's connection:
 *  - OnBegin()		    called when player starts or resumes
 *  - OnStop()		    called when player stops
 *  - OnPause()		    called when the playback is paused
 *  - OnDone()		    called when the player has disconnected
 *
 *  After authentication has been requested by this plug-in:
 *  - ChallengeReady()	    called with the success or failure of the 
 *			    authentication attempt.
 *
 *  Note:
 *  The routines are listed roughly in the order in which they are called.
 */


/****************************************************************************
 * Defines
 */
#define	INITGUID     /* Interface ID's */
#define AUTHENTICATION_REALM "config.AuthenticationRealms.SecureAdmin.Realm"

/****************************************************************************
 * Includes
 */
#include <stdio.h>     /* printf, sprintf */
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "pntypes.h" 
#include "pncom.h"	/* IUnknown */
#include "rmacomm.h"	/* IRMACommonClassFactory */
#include "rmapckts.h"	/* IRMABuffer, IRMAPacket, IRMAValues */
#include "rmaplugn.h"	/* IRMAPlugin */
#include "rmaallow.h"	/* IRMAPlayerConnectionAdviseSink */
#include "rmaauth.h"	/* IRMAAuthenticator */
#include "rmaauthn.h"	/* IRMAServerAuthResponse */
#include "rmamon.h"	/* IRMAPNRegistry */
#include "rmafiles.h"	/* IRMARequest */
#include "rmaplgns.h"	/* IRMAObjectConfiguration */
#include "exppvpln.h"   /* CExamplePayPerView */


/****************************************************************************
 *  RMACreateInstance                                        ref:  rmaplugn.h
 *
 *  This routine creates a new instance of the CExamplePayPerView class.
 *  It is called when the RMA server application is launched.
 */
STDAPI
RMACreateInstance(IUnknown** ppExPayPerViewObj)
{
    *ppExPayPerViewObj = (IUnknown*)(IRMAPlugin*)new CExamplePayPerView();
    if (*ppExPayPerViewObj != NULL)
    {
	(*ppExPayPerViewObj)->AddRef();
	return PNR_OK;
    }

    return PNR_OUTOFMEMORY;
}


/****************************************************************************
 *  CExamplePayPerView static variables                       ref: exppvpln.h
 *
 *  These variables are passed to the RMA server to provide information about
 *  this plug-in. They are required to be static in order to remain valid
 *  for the lifetime of the plug-in.
 */
const char* CExamplePayPerView::zm_pDescription = MY_DESCRIPTION;
const char* CExamplePayPerView::zm_pCopyright   = MY_COPYRIGHT;
const char* CExamplePayPerView::zm_pMoreInfoURL = MY_MORE_INFO_URL;


/****************************************************************************
 *  CExamplePayPerView::CExamplePayPerView                    ref: exppvpln.h
 *
 *  Constructor
 */
CExamplePayPerView::CExamplePayPerView(void)
    : m_RefCount	    (0),
      m_pContext	    (NULL),
      m_pClassFactory	    (NULL),
      m_pRegistry	    (NULL),
      m_pPCResponse	    (NULL),
      m_pPlayerController   (NULL),
      m_pAuthConversation   (NULL),
      m_pRequest	    (NULL),
      m_ulPlayerRegistryID  (0),
      m_pPlayerRegistryName (NULL)
{
}


/****************************************************************************
 *  IRMAPlugin::GetPluginInfo                                ref:  rmaplugn.h
 *
 *  This routine returns descriptive information about the plug-in. It is 
 *  called when the RMA server application is launched.
 */
STDMETHODIMP
CExamplePayPerView::GetPluginInfo
(
    REF(BOOL)        bLoadMultiple,
    REF(const char*) pDescription,
    REF(const char*) pCopyright,
    REF(const char*) pMoreInfoURL,
    REF(UINT32)      versionNumber
)
{
    bLoadMultiple = TRUE;
    pDescription  = zm_pDescription;
    pCopyright    = zm_pCopyright;
    pMoreInfoURL  = zm_pMoreInfoURL;
    versionNumber = MY_PLUGIN_VERSION;

    return PNR_OK;
}


/****************************************************************************
 *  IRMAPlugin::InitPlugin                                   ref:  rmaplugn.h
 *
 *  This routine performs initialization steps such as determining if
 *  required interfaces are available. It is called when the RMA core 
 *  application is launched, and whenever a new player Allowance object
 *  is needed.
 */
STDMETHODIMP
CExamplePayPerView::InitPlugin(IUnknown* pContext)
{
    m_pContext = pContext;
    m_pContext->AddRef();

    /*
     * Store a reference to the IRMACommonClassFactory interface which is
     * used to create commonly used RMA objects such as IRMAPacket, 
     * IRMAValues, and IRMABuffers.
     */
    pContext->QueryInterface(IID_IRMACommonClassFactory, 
	                    (void**)&m_pClassFactory);
    if (m_pClassFactory == NULL)
    {
	return PNR_NOTIMPL;
    }

    /*
     * Store a reference to the IRMAPNRegistry interface which is
     * used to retrieve useful information about the connected player
     */
    pContext->QueryInterface(IID_IRMAPNRegistry, 
	                    (void**)&m_pRegistry);
    if (m_pRegistry == NULL)
    {
	return PNR_NOTIMPL;
    }

    /* 
     * Note that QueryInterface() takes care of adding a reference to the
     * interface for us. You are however responsible for releasing the
     * reference to the object when you are done using it by calling
     * the Release() routine.
     */

    return PNR_OK;
}


/****************************************************************************
 *  IRMAPlayerConnectionAdviseSink::OnConnection             ref:  rmaallow.h
 *
 *  The server core calls this routine when a new player has connected
 *  to the server.
 */
STDMETHODIMP 
CExamplePayPerView::OnConnection(THIS_ 
				 IRMAPlayerConnectionResponse* pResponse)
{
    m_pPCResponse = pResponse;
    m_pPCResponse->AddRef();

    m_pPCResponse->OnConnectionDone(PNR_OK);

    return PNR_OK;
}

/****************************************************************************
 *  IRMAPlayerConnectionAdviseSink::SetPlayerController      ref:  rmaallow.h
 *
 *  The server core calls this routine to supply the Allowance plug-in with
 *  an interface by which the player can be controlled. For example, this 
 *  interface can be used to disconnect the player, send them an alert 
 *  message, or redirect them to another URL.
 */
STDMETHODIMP
CExamplePayPerView::SetPlayerController(THIS_ 
				        IRMAPlayerController* pPlayerController)
{
    m_pPlayerController = pPlayerController;
    m_pPlayerController->AddRef();

    return PNR_OK;
}

/****************************************************************************
 *  IRMAPlayerConnectionAdviseSink::SetRegistryID            ref:  rmaallow.h
 *
 *  The server core calls this routine to supply the Allowance plug-in with
 *  a server registry ID, which can be used to find out various pieces of
 *  information about the connected player. For example, the player's IP
 *  Address, GUID, and Protocol are all stored under this key in the server
 *  registry.
 */
STDMETHODIMP
CExamplePayPerView::SetRegistryID(THIS_ UINT32 ulPlayerRegistryID)
{
    m_ulPlayerRegistryID = ulPlayerRegistryID;

    // Obtain the name of the key for obtaining subkeys later
    m_pRegistry->GetPropName(ulPlayerRegistryID, m_pPlayerRegistryName);

    return PNR_OK;
}

/****************************************************************************
 *  IRMAPlayerConnectionAdviseSink::OnURL                    ref:  rmaallow.h
 *
 *  The server core calls this routine when the connected player requests
 *  a URL. The requested URL is stored in the passed IRMARequest object,
 *  along with all 822 Headers that accompanied the request.
 */
STDMETHODIMP 
CExamplePayPerView::OnURL(THIS_ IRMARequest* pRequest)
{
    const char* pURL = NULL;

    if (m_pRequest)
    {
	m_pRequest->Release();
	m_pRequest = NULL;
    }
    m_pRequest = pRequest;
    m_pRequest->AddRef();

    // Obtain the requested URL
    m_pRequest->GetURL(pURL);

    if (strstr(pURL, "protected"))
    {
	/*
	 * The player requested a protected URL
	 *
	 * Note: We won't call OnURLDone() until we have determined if this 
	 * user is authorized to view the secure content or not.
	 */
	HandleSecureContentRequest();
    }
    else
    {
	/*
	 * Note: Most IRMAPlayerConnectionAdviseSink interface functions have
	 * a corresponding response function in the 
	 * IRMAPlayerConnectionResponse interface. This function should be 
	 * called when the plug-in has finished processing the notification.
	 *
	 * The connected player will not be permitted to begin playback until 
	 * all allowance plugins have called OnURLDone() with a Status of 
	 * PNR_OK.
	 */
	m_pPCResponse->OnURLDone(PNR_OK);
    }

    return PNR_OK;
}

/****************************************************************************
 *  IRMAPlayerConnectionAdviseSink::OnBegin                  ref:  rmaallow.h
 *
 *  The server core calls this routine when the player begins or resumes
 *  playback.
 */
STDMETHODIMP 
CExamplePayPerView::OnBegin(void)
{
    m_pPCResponse->OnBeginDone(PNR_OK);

    return PNR_OK;
}

/****************************************************************************
 *  IRMAPlayerConnectionAdviseSink::OnPause                  ref:  rmaallow.h
 *
 *  The server core calls this routine when the player pauses playback.
 */
STDMETHODIMP 
CExamplePayPerView::OnPause(void)
{
    m_pPCResponse->OnPauseDone(PNR_OK);

    return PNR_OK;
}

/****************************************************************************
 *  IRMAPlayerConnectionAdviseSink::OnStop                   ref:  rmaallow.h
 *
 *  The server core calls this routine when the player stops playback.
 */
STDMETHODIMP 
CExamplePayPerView::OnStop(void)
{
    m_pPCResponse->OnStopDone(PNR_OK);

    return PNR_OK;
}

/****************************************************************************
 *  IRMAPlayerConnectionAdviseSink::OnDone                   ref:  rmaallow.h
 *
 *  The server core calls this routine when the player has disconnected
 *  from the server.
 *  
 */
STDMETHODIMP 
CExamplePayPerView::OnDone(void)
{
    return PNR_OK;
}


/****************************************************************************
 *  CExamplePayPerView::~CExamplePayPerView                 ref:   exppvpln.h
 *
 *  Destructor
 */
CExamplePayPerView::~CExamplePayPerView(void)
{
    if (m_pContext)
    {
	m_pContext->Release();
	m_pContext = NULL;
    }

    if (m_pClassFactory)
    {
	m_pClassFactory->Release();
	m_pClassFactory = NULL;
    }

    if (m_pRegistry)
    {
	m_pRegistry->Release();
	m_pRegistry = NULL;
    }

    if (m_pPCResponse)
    {
	m_pPCResponse->Release();
	m_pPCResponse = NULL;
    }

    if (m_pPlayerController)
    {
	m_pPlayerController->Release();
	m_pPlayerController = NULL;
    }

    if (m_pAuthConversation)
    {
	m_pAuthConversation->Release();
	m_pAuthConversation = NULL;
    }

    if (m_pRequest)
    {
	m_pRequest->Release();
	m_pRequest = NULL;
    }

    if (m_pPlayerRegistryName)
    {
	m_pPlayerRegistryName->Release();
	m_pPlayerRegistryName = NULL;
    }
}


// IUnknown COM Interface Methods

/****************************************************************************
 *  IUnknown::AddRef                                            ref:  pncom.h
 *
 *  This routine increases the object reference count in a thread safe
 *  manner. The reference count is used to manage the lifetime of an object.
 *  This method must be explicitly called by the user whenever a new
 *  reference to an object is used.
 */
STDMETHODIMP_(UINT32)
CExamplePayPerView::AddRef(void)
{
    return InterlockedIncrement(&m_RefCount);
}


/****************************************************************************
 *  IUnknown::Release                                           ref:  pncom.h
 *
 *  This routine decreases the object reference count in a thread safe
 *  manner, and deletes the object if no more references to it exist. It must
 *  be called explicitly by the user whenever an object is no longer needed.
 */
STDMETHODIMP_(UINT32)
CExamplePayPerView::Release(void)
{
    if (InterlockedDecrement(&m_RefCount) > 0)
    {
	return m_RefCount;
    }

    delete this;
    return 0;
}


/****************************************************************************
 *  IUnknown::QueryInterface                                    ref:  pncom.h
 *
 *  This routine indicates which interfaces this object supports. If a given
 *  interface is supported, the object's reference count is incremented, and
 *  a reference to that interface is returned. Otherwise a NULL object and
 *  error code are returned. This method is called by other objects to
 *  discover the functionality of this object.
 */
STDMETHODIMP
CExamplePayPerView::QueryInterface
(
    REFIID interfaceID,
    void** ppInterfaceObj
)
{
    // By definition all COM objects support the IUnknown interface
    if (IsEqualIID(interfaceID, IID_IUnknown))
    {
	AddRef();
	*ppInterfaceObj = (IUnknown*)(IRMAPlugin*)this;
	return PNR_OK;
    }

    // IRMAPlugin interface is supported
    else if (IsEqualIID(interfaceID, IID_IRMAPlugin))
    {
	AddRef();
	*ppInterfaceObj = (IRMAPlugin*)this;
	return PNR_OK;
    }

    // IRMAPlayerConnectionAdviseSink interface is supported
    else if (IsEqualIID(interfaceID, IID_IRMAPlayerConnectionAdviseSink))
    {
	AddRef();
	*ppInterfaceObj = (IRMAPlayerConnectionAdviseSink*)this;
	return PNR_OK;
    }

    // IRMAServerAuthResponse interface is supported
    else if (IsEqualIID(interfaceID, IID_IRMAServerAuthResponse))
    {
	AddRef();
	*ppInterfaceObj = (IRMAServerAuthResponse*)this;
	return PNR_OK;
    }

    // No other interfaces are supported
    *ppInterfaceObj = NULL;
    return PNR_NOINTERFACE;
}

void
CExamplePayPerView::HandleSecureContentRequest()
{
    PN_RESULT		     hResult	    = PNR_OK;
    IRMABuffer*		     pBuffer	    = NULL;
    IRMAValues*		     pValues	    = NULL;
    IUnknown*		     pAuthenticator = NULL;
    IRMAObjectConfiguration* pAuthConfig    = NULL;
    
    if (m_pAuthConversation)
    {
	m_pAuthConversation->Release();
	m_pAuthConversation = NULL;
    }
    
    /*
     * Start the authentication process by creating an authenticator
     * object, and giving it all of the configuration information
     * it needs to properly challenge the user.
     */
    m_pClassFactory->CreateInstance(CLSID_CRMAServerAuthenticator,
	(void**)&pAuthenticator);

    hResult = pAuthenticator->QueryInterface(IID_IRMAObjectConfiguration, 
	(void**)&pAuthConfig);

    if (PNR_OK == hResult && pAuthConfig)
    {
	// Set the Authenticator's Context
	pAuthConfig->SetContext(m_pContext);

	// Get the Admin Realm from the default server config file
	// Only users registered in that Realm will be allowed to 
	// access this secure piece of content.
	m_pRegistry->GetStrByName((const char*)AUTHENTICATION_REALM,
	    pBuffer);

	// Set the Authenticator's Configuration
	m_pClassFactory->CreateInstance(CLSID_IRMAValues, (void**)&pValues);
	pValues->SetPropertyCString("Realm", pBuffer);
	pAuthConfig->SetConfiguration(pValues);

	if (pValues)
	{
	    pValues->Release();
	    pValues = NULL;
	}

	if (pBuffer)
	{
	    pBuffer->Release();
	    pBuffer = NULL;
	}

	hResult = pAuthenticator->
	    QueryInterface(IID_IRMAServerAuthConversation, 
		(void**)&m_pAuthConversation);

	if (PNR_OK == hResult && m_pAuthConversation)
	{
	    /*
	     * Actually Challenge the user. This call will result in a
	     * ChallengeReady() response which will tell us if the user
	     * has responded successfully to the authentication request.
	     */
	    m_pAuthConversation->MakeChallenge(this, m_pRequest);
	}

	// Flow continues in IRMAServerAuthResponse::ChallengeReady()...
    }

    if (pAuthenticator)
    {
	pAuthenticator->Release();
	pAuthenticator = NULL;
    }

    if (pAuthConfig)
    {
	pAuthConfig->Release();
	pAuthConfig = NULL;
    }
}

/****************************************************************************
 *  IRMAServerAuthResponse::ChallengeReady                   ref:  rmaauthn.h
 *
 *  The server core calls this routine in response to a MakeChallenge() call.
 *  The status passed as the first parameter indicates whether or not the
 *  client successfully responded to the authentication request. The second
 *  parameter contains the challenge headers, from which we can obtain the
 *  username that the user supplied, if they were authenticated successfully.
 */
STDMETHODIMP
CExamplePayPerView::ChallengeReady(PN_RESULT pnrStatus, 
			           IRMARequest* pIRMARequestChallengeHeaders)
{
    PN_RESULT		hResult		= PNR_OK;
    IRMABuffer*		pUserid		= NULL;
    IRMARequestContext* pRequestContext = NULL;
    IUnknown*		pUserContext	= NULL;
    IRMAUserProperties* pUserProperties = NULL;

    if (!m_pAuthConversation)
    {
	return PNR_UNEXPECTED;
    }

    if (!m_pAuthConversation->IsAuthenticated())
    {
	/*
	 * The user was NOT successfully authenticated! They are 
	 * not authorized to play the clip they requested
	 */

	m_pPCResponse->OnURLDone(PNR_NOT_AUTHORIZED);
	return PNR_OK;
    }

    /*
     * The user was successfully authenticated!!! We will 
     * let them play the clip they requested
     */

    // Obtain the Username
    hResult = pIRMARequestChallengeHeaders->
	QueryInterface(IID_IRMARequestContext, (void**)&pRequestContext);
    if (PNR_OK == hResult)
    {
	hResult = pRequestContext->GetUserContext(pUserContext);
	if (PNR_OK == hResult)
	{
	    hResult = pUserContext->QueryInterface(IID_IRMAUserProperties, 
		(void**)&pUserProperties);
	    if (PNR_OK == hResult)
	    {
		// This PrincipalID is the username that the client used
		// to successfully complete authentication...
		hResult = pUserProperties->GetPrincipalID(pUserid);
	    }
	}
    }

    if (pRequestContext)
    {
	pRequestContext->Release();
	pRequestContext = NULL;
    }

    if (pUserContext)
    {
	pUserContext->Release();
	pUserContext = NULL;
    }

    if (pUserProperties)
    {
	pUserProperties->Release();
	pUserProperties = NULL;
    }

    if (pUserid)
    {
	pUserid->Release();
	pUserid = NULL;
    }

    // Return a status of PNR_OK in the OnURLDone() call to indicate 
    // that the client is permitted to play the content they requested.
    m_pPCResponse->OnURLDone(PNR_OK);

    return PNR_OK;
}
