/****************************************************************************
 * 
 *  $Id: pstmixrd.cpp,v 1.22 1999/02/25 21:08:25 jfarr Exp $
 *
 *  Copyright (C) 1995,1996,1997 RealNetworks, Inc.
 *  All rights reserved.
 *  
 *  http://www.real.com/devzone
 *
 *  This program contains proprietary information of RealNetworks, Inc.,
 *  and is licensed subject to restrictions on use and distribution.
 *
 *  pstmixrd.cpp
 *
 *  Post-mix example Renderer for RealMedia Architecture (RMA).
 *
 *  There are three scenarios in which one can use pre/post-mix audio hooks.
 *
 *  1. Implement them in a top level client, for example how the RealPlayer
 *     implements the graphical EQ and spectrum analyzer.
 *
 *  2. Implement them in a generic plugin, which gets instantiated at core
 *     startup time and remains for the life of the core.
 *
 *  3. Implement them in a rendering plugin, which must be loaded in a SMIL 
 *     file in parallel with the audio content. This is how the example 
 *     plugins work.
 *
 */


/****************************************************************************
 * Defines
 */
#define INITGUID     /* Interface ID's */
#define STRICT

#ifdef _DEBUG
#undef PN_THIS_FILE		
static char PN_THIS_FILE[] = __FILE__;
#endif


/****************************************************************************
 * Includes
 */
#include <stdio.h>
#include <signal.h>
#ifdef _UNIX
    #include <sys/types.h>
    #include <sys/stat.h>
    #include <unistd.h>
    #include <fcntl.h>
#endif

#include "pntypes.h"
#include "pncom.h"
#include "rmacomm.h"
#include "rmapckts.h"
#include "rmafiles.h"
#include "rmacore.h"
#include "rmarendr.h"
#include "rmaplugn.h"
#include "rmaausvc.h"

#include "pstmixrd.h"
#include "pmrendr.ver"


/****************************************************************************
 *  RMACreateInstance                                        ref:  rmaplugn.h
 *
 *  This routine creates a new instance of the CPostMixRenderer class.
 *  It is called when the RMA core application is launched, and whenever
 *  a stream associated with this plug-in needs to be rendered.
 */
STDAPI
RMACreateInstance(IUnknown** ppIUnknown)
{
    *ppIUnknown = (IUnknown*)(IRMAPlugin*)new CPostMixRenderer();
    if (*ppIUnknown)
    {
	(*ppIUnknown)->AddRef();
	return PNR_OK;
    }
    return PNR_OUTOFMEMORY;
}

/****************************************************************************
 *  CPostMixRenderer static variables                        ref:  pstmixrd.h
 *
 *  These variables are passed to the RMA core to provide information about
 *  this plug-in. They are required to be static in order to remain valid
 *  for the lifetime of the plug-in.
 */
const char* CPostMixRenderer::zm_pDescription    = "Post Mix Test Renderer Plugin";
const char* CPostMixRenderer::zm_pCopyright      = "(c) 1995,1996,1997 RealNetworks, All rights reserved.";
const char* CPostMixRenderer::zm_pMoreInfoURL    = "http://www.real.com";
const char* CPostMixRenderer::zm_pStreamMimeTypes[]= {"audio/x-pn-pst",NULL};


/****************************************************************************
 *  CPostMixRenderer::CPostMixRenderer                       ref:  pstmixrd.h
 *
 *  Constructor
 */
CPostMixRenderer::CPostMixRenderer() : 
  m_lRefCount(0)
, m_pContext(NULL)
, m_pStream(NULL)
, m_pPlayer(NULL)
, m_pValues(NULL)
, m_ulLastTime(0)
, m_pAudioPlayer(0)
, m_pHook(NULL)
, m_bInSeekMode(FALSE)
{ 
}


/****************************************************************************
 *  IRMAPlugin::GetPluginInfo                                ref:  rmaplugn.h
 *
 *  This routine returns descriptive information about the plug-in, most
 *  of which is used in the About box of the user interface. It is called
 *  when the RMA core application is launched.
 */
STDMETHODIMP
CPostMixRenderer::GetPluginInfo
(
    REF(BOOL)        /*OUT*/ bLoadMultiple,
    REF(const char*) /*OUT*/ pDescription,
    REF(const char*) /*OUT*/ pCopyright,
    REF(const char*) /*OUT*/ pMoreInfoURL,
    REF(ULONG32)     /*OUT*/ ulVersionNumber
)
{
    bLoadMultiple = FALSE;   // Must be true for file formats.

    pDescription    = zm_pDescription;
    pCopyright	    = zm_pCopyright;
    pMoreInfoURL    = zm_pMoreInfoURL;
    ulVersionNumber = TARVER_ULONG32_VERSION;

    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::GetRendererInfo                            ref:  rmarendr.h
 *
 *  This routine returns crucial information required to associate this
 *  plug-in with a given stream MIME type. This information tells the RMA
 *  core which Renderer to use to display a particular type of stream. This
 *  method is called when the RMA core application is launched.
 */
STDMETHODIMP
CPostMixRenderer::GetRendererInfo
(
    REF(const char**) /*OUT*/ pStreamMimeTypes,
    REF(UINT32)       /*OUT*/ unInitialGranularity
)
{
    pStreamMimeTypes = zm_pStreamMimeTypes;
    unInitialGranularity = 50;  // OnTimeSync() every 50ms (this is the same as
				// the TIME_PER_PACKET in wvffplin.cpp
    m_ulGranularity = unInitialGranularity;
    return PNR_OK;
}


/****************************************************************************
 *  IRMAPlugin::InitPlugin                                   ref:  rmaplugn.h
 *
 *  This routine performs initialization steps such as determining if
 *  required interfaces are available. It is called when the RMA core 
 *  application is launched, and whenever a stream associated with this
 *  plug-in needs to be rendered.
 */
STDMETHODIMP
CPostMixRenderer::InitPlugin(IUnknown* /*IN*/ pContext)
{
    m_pContext = pContext;
    m_pContext->AddRef();

    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::StartStream                                ref:  rmarendr.h
 *
 *  The RMA core calls this routine to provide access to the stream and
 *  player. It is called when the plug-in is being initialized.
 */
STDMETHODIMP
CPostMixRenderer::StartStream
(
    IRMAStream*	    pStream,
    IRMAPlayer*	    pPlayer
)
{

    // Save for later use!
    m_pStream  = pStream;
    m_pPlayer  = pPlayer;

    if (m_pStream ) m_pStream->AddRef();
    if (m_pPlayer ) m_pPlayer->AddRef();

    // get interface to audio player object
    if ( PNR_OK != m_pPlayer->QueryInterface(IID_IRMAAudioPlayer, (void**) &m_pAudioPlayer ))
    {
        return PNR_FAILED;
    }

    // create a post mix hook object
    m_pHook   = (CPostMixHook *) new CPostMixHook;
    if ( m_pHook ) 
    {
	m_pHook->AddRef();
	/* we need the context to create an IRMABuffer. */
	m_pHook->SetIContext(m_pContext);
    }

    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnHeader                                   ref:  rmarendr.h
 *
 *  This routine is passed the stream header object created by the associated
 *  file format plug-in. Use the methods defined in the IRMAValues interface
 *  to access the stream header data. This method is called from the RMA
 *  core when the plug-in is being initialized.
 */
STDMETHODIMP
CPostMixRenderer::OnHeader(IRMAValues* pValues)
{
    m_pValues = pValues;
    m_pValues->AddRef();

    IRMABuffer* pOpaqueData = 0;
    ULONG32 ulStreamNumber;
    ULONG32 ulMaxBitRate;
    ULONG32 ulAvgBitRate;
    ULONG32 ulMaxPacketSize;
    ULONG32 ulAvgPacketSize;
    ULONG32 ulStartTime;
    ULONG32 ulPreroll;
    ULONG32 ulDuration;
    IRMABuffer* pStreamName = 0;
    IRMABuffer* pMimeType = 0;

    pValues->GetPropertyBuffer ("OpaqueData",       pOpaqueData);
    pValues->GetPropertyULONG32("StreamNumber",     ulStreamNumber);
    pValues->GetPropertyULONG32("MaxBitRate",       ulMaxBitRate);
    pValues->GetPropertyULONG32("AvgBitRate",       ulAvgBitRate);
    pValues->GetPropertyULONG32("MaxPacketSize",    ulMaxPacketSize);
    pValues->GetPropertyULONG32("AvgPacketSize",    ulAvgPacketSize);
    pValues->GetPropertyULONG32("StartTime",        ulStartTime);
    pValues->GetPropertyULONG32("Preroll",          ulPreroll);
    pValues->GetPropertyULONG32("Duration",         ulDuration);
    pValues->GetPropertyBuffer ("StreamName",       pStreamName);
    pValues->GetPropertyBuffer ("MimeType",         pMimeType);

    /* Add post mix hook            */
    BOOL bDisableWrite = FALSE;
    BOOL bFinal = FALSE;
    m_pAudioPlayer->AddPostMixHook(
				(IRMAAudioHook*) m_pHook, 
				bDisableWrite,
				bFinal);

    if (pOpaqueData)	pOpaqueData->Release();
    if (pStreamName)	pStreamName->Release();
    if (pMimeType)	pMimeType->Release();

    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnBegin                                   ref:  rmarendr.h
 *
 *  This routine is called by the RMA core to inform the Renderer that 
 *  playback has just begun or has been resumed. The stream's time value just
 *  after resuming the playback is provided.
 */
STDMETHODIMP
CPostMixRenderer::OnBegin(ULONG32 ulTime)
{
    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::GetDisplayType                             ref:  rmarendr.h
 *
 *  This routine returns the preferred display type of the Renderer. Any
 *  other additional information necessary for the display can also be
 *  provided. This method is called by the RMA core when the plug-in is being
 *  initialized.
 */
STDMETHODIMP
CPostMixRenderer::GetDisplayType
(
    REF(RMA_DISPLAY_TYPE)   ulFlags,
    REF(IRMABuffer*)	    pBuffer
)
{
    //  Our renderer needs no visible display.
    ulFlags = RMA_DISPLAY_NONE;

    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnPacket                                   ref:  rmarendr.h
 *
 *  This routine is passed the packet object streamed by the associated
 *  file format plug-in. It is called by the RMA core when a packet is due
 *  to be delivered. In most cases, the actual rendering of the packet is
 *  done in this method. The stream's time offset with respect to the master
 *  timeline is provided.
 */
STDMETHODIMP
CPostMixRenderer::OnPacket(IRMAPacket* pPacket, LONG32 lTimeOffset)
{
    /* Ignore pre-seek packets. Some renderers may want to use them though */
    if (m_bInSeekMode)
    {
	return PNR_OK;
    }

    // did we get a valid packet?
    if (pPacket->IsLost())
	return PNR_OK;

    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnTimeSync                                 ref:  rmarendr.h
 *
 *  This routine is called by the RMA core periodically passing the current
 *  playback time. The Renderer should use this time value to synchronize the
 *  playback of various streams.
 */
STDMETHODIMP
CPostMixRenderer::OnTimeSync(ULONG32 ulTime)
{
    m_ulLastTime = ulTime;
    
    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnPreSeek                                  ref:  rmarendr.h
 *
 *  This routine is called by the RMA core before a seek is about to occur.
 *  The stream's time value just before the seek, and the time value just
 *  after the seek are provided.
 */
STDMETHODIMP
CPostMixRenderer::OnPreSeek(ULONG32 ulOldTime, ULONG32 ulNewTime)
{
    m_bInSeekMode   = TRUE;
    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnPostSeek                                 ref:  rmarendr.h
 *
 *  This routine is called by the RMA core just after a seek has occured.
 *  The stream's time value just before the seek, and the time value just
 *  after the seek are provided.
 */
STDMETHODIMP
CPostMixRenderer::OnPostSeek(ULONG32 ulOldTime, ULONG32 ulNewTime)
{
    m_bInSeekMode   = FALSE;
    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnPause                                    ref:  rmarendr.h
 *
 *  This routine is called by the RMA core just after a pause has occured.
 *  The stream's time value just before pausing is provided.
 */
STDMETHODIMP
CPostMixRenderer::OnPause(ULONG32 ulTime)
{
    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnBuffering                                ref:  rmarendr.h
 *
 *  This routine is called by the RMA core to inform the Renderer that
 *  buffering of data is occuring. The reason for buffering (e.g. start-up
 *  of stream, seek has occured, network congestion, etc.), as well as the
 *  percentage of the buffering process complete are provided.
 */
STDMETHODIMP
CPostMixRenderer::OnBuffering(ULONG32 ulFlags, UINT16 unPercentComplete)
{
    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnEndofPackets                             ref:  rmarendr.h
 *
 *  Called by client engine to inform the renderer that all the
 *  packets have been delivered. However, if the user seeks before
 *  EndStream() is called, renderer may start getting packets again
 *  and the client engine will eventually call this function again.
 */
STDMETHODIMP
CPostMixRenderer::OnEndofPackets(void)
{
    return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::EndStream                                  ref:  rmarendr.h
 *
 *  This routine is called by the RMA core when the rendering of the stream
 *  has completed. Deallocation of any resources should be done here.
 */
STDMETHODIMP
CPostMixRenderer::EndStream()
{
    // We're done with these...
    if (m_pValues) m_pValues->Release();	    m_pValues	   = NULL;
    if (m_pContext) m_pContext->Release();	    m_pContext	   = NULL;
    if (m_pStream) m_pStream->Release();	    m_pStream	   = NULL;
    if (m_pPlayer) m_pPlayer->Release();	    m_pPlayer	   = NULL;

    if (m_pHook) 
    {
	if (m_pAudioPlayer)
	{
	    m_pAudioPlayer->RemovePostMixHook((IRMAAudioHook*) m_pHook);
	}

	m_pHook->Release();     	    
	m_pHook	   = NULL;
    }

    if (m_pAudioPlayer) m_pAudioPlayer->Release();  m_pAudioPlayer = NULL;

    return PNR_OK;
}


// IUnknown COM Interface Methods

/****************************************************************************
 *  IUnknown::AddRef                                            ref:  pncom.h
 *
 *  This routine increases the object reference count in a thread safe
 *  manner. The reference count is used to manage the lifetime of an object.
 *  This method must be explicitly called by the user whenever a new
 *  reference to an object is used.
 */
STDMETHODIMP_(ULONG32)
CPostMixRenderer::AddRef()
{
    return InterlockedIncrement(&m_lRefCount);
}


/****************************************************************************
 *  IUnknown::Release                                           ref:  pncom.h
 *
 *  This routine decreases the object reference count in a thread safe
 *  manner, and deletes the object if no more references to it exist. It must
 *  be called explicitly by the user whenever an object is no longer needed.
 */
STDMETHODIMP_(ULONG32)
CPostMixRenderer::Release()
{
    if (InterlockedDecrement(&m_lRefCount) > 0)
    {
        return m_lRefCount;
    }

    delete this;
    return 0;
}


/****************************************************************************
 *  IUnknown::QueryInterface                                    ref:  pncom.h
 *
 *  This routine indicates which interfaces this object supports. If a given
 *  interface is supported, the object's reference count is incremented, and
 *  a reference to that interface is returned. Otherwise a NULL object and
 *  error code are returned. This method is called by other objects to
 *  discover the functionality of this object.
 */
STDMETHODIMP
CPostMixRenderer::QueryInterface(REFIID riid, void** ppvObj)
{
    if (IsEqualIID(riid, IID_IUnknown))
    {
	AddRef();
	*ppvObj = (IUnknown*)(IRMAPlugin*)this;
	return PNR_OK;
    }
    else if (IsEqualIID(riid, IID_IRMAPlugin))
    {
	AddRef();
	*ppvObj = (IRMAPlugin*)this;
	return PNR_OK;
    }
    else if (IsEqualIID(riid, IID_IRMARenderer))
    {
	AddRef();
	*ppvObj = (IRMARenderer*)this;
	return PNR_OK;
    }
    else if (IsEqualIID(riid, IID_IRMAAudioPlayer))
    {
        AddRef();
        *ppvObj = (IRMAAudioPlayer*)this;
        return PNR_OK;
    }

    *ppvObj = NULL;
    return PNR_NOINTERFACE;
}


// CPostMixHook  Class Implementation


/****************************************************************************
 *  CPostMixRenderer::CPostMixHook::CPostMixHook            ref:  pstmixrd.h
 *
 *  Constructor
 */
CPostMixRenderer::CPostMixHook::CPostMixHook() :
     m_lRefCount (0)
    ,m_uChannels(0)
    ,m_uBitsPerSample(0)
    ,m_ulSamplesPerSec(0)
    ,m_uMaxBlockSize(0)
    ,m_pContext(0)
    ,m_pCommonClassFactory(0)
{
}

/****************************************************************************
 *  CPostMixRenderer::CPostMixHook::~CPostMixHook           ref:  pstmixrd.h
 *
 *  Destructor
 */
CPostMixRenderer::CPostMixHook::~CPostMixHook()
{
    if (m_pContext)
    {
	m_pContext->Release();
	m_pContext = 0;
    }

    if (m_pCommonClassFactory)
    {
	m_pCommonClassFactory->Release();
	m_pCommonClassFactory = 0;
    }
}


/****************************************************************************
 *  CPostMixRenderer::CPostMixHook::SetIContext             ref:  pstmixrd.h
 *
 */
int CPostMixRenderer::CPostMixHook::SetIContext
( 
    IUnknown* pIU
)
{
    if (!pIU)
    {
	return PNR_INVALID_PARAMETER;
    }

    m_pContext = pIU;
    m_pContext->AddRef();

    if (PNR_OK != m_pContext->QueryInterface(IID_IRMACommonClassFactory,
                                (void **)&m_pCommonClassFactory))
    {
        return PNR_UNEXPECTED;
    }

    return PNR_OK;
}


// IRMAAudioHook Interface Methods

/****************************************************************************
 *  IRMAAudioHook::OnInit                                    ref:  rmaausvc.h
 *
 */
STDMETHODIMP
CPostMixRenderer::CPostMixHook::OnInit( RMAAudioFormat* pFormat )
{
#ifndef _WIN16    
    printf("Post-mix renderer audio format:\n");
    printf("Channels:            %d\n",  pFormat->uChannels);
    printf("BitsPerSample:       %d\n",  pFormat->uBitsPerSample);
    printf("Sampling Rate:       %ld\n", pFormat->ulSamplesPerSec);
    printf("Block Size:          %d\n",  pFormat->uMaxBlockSize);
#endif

    m_uChannels         = pFormat->uChannels;
    m_uBitsPerSample    = pFormat->uBitsPerSample;
    m_ulSamplesPerSec   = pFormat->ulSamplesPerSec;
    m_uMaxBlockSize     = pFormat->uMaxBlockSize; 

    return PNR_OK;
}

/****************************************************************************
 *  IRMAAudioHook::OnBuffer                                  ref:  rmaausvc.h
 *
 */
STDMETHODIMP
CPostMixRenderer::CPostMixHook::OnBuffer
( 
    RMAAudioData* pInData,
    RMAAudioData* pOutData 
)
{
    ULONG32 ulLen = 0;
    IRMABuffer* pModifiedData;

    /* Create a temporary buffer for audio data. */
    ulLen = pInData->pData->GetSize();

    /* Create an IRMABuffer for modified data. */
    if (PNR_OK != m_pCommonClassFactory->CreateInstance(CLSID_IRMABuffer,
                                (void**)&pModifiedData))
    {
        return PNR_UNEXPECTED;
    }

    pModifiedData->SetSize(ulLen);

    if (m_uBitsPerSample == 16)
    {
	short* pInBuf = (short*) pInData->pData->GetBuffer();
	short* pbuf   = (short*) pModifiedData->GetBuffer();
	/* Just modify the volume... */
	for(UINT16 i = 0; i < (ulLen/2); i++)   // divide by 2 for 16-bit samples
	{
	    pbuf[i] = (short) ((float) pInBuf[i] * (float) 0.3);
	}
    }
    else
    {
	UCHAR* pInBuf = pInData->pData->GetBuffer();
	UCHAR* pbuf   = pModifiedData->GetBuffer();
	/* Just modify the volume... */
	for(UINT16 i = 0; i < ulLen; i++)   // divide by 2 for 16-bit samples
	{
	    pbuf[i] = (UCHAR) ((float) pInBuf[i] * (float) 0.3);
	}
    }

    /* Since we are going to modify the data, we need to set 
     * the IRMABuffer in the out data.
     */
    pOutData->pData		= pModifiedData;
    pOutData->ulAudioTime	= pInData->ulAudioTime;
    pOutData->uAudioStreamType	= pInData->uAudioStreamType;

    return PNR_OK;
}


// IUnknown COM Interface Methods

/****************************************************************************
 *  IUnknown::AddRef                                            ref:  pncom.h
 *
 */
STDMETHODIMP_(ULONG32)
CPostMixRenderer::CPostMixHook::AddRef()
{
    return InterlockedIncrement(&m_lRefCount);
}

/****************************************************************************
 *  IUnknown::Release                                           ref:  pncom.h
 *
 */
STDMETHODIMP_(ULONG32)
CPostMixRenderer::CPostMixHook::Release()
{
    if (InterlockedDecrement(&m_lRefCount) > 0)
    {
        return m_lRefCount;
    }

    delete this;
    return 0;
}

/****************************************************************************
 *  IUnknown::QueryInterface                                    ref:  pncom.h
 *
 */
STDMETHODIMP
CPostMixRenderer::CPostMixHook::QueryInterface(REFIID riid, void** ppvObj)
{
    if (IsEqualIID(riid, IID_IUnknown))
    {
        AddRef();
        *ppvObj = (IUnknown*)(IRMAAudioHook*)this;
        return PNR_OK;
    }
    else if (IsEqualIID(riid, IID_IRMAAudioHook))
    {
        AddRef();
        *ppvObj = (IRMAAudioHook*)this;
        return PNR_OK;
    }
    *ppvObj = NULL;
    return PNR_NOINTERFACE;
}

