/****************************************************************************
 *
 *  $Id: exallow.cpp,v 1.1 1998/12/03 18:44:43 darrens Exp $
 *
 *  Copyright (C) 1998 RealNetworks, Inc.
 *  All rights reserved.
 *
 *  http://www.real.com/devzone
 *
 *  This program contains proprietary information of RealNetworks, Inc.,
 *  and is licensed subject to restrictions on use and distribution.
 *
 *  exallow.cpp
 *
 *  Example Allowance plug-in for the RealMedia Architecture (RMA).
 */

/****************************************************************************
 *  This is an example of an Allowance plug-in. The purpose of an Allowance
 *  server plug-in is to monitor player connections to the server and make 
 *  decisions about what each player is allowed to do. The Allowance plug-in
 *  receives a notification for each significant event in a player's 
 *  connection lifetime, and can act to restrict the player's rights after
 *  receiving any one of these notifications. The ways in which an Allowance
 *  plug-in can restrict a player are varied, and include disconnecting,
 *  redirecting, authenticating, or granting limited permissions to the
 *  player in terms of which content can be played.
 *
 *  In order for your plug-in to achieve this functionality, it must
 *  implement the IRMAPlayerConnectionAdviseSink interface which has 
 *  methods for receiving notifications of many events in the course of
 *  each player connection. The IRMAPlugin interface must also be 
 *  implemented to handle fundamental plug-in operations.
 *
 *
 *  Example: exallow.cpp
 *
 *  This example demonstrates a simple Allowance Plug-in. It will impose the
 *  following two restrictions upon incoming player connections, to
 *  demonstrate just a few of the ways in which Allowance Plug-ins can be
 *  used to customize and extend the capabilities of the RMA server:
 *
 *  1) A maximum of 5 players will be allowed to view the 
 *     rtsp://server:port/limited.rm URL at any given time. If a player 
 *     requests this URL while 5 players are already viewing it, the 6th
 *     player will receive an alert message, asking them to try again later.
 *
 *  2) A player attempting to view the rtsp://server:port/jukebox.rm URL
 *     will be randomly redirected to one of the following URLs:
 *
 *                  - rtsp://server:port/song1.rm
 *                  - rtsp://server:port/song2.rm
 *                  - rtsp://server:port/song3.rm
 *
 *  To test this example plug-in, simply build the plug-in, place it in
 *  your RMA server plug-in directory, and place RealMedia (.rm) files in
 *  your content directory with the names mentioned above (limited.rm, 
 *  song1.rm, song2.rm, and song3.rm). Connect to one of the two special URLs
 *  with a RealPlayer to verify that the Allowance plug-in is working 
 *  properly.
 *
 *
 *  The RMA server calls the various interface routines at well defined times.
 *  The sequence of the major events is as follows:
 *
 *  Upon RMA server startup:
 *  - RMACreateInstance() creates new instance of CMyExampleRenderer class
 *  - GetPlugInInfo()     returns descriptive information about the plug-in
 *
 *  When a player first connects to the server:
 *  - OnConnection()	    called when a player connects
 *  - SetPlayerController() called to provide a player controlling interface
 *  - SetRegistryID()	    called to provide the location of player info.
 *  - OnURL()		    called when the player requests a URL
 *
 *  During the player's connection:
 *  - OnBegin()		    called when player starts or resumes
 *  - OnStop()		    called when player stops
 *  - OnPause()		    called when the playback is paused
 *  - OnDone()		    called when the player has disconnected
 *
 *
 *  Note:
 *  The routines are listed roughly in the order in which they are called.
 */


/****************************************************************************
 * Defines
 */
#define	INITGUID     /* Interface ID's */
#define	MAX_LIMITED_URL_COUNT 5
#define LIMITED_URL_ALERT "Too many users are viewing this stream. Please try again later."

/****************************************************************************
 * Includes
 */
#include <stdio.h>     /* printf, sprintf */
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "pntypes.h" 
#include "pncom.h"     /* IUnknown */
#include "rmacomm.h"   /* IRMACommonClassFactory */
#include "rmapckts.h"  /* IRMABuffer, IRMAPacket, IRMAValues */
#include "rmaplugn.h"  /* IRMAPlugin */
#include "rmaallow.h"  /* IRMAPlayerConnectionAdviseSink */
#include "rmamon.h"    /* IRMAPNRegistry */
#include "rmafiles.h"  /* IRMARequest */
#include "exallow.h"   /* CExampleAllowance */


/****************************************************************************
 *  RMACreateInstance                                        ref:  rmaplugn.h
 *
 *  This routine creates a new instance of the CExampleAllowance class.
 *  It is called when the RMA server application is launched.
 */
STDAPI
RMACreateInstance(IUnknown** ppExAllowanceObj)
{
    *ppExAllowanceObj = (IUnknown*)(IRMAPlugin*)new CExampleAllowance();
    if (*ppExAllowanceObj != NULL)
    {
	(*ppExAllowanceObj)->AddRef();
	return PNR_OK;
    }

    return PNR_OUTOFMEMORY;
}


/****************************************************************************
 *  CExampleAllowance static variables                        ref:  exallow.h
 *
 *  These variables are passed to the RMA server to provide information about
 *  this plug-in. They are required to be static in order to remain valid
 *  for the lifetime of the plug-in.
 */
const char* CExampleAllowance::zm_pDescription = MY_DESCRIPTION;
const char* CExampleAllowance::zm_pCopyright   = MY_COPYRIGHT;
const char* CExampleAllowance::zm_pMoreInfoURL = MY_MORE_INFO_URL;


/****************************************************************************
 *  CExampleAllowance::CExampleAllowance                      ref:  exallow.h
 *
 *  Constructor
 */
CExampleAllowance::CExampleAllowance(void)
    : m_RefCount	    (0),
      m_pClassFactory	    (NULL),
      m_pRegistry	    (NULL),
      m_pPCResponse	    (NULL),
      m_pPlayerController   (NULL),
      m_ulPlayerRegistryID  (0),
      m_pPlayerRegistryName (NULL),
      m_bViewingLimitedURL  (FALSE)
{
}


/****************************************************************************
 *  IRMAPlugin::GetPluginInfo                                ref:  rmaplugn.h
 *
 *  This routine returns descriptive information about the plug-in. It is 
 *  called when the RMA server application is launched.
 */
STDMETHODIMP
CExampleAllowance::GetPluginInfo
(
    REF(BOOL)        bLoadMultiple,
    REF(const char*) pDescription,
    REF(const char*) pCopyright,
    REF(const char*) pMoreInfoURL,
    REF(UINT32)      versionNumber
)
{
    bLoadMultiple = TRUE;
    pDescription  = zm_pDescription;
    pCopyright    = zm_pCopyright;
    pMoreInfoURL  = zm_pMoreInfoURL;
    versionNumber = MY_PLUGIN_VERSION;

    return PNR_OK;
}


/****************************************************************************
 *  IRMAPlugin::InitPlugin                                   ref:  rmaplugn.h
 *
 *  This routine performs initialization steps such as determining if
 *  required interfaces are available. It is called when the RMA core 
 *  application is launched, and whenever a new player Allowance object
 *  is needed.
 */
STDMETHODIMP
CExampleAllowance::InitPlugin(IUnknown* pContext)
{
    /*
     * Store a reference to the IRMACommonClassFactory interface which is
     * used to create commonly used RMA objects such as IRMAPacket, 
     * IRMAValues, and IRMABuffers.
     */
    pContext->QueryInterface(IID_IRMACommonClassFactory, 
	                    (void**)&m_pClassFactory);
    if (m_pClassFactory == NULL)
    {
	return PNR_NOTIMPL;
    }

    /*
     * Store a reference to the IRMAPNRegistry interface which is
     * used to retrieve useful information about the connected player
     */
    pContext->QueryInterface(IID_IRMAPNRegistry, 
	                    (void**)&m_pRegistry);
    if (m_pRegistry == NULL)
    {
	return PNR_NOTIMPL;
    }

    /* 
     * Note that QueryInterface() takes care of adding a reference to the
     * interface for us. You are however responsible for releasing the
     * reference to the object when you are done using it by calling
     * the Release() routine.
     */

    return PNR_OK;
}


/****************************************************************************
 *  IRMAPlayerConnectionAdviseSink::OnConnection             ref:  rmaallow.h
 *
 *  The server core calls this routine when a new player has connected
 *  to the server.
 */
STDMETHODIMP 
CExampleAllowance::OnConnection(THIS_ IRMAPlayerConnectionResponse* pResponse)
{
    m_pPCResponse = pResponse;
    m_pPCResponse->AddRef();

    m_pPCResponse->OnConnectionDone(PNR_OK);

    return PNR_OK;
}

/****************************************************************************
 *  IRMAPlayerConnectionAdviseSink::SetPlayerController      ref:  rmaallow.h
 *
 *  The server core calls this routine to supply the Allowance plug-in with
 *  an interface by which the player can be controlled. For example, this 
 *  interface can be used to disconnect the player, send them an alert 
 *  message, or redirect them to another URL.
 */
STDMETHODIMP
CExampleAllowance::SetPlayerController(THIS_ 
				       IRMAPlayerController* pPlayerController)
{
    m_pPlayerController = pPlayerController;
    m_pPlayerController->AddRef();

    return PNR_OK;
}

/****************************************************************************
 *  IRMAPlayerConnectionAdviseSink::SetRegistryID            ref:  rmaallow.h
 *
 *  The server core calls this routine to supply the Allowance plug-in with
 *  a server registry ID, which can be used to find out various pieces of
 *  information about the connected player. For example, the player's IP
 *  Address, GUID, and Protocol are all stored under this key in the server
 *  registry.
 */
STDMETHODIMP
CExampleAllowance::SetRegistryID(THIS_ UINT32 ulPlayerRegistryID)
{
    m_ulPlayerRegistryID = ulPlayerRegistryID;

    // Obtain the name of the key for obtaining subkeys later
    m_pRegistry->GetPropName(ulPlayerRegistryID, m_pPlayerRegistryName);

    return PNR_OK;
}

/****************************************************************************
 *  IRMAPlayerConnectionAdviseSink::OnURL                    ref:  rmaallow.h
 *
 *  The server core calls this routine when the connected player requests
 *  a URL. The requested URL is stored in the passed IRMARequest object,
 *  along with all 822 Headers that accompanied the request.
 */
STDMETHODIMP 
CExampleAllowance::OnURL(THIS_ IRMARequest* pRequest)
{
    PN_RESULT	hResult = PNR_OK;
    const char* pURL	= NULL;

    // Obtain the requested URL
    pRequest->GetURL(pURL);

    if (!strcmp(pURL, "limited.rm"))
    {
	// The player requested the limited.rm URL
	hResult = HandleLimitedURLRequest();
    }
    else if (!strcmp(pURL, "jukebox.rm"))
    {
	// The player requested the jukebox.rm URL
	hResult = HandleJukeboxURLRequest();
    }

    /*
     * Note: Most IRMAPlayerConnectionAdviseSink interface functions have
     * a corresponding response function in the IRMAPlayerConnectionResponse
     * interface. This function should be called when the plug-in has 
     * finished processing the notification.
     *
     * The connected player will not be permitted to begin playback until 
     * all allowance plugins have called OnURLDone() with a Status of PNR_OK.
     */
    m_pPCResponse->OnURLDone(hResult);

    return PNR_OK;
}

/****************************************************************************
 *  IRMAPlayerConnectionAdviseSink::OnBegin                  ref:  rmaallow.h
 *
 *  The server core calls this routine when the player begins or resumes
 *  playback.
 */
STDMETHODIMP 
CExampleAllowance::OnBegin(void)
{
    m_pPCResponse->OnBeginDone(PNR_OK);

    return PNR_OK;
}

/****************************************************************************
 *  IRMAPlayerConnectionAdviseSink::OnPause                  ref:  rmaallow.h
 *
 *  The server core calls this routine when the player pauses playback.
 */
STDMETHODIMP 
CExampleAllowance::OnPause(void)
{
    m_pPCResponse->OnPauseDone(PNR_OK);

    return PNR_OK;
}

/****************************************************************************
 *  IRMAPlayerConnectionAdviseSink::OnStop                   ref:  rmaallow.h
 *
 *  The server core calls this routine when the player stops playback.
 */
STDMETHODIMP 
CExampleAllowance::OnStop(void)
{
    if (m_bViewingLimitedURL)
    {
	// This player has finished viewing the "limited.rm" URL
	DecrementLimitedURLCount();
	m_bViewingLimitedURL = FALSE;
    }

    m_pPCResponse->OnStopDone(PNR_OK);

    return PNR_OK;
}

/****************************************************************************
 *  IRMAPlayerConnectionAdviseSink::OnDone                   ref:  rmaallow.h
 *
 *  The server core calls this routine when the player has disconnected
 *  from the server.
 *  
 */
STDMETHODIMP 
CExampleAllowance::OnDone(void)
{
    return PNR_OK;
}


/****************************************************************************
 *  CExampleAllowance::~CExampleAllowance                    ref:   exallow.h
 *
 *  Destructor
 */
CExampleAllowance::~CExampleAllowance(void)
{
    if (m_pClassFactory)
    {
	m_pClassFactory->Release();
	m_pClassFactory = NULL;
    }

    if (m_pRegistry)
    {
	m_pRegistry->Release();
	m_pRegistry = NULL;
    }

    if (m_pPCResponse)
    {
	m_pPCResponse->Release();
	m_pPCResponse = NULL;
    }

    if (m_pPlayerController)
    {
	m_pPlayerController->Release();
	m_pPlayerController = NULL;
    }

    if (m_pPlayerRegistryName)
    {
	m_pPlayerRegistryName->Release();
	m_pPlayerRegistryName = NULL;
    }
}


// IUnknown COM Interface Methods

/****************************************************************************
 *  IUnknown::AddRef                                            ref:  pncom.h
 *
 *  This routine increases the object reference count in a thread safe
 *  manner. The reference count is used to manage the lifetime of an object.
 *  This method must be explicitly called by the user whenever a new
 *  reference to an object is used.
 */
STDMETHODIMP_(UINT32)
CExampleAllowance::AddRef(void)
{
    return InterlockedIncrement(&m_RefCount);
}


/****************************************************************************
 *  IUnknown::Release                                           ref:  pncom.h
 *
 *  This routine decreases the object reference count in a thread safe
 *  manner, and deletes the object if no more references to it exist. It must
 *  be called explicitly by the user whenever an object is no longer needed.
 */
STDMETHODIMP_(UINT32)
CExampleAllowance::Release(void)
{
    if (InterlockedDecrement(&m_RefCount) > 0)
    {
	return m_RefCount;
    }

    delete this;
    return 0;
}


/****************************************************************************
 *  IUnknown::QueryInterface                                    ref:  pncom.h
 *
 *  This routine indicates which interfaces this object supports. If a given
 *  interface is supported, the object's reference count is incremented, and
 *  a reference to that interface is returned. Otherwise a NULL object and
 *  error code are returned. This method is called by other objects to
 *  discover the functionality of this object.
 */
STDMETHODIMP
CExampleAllowance::QueryInterface
(
    REFIID interfaceID,
    void** ppInterfaceObj
)
{
    // By definition all COM objects support the IUnknown interface
    if (IsEqualIID(interfaceID, IID_IUnknown))
    {
	AddRef();
	*ppInterfaceObj = (IUnknown*)(IRMAPlugin*)this;
	return PNR_OK;
    }

    // IRMAPlugin interface is supported
    else if (IsEqualIID(interfaceID, IID_IRMAPlugin))
    {
	AddRef();
	*ppInterfaceObj = (IRMAPlugin*)this;
	return PNR_OK;
    }

    // IRMAPlayerConnectionAdviseSink interface is supported
    else if (IsEqualIID(interfaceID, IID_IRMAPlayerConnectionAdviseSink))
    {
	AddRef();
	*ppInterfaceObj = (IRMAPlayerConnectionAdviseSink*)this;
	return PNR_OK;
    }

    // No other interfaces are supported
    *ppInterfaceObj = NULL;
    return PNR_NOINTERFACE;
}

PN_RESULT
CExampleAllowance::HandleLimitedURLRequest()
{
    PN_RESULT	hResult	     = PNR_OK;
    INT32	nClientCount = 0;
    IRMABuffer* pBuffer	     = NULL;

    // Find out how many players are currently watching this URL
    if (PNR_OK != m_pRegistry->GetIntByName("server.LimitedURLCount", 
					    nClientCount))
    {
	// The registry item does not exist yet, so create it
	nClientCount = 0;
	m_pRegistry->AddInt("server.LimitedURLCount", nClientCount);
    }

    // Test the limit, to determine if it has reached capacity
    if (nClientCount >= MAX_LIMITED_URL_COUNT)
    {
	// We have reached capacity, so send them an alert message
	m_pClassFactory->CreateInstance(CLSID_IRMABuffer, (void**)&pBuffer);
	pBuffer->Set((const UCHAR*)LIMITED_URL_ALERT, 
	    strlen(LIMITED_URL_ALERT) + 1);
	m_pPlayerController->AlertAndDisconnect(pBuffer);
	PN_RELEASE(pBuffer);

	hResult = PNR_FAIL;
    }
    else
    {
	// We have not reached capacity, so let them have access
	m_bViewingLimitedURL = TRUE;

	// Increment the count and update its registry value
	nClientCount++;
	m_pRegistry->SetIntByName("server.LimitedURLCount", nClientCount);
    }

    return hResult;
}

PN_RESULT
CExampleAllowance::HandleJukeboxURLRequest()
{
    UINT32	ulRand  = 0;
    char	pNewURL[256];
    IRMABuffer* pBuffer = NULL;

    // Generate a random number between 0 and 2
    srand((unsigned)time(NULL));
    ulRand = rand() % 3;

    // Choose a redirect URL based on the result
    switch(ulRand)
    {
	case 0:
	{
	    sprintf(pNewURL, "song1.rm");
	}
	break;

	case 1:
	{
	    sprintf(pNewURL, "song2.rm");
	}
	break;

	case 2:
	default:
	{
	    sprintf(pNewURL, "song3.rm");
	}
	break;
    }

    // Redirect the player
    m_pClassFactory->CreateInstance(CLSID_IRMABuffer, (void**)&pBuffer);
    pBuffer->Set((const UCHAR*)pNewURL, strlen(pNewURL) + 1);
    m_pPlayerController->Redirect(pBuffer);
    PN_RELEASE(pBuffer);

    return PNR_OK;
}

void
CExampleAllowance::DecrementLimitedURLCount(void)
{
    INT32 nClientCount = 0;

    // Obtain the count of players currently watching the "limited.rm" URL
    if (PNR_OK == m_pRegistry->GetIntByName("server.LimitedURLCount", 
					    nClientCount))
    {
	// Decrement the count and update its registry value
	if (nClientCount > 0)
	{
	    nClientCount--;
	}
	m_pRegistry->SetIntByName("server.LimitedURLCount", nClientCount);
    }
}
