/****************************************************************************
 *
 *  $Id: ff1rendr.cpp,v 1.21 1997/11/21 19:16:52 jfarr Exp $
 *
 *  Copyright (C) 1995,1996,1997 RealNetworks, Inc.
 *  All rights reserved.
 *
 *  http://www.real.com/devzone
 *
 *  This program contains proprietary information of RealNetworks, Inc.,
 *  and is licensed subject to restrictions on use and distribution.
 *
 *  ff1rendr.cpp
 *
 *  Example Renderer plug-in for the RealMedia Architecture (RMA).
 */

/****************************************************************************
 *  This is an example of a Renderer plug-in. The purpose of a Renderer is to
 *  "render" a stream of data in some human understandable form such as
 *  audio, video, text, etc. The Renderer receives data in the form of a
 *  stream of packets, which are just small chunks of data of a known type.
 *  Knowing the format of the packets of a particular stream type, the
 *  Renderer can reassemble these packets and "display" them on the computer.
 *
 *  In order for your plug-in to achieve this functionality, it must
 *  implement the IRMARenderer interface which has routines for handling
 *  rendering events and operations. The IRMAPlugin interface must also be
 *  implemented to handle fundamental plug-in operations.
 *
 *
 *  Example: ff1rendr.cpp
 *
 *  This example renders a very simple file-format which was created in the
 *  File Format 1 plug-in example. That example streamed a text file of the
 *  following structure:
 *
 *        *HEADER*STREAM_HEADER*NO_PACKETS=002*PACKET_001*PACKET_002
 *
 *  The format consists of one file header (*HEADER) , one stream header
 *  (*STREAM_HEADER*NO_PACKETS=002), and one stream (*PACKET_001*PACKET_002),
 *  of data consisting of less than 1000 packets. The number of packets in a
 *  given file is given as part of the stream header (*NO_PACKETS=002). Each
 *  packet simply consists of a text label (*PACKET_###) of fixed size. This
 *  plug-in renders these packets by just printing out the packet label to
 *  the console once the packet has arrived.
 *
 *
 *  The RMA core calls the various interface routines at well defined times.
 *  The sequence of the major events is as follows:
 *
 *  Upon RMA core startup:
 *  - RMACreateInstance() creates new instance of CMyExampleRenderer class
 *  - GetPlugInInfo()     returns descriptive information about the plug-in
 *  - GetRendererInfo()   returns crucial info linking plug-in and stream
 *
 *  When a stream associated with this plug-in needs to be rendered:
 *  - InitPlugIn()        initializes plug-in
 *  - StartStream()       called to provide access to stream and player
 *  - OnHeader()          called when the stream header arrives
 *  - OnBegin()           called when player starts or resumes
 *  - GetDisplayType()    returns the preferred display type of Renderer
 *  - OnPacket()          called when a packet arrives
 *  - OnTimeSynch()       called periodically to synchronize rendering
 *  - OnBuffering()       called when Renderer is buffering data
 *  - OnEndofPackets()    called when Renderer has received all of its packets
 *  - EndStream()         called when rendering of stream has completed
 *
 *  During user interaction:
 *  - OnPreSeek()         called before a seek is about to occur
 *  - OnPostSeek()        called just after a seek has occured
 *  - OnPause()           called when the playback is paused
 *
 *
 *  Note:
 *  The routines are listed roughly in the order in which they are called.
 */


/****************************************************************************
 * Defines
 */
#define   INITGUID     /* Interface ID's */


/****************************************************************************
 * Includes
 */
#include <stdio.h>     /* printf, sprintf */
#include "pntypes.h" 
#include "pncom.h"     /* IUnknown */
#include "rmacomm.h"   /* IRMACommonClassFactory */
#include "rmapckts.h"  /* IRMABuffer, IRMAPacket, IRMAValues */
#include "rmaplugn.h"  /* IRMAPlugin */
#include "rmarendr.h"  /* IRMARenderer */
#include "ff1rendr.h"  /* CMyExampleRenderer */


/****************************************************************************
 *  RMACreateInstance                                        ref:  rmaplugn.h
 *
 *  This routine creates a new instance of the CMyExampleRenderer class.
 *  It is called when the RMA core application is launched, and whenever
 *  a stream associated with this plug-in needs to be rendered.
 */
STDAPI
RMACreateInstance(IUnknown** ppExRendererObj)
{
	*ppExRendererObj = (IUnknown*)(IRMAPlugin*)new CMyExampleRenderer();
	if (*ppExRendererObj != NULL)
	{
		(*ppExRendererObj)->AddRef();
		return PNR_OK;
	}

	return PNR_OUTOFMEMORY;
}


/****************************************************************************
 *  CMyExampleRenderer static variables                      ref:  ff1rendr.h
 *
 *  These variables are passed to the RMA core to provide information about
 *  this plug-in. They are required to be static in order to remain valid
 *  for the lifetime of the plug-in.
 */
const char* CMyExampleRenderer::zm_pDescription = MY_DESCRIPTION;
const char* CMyExampleRenderer::zm_pCopyright   = MY_COPYRIGHT;
const char* CMyExampleRenderer::zm_pMoreInfoURL = MY_MORE_INFO_URL;
const char* CMyExampleRenderer::zm_pStreamMimeTypes[] = MY_STREAM_MIME_TYPES;


/****************************************************************************
 *  CMyExampleRenderer::CMyExampleRenderer                   ref:  ff1rendr.h
 *
 *  Constructor
 */
CMyExampleRenderer::CMyExampleRenderer(void)
	: m_RefCount      (0),
	  m_pClassFactory (NULL),
	  m_bInSeekMode   (FALSE)
{
}


/****************************************************************************
 *  IRMAPlugin::GetPluginInfo                                ref:  rmaplugn.h
 *
 *  This routine returns descriptive information about the plug-in, most
 *  of which is used in the About box of the user interface. It is called
 *  when the RMA core application is launched.
 */
STDMETHODIMP
CMyExampleRenderer::GetPluginInfo
(
	REF(BOOL)        bLoadMultiple,
	REF(const char*) pDescription,
	REF(const char*) pCopyright,
	REF(const char*) pMoreInfoURL,
	REF(UINT32)      versionNumber
)
{
	bLoadMultiple = TRUE;
	pDescription  = zm_pDescription;
	pCopyright    = zm_pCopyright;
	pMoreInfoURL  = zm_pMoreInfoURL;
	versionNumber = MY_PLUGIN_VERSION;

	return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::GetRendererInfo                            ref:  rmarendr.h
 *
 *  This routine returns crucial information required to associate this
 *  plug-in with a given stream MIME type. This information tells the RMA
 *  core which Renderer to use to display a particular type of stream. This
 *  method is called when the RMA core application is launched.
 */
STDMETHODIMP
CMyExampleRenderer::GetRendererInfo
(
	REF(const char**) pStreamMimeTypes,
	REF(UINT32)       initialGranularity
)
{
	// Associate this Renderer with a given stream MIME type
	pStreamMimeTypes = zm_pStreamMimeTypes;

	// Set frequency of OnTimeSync() updates
	initialGranularity = MY_TIME_SYNC_FREQ;

	return PNR_OK;
}


/****************************************************************************
 *  IRMAPlugin::InitPlugin                                   ref:  rmaplugn.h
 *
 *  This routine performs initialization steps such as determining if
 *  required interfaces are available. It is called when the RMA core 
 *  application is launched, and whenever a stream associated with this
 *  plug-in needs to be rendered.
 */
STDMETHODIMP
CMyExampleRenderer::InitPlugin(IUnknown* pRMACore)
{
	/*
	 * Store a reference to the IRMACommonClassFactory interface which is
	 * used to create commonly used RMA objects such as IRMAPacket, 
	 * IRMAValues, and IRMABuffers.
	 */
	pRMACore->QueryInterface(IID_IRMACommonClassFactory, 
	                        (void**)&m_pClassFactory);
	if (m_pClassFactory == NULL)
	{
		return PNR_NOTIMPL;
	}

	/* 
	 * Note that QueryInterface() takes care of adding a reference to the
	 * interface for us. You are however responsible for releasing the
	 * reference to the object when you are done using it by calling
	 * the Release() routine.
	 */

	return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::StartStream                                ref:  rmarendr.h
 *
 *  The RMA core calls this routine to provide access to the stream and
 *  player. It is called when the plug-in is being initialized.
 */
STDMETHODIMP
CMyExampleRenderer::StartStream
(
	IRMAStream* /* pStream */,
	IRMAPlayer* /* pPlayer */
)
{
	// This example does not use either the stream or player objects
	return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnHeader                                   ref:  rmarendr.h
 *
 *  This routine is passed the stream header object created by the associated
 *  file format plug-in. Use the methods defined in the IRMAValues interface
 *  to access the stream header data. This method is called from the RMA
 *  core when the plug-in is being initialized.
 */
STDMETHODIMP
CMyExampleRenderer::OnHeader(IRMAValues* pStreamHeaderObj)
{
	IRMABuffer* pOpaqueData = NULL;
	pStreamHeaderObj->GetPropertyBuffer("OpaqueData", pOpaqueData);
	if (pOpaqueData != NULL)
	{
		/*
		 * Do something with the stream's opaque data here
		 */

		// Release the object once you are done with it
		pOpaqueData->Release();
	}

	return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnBegin                                   ref:  rmarendr.h
 *
 *  This routine is called by the RMA core to inform the Renderer that 
 *  playback has just begun or has been resumed. The stream's time value just
 *  after resuming the playback is provided.
 */
STDMETHODIMP
CMyExampleRenderer::OnBegin(UINT32 /* timeAfterBegin */)
{
	return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::GetDisplayType                             ref:  rmarendr.h
 *
 *  This routine returns the preferred display type of the Renderer. Any
 *  other additional information necessary for the display can also be
 *  provided. This method is called by the RMA core when the plug-in is being
 *  initialized.
 */
STDMETHODIMP
CMyExampleRenderer::GetDisplayType
(
	REF(RMA_DISPLAY_TYPE) displayType,
	REF(IRMABuffer*)   /* pDisplayInfo */
)
{
	displayType = RMA_DISPLAY_NONE;

	return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnPacket                                   ref:  rmarendr.h
 *
 *  This routine is passed the packet object streamed by the associated
 *  file format plug-in. It is called by the RMA core when a packet is due
 *  to be delivered. In most cases, the actual rendering of the packet is
 *  done in this method. The stream's time offset with respect to the master
 *  timeline is provided.
 */
STDMETHODIMP
CMyExampleRenderer::OnPacket
(
	IRMAPacket* pPacketObj,
	INT32    /* streamOffsetTime */
)
{
	/* Ignore pre-seek packets. Some renderers may want to use them though */
	if (m_bInSeekMode)
	{
	    return PNR_OK;
	}

	// Retrieve the data stored in the packet
	IRMABuffer* pPacketBuff = pPacketObj->GetBuffer();
	if (pPacketBuff != NULL)
	{
		UCHAR* pPacketData = pPacketBuff->GetBuffer();
		if (pPacketData != NULL)
		{
			// Convert packet data to C string
			char packetStr[MY_PACKET_LENGTH + 1];// Add 1 for null terminator
			::sprintf(packetStr, "%.*s", MY_PACKET_LENGTH, pPacketData);
			
			// Render the packet
			#ifndef _MACINTOSH
				#ifndef _WIN16
					::printf("%s Received!\n", packetStr);
				#endif
			#else
				// Make sure console window doesn't interfere with player
				GrafPtr savedPort;
				::GetPort(&savedPort);
				::printf("%s Received!\n", packetStr);
				::SetPort(savedPort);
			#endif		
		}
		// Release the object once you are done with it
		pPacketBuff->Release();
	}

	return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnTimeSync                                 ref:  rmarendr.h
 *
 *  This routine is called by the RMA core periodically passing the current
 *  playback time. The Renderer should use this time value to synchronize the
 *  playback of various streams.
 */
STDMETHODIMP
CMyExampleRenderer::OnTimeSync(UINT32 /* currentPlayBackTime */)
{
	return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnPreSeek                                  ref:  rmarendr.h
 *
 *  This routine is called by the RMA core before a seek is about to occur.
 *  The stream's time value just before the seek, and the time value just
 *  after the seek are provided.
 */
STDMETHODIMP
CMyExampleRenderer::OnPreSeek
(
	UINT32 /* timeBeforeSeek */,
	UINT32 /* timeAfterSeek */
)
{
	m_bInSeekMode   = TRUE;
	return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnPostSeek                                 ref:  rmarendr.h
 *
 *  This routine is called by the RMA core just after a seek has occured.
 *  The stream's time value just before the seek, and the time value just
 *  after the seek are provided.
 */
STDMETHODIMP
CMyExampleRenderer::OnPostSeek
(
	UINT32 /* timeBeforeSeek */,
	UINT32 /* timeAfterSeek */
)
{
	m_bInSeekMode   = FALSE;
	return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnPause                                    ref:  rmarendr.h
 *
 *  This routine is called by the RMA core just after a pause has occured.
 *  The stream's time value just before pausing is provided.
 */
STDMETHODIMP
CMyExampleRenderer::OnPause(UINT32 /* timeBeforePause */)
{
	return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnBuffering                                ref:  rmarendr.h
 *
 *  This routine is called by the RMA core to inform the Renderer that
 *  buffering of data is occuring. The reason for buffering (e.g. start-up
 *  of stream, seek has occured, network congestion, etc.), as well as the
 *  percentage of the buffering process complete are provided.
 */
STDMETHODIMP
CMyExampleRenderer::OnBuffering
(
	UINT32 /* reason */,
	UINT16 /* percentComplete */
)
{
	return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::OnEndofPackets                             ref:  rmarendr.h
 *
 *  Called by client engine to inform the renderer that all the
 *  packets have been delivered. However, if the user seeks before
 *  EndStream() is called, renderer may start getting packets again
 *  and the client engine will eventually call this function again.
 */
STDMETHODIMP
CMyExampleRenderer::OnEndofPackets(void)
{
	return PNR_OK;
}


/****************************************************************************
 *  IRMARenderer::EndStream                                  ref:  rmarendr.h
 *
 *  This routine is called by the RMA core when the rendering of the stream
 *  has completed. Deallocation of any resources should be done here.
 */
STDMETHODIMP
CMyExampleRenderer::EndStream(void)
{
	if (m_pClassFactory != NULL)
	{
		m_pClassFactory->Release();
		m_pClassFactory = 0;
	}

	return PNR_OK;
}


/****************************************************************************
 *  CMyExampleRenderer::~CMyExampleRenderer                  ref:  ff1rendr.h
 *
 *  Destructor
 */
CMyExampleRenderer::~CMyExampleRenderer(void)
{
	EndStream();
}


// IUnknown COM Interface Methods

/****************************************************************************
 *  IUnknown::AddRef                                            ref:  pncom.h
 *
 *  This routine increases the object reference count in a thread safe
 *  manner. The reference count is used to manage the lifetime of an object.
 *  This method must be explicitly called by the user whenever a new
 *  reference to an object is used.
 */
STDMETHODIMP_(UINT32)
CMyExampleRenderer::AddRef(void)
{
	return InterlockedIncrement(&m_RefCount);
}


/****************************************************************************
 *  IUnknown::Release                                           ref:  pncom.h
 *
 *  This routine decreases the object reference count in a thread safe
 *  manner, and deletes the object if no more references to it exist. It must
 *  be called explicitly by the user whenever an object is no longer needed.
 */
STDMETHODIMP_(UINT32)
CMyExampleRenderer::Release(void)
{
	if (InterlockedDecrement(&m_RefCount) > 0)
	{
		return m_RefCount;
	}

	delete this;
	return 0;
}


/****************************************************************************
 *  IUnknown::QueryInterface                                    ref:  pncom.h
 *
 *  This routine indicates which interfaces this object supports. If a given
 *  interface is supported, the object's reference count is incremented, and
 *  a reference to that interface is returned. Otherwise a NULL object and
 *  error code are returned. This method is called by other objects to
 *  discover the functionality of this object.
 */
STDMETHODIMP
CMyExampleRenderer::QueryInterface
(
	REFIID interfaceID,
	void** ppInterfaceObj
)
{
	// By definition all COM objects support the IUnknown interface
	if (IsEqualIID(interfaceID, IID_IUnknown))
	{
		AddRef();
		*ppInterfaceObj = (IUnknown*)(IRMAPlugin*)this;
		return PNR_OK;
	}

	// IRMAPlugin interface is supported
	else if (IsEqualIID(interfaceID, IID_IRMAPlugin))
	{
		AddRef();
		*ppInterfaceObj = (IRMAPlugin*)this;
		return PNR_OK;
	}

	// IRMARenderer interface is supported
	else if (IsEqualIID(interfaceID, IID_IRMARenderer))
	{
		AddRef();
		*ppInterfaceObj = (IRMARenderer*)this;
		return PNR_OK;
	}

	// No other interfaces are supported
	*ppInterfaceObj = NULL;
	return PNR_NOINTERFACE;
}
