/****************************************************************************
 *
 *  $Id: filefmt1.cpp,v 1.23 1999/03/30 17:57:16 darrens Exp $
 *
 *  Copyright (C) 1995,1996,1997 RealNetworks, Inc.
 *  All rights reserved.
 *
 *  http://www.real.com/devzone
 *
 *  This program contains proprietary information of RealNetworks, Inc.,
 *  and is licensed subject to restrictions on use and distribution.
 *
 *  filefmt1.cpp
 *
 *  Example File Format plug-in for the RealMedia Architecture (RMA).
 */

/****************************************************************************
 *
 *  This is an example of a File Format plug-in. In general a File Format
 *  plug-in converts your native file into streams of packets, which
 *  flow through the RealMedia platform. Eventually, these streams are
 *  delivered to the renderer, a separate plug-in responsible for 
 *  reassembling and displaying these packets in some human understandable
 *  form (audio, video, text, etc.).
 *
 *  A file format generally consists of a file header, one or more stream
 *  headers, and one or more streams of data which are to be rendered. For
 *  example, a single file could contain two streams of data- one stream of
 *  audio data, and one stream of video data. Each stream "type" has a
 *  corresponding renderer plug-in which knows how to "display" that stream's
 *  data. The header data usually contains information used to initialize the
 *  file or stream. The stream data should be broken up into a series of
 *  small chunks of data, called packets, in order to be easily transported
 *  by the RealMedia platform.
 *
 *  In order for your plug-in to achieve this functionality, it must
 *  implement the IRMAFileFormatObject interface which has routines
 *  specifically geared toward manipulating such headers and packets.
 *  Additionally, since the RMA file system is asynchronous in nature, the
 *  IRMAFileResponse interface must be implemented in order to receive
 *  notification when a file operation (read, write, seek, etc.) has
 *  completed. As is the case for all plug-ins, the IRMAPlugin interface must
 *  also be implemented to handle fundamental plug-in operations.
 *
 *
 *  Example: filefmt1.cpp
 *
 *  This is an example of a File Format plug-in which converts a very simple
 *  text file into a series of packets. The format consists of one file
 *  header (*HEADER), one stream header (*STREAM_HEADER*NO_PACKETS=002), and
 *  one stream of data (*PACKET_001*PACKET_002) consisting of less than
 *  1000 packets. The number of packets in a given file is given as part
 *  of the stream header (*NO_PACKETS=002). Here is an example of a file
 *  consisting of two packets:
 *
 *        *HEADER*STREAM_HEADER*NO_PACKETS=002*PACKET_001*PACKET_002
 *
 *  This simple file format was chosen for illustrative purposes, the same
 *  concepts apply for arbitrarily complex file formats.
 *
 *
 *  The RMA core calls the various interface routines at well defined times.
 *  The sequence of the major events is as follows:
 *
 *  Upon RMA core startup:
 *  - RMACreateInstance()  creates new instance of CMyExampleFileFormat class
 *  - GetPlugInInfo()      returns descriptive information about the plug-in
 *  - GetFileFormatInfo()  returns crucial info linking plug-in and file type
 *
 *  When an URL is requested:
 *  - InitPlugIn()         initializes plug-in
 *  - InitFileFormat()     initializes objects used by File Format interface
 *  - InitDone()           called when file initialization is done
 *  - GetFileHeader()      reads file header from file
 *  - FileHeaderReady()    passes file header object to RMA core
 *
 *    For each stream in the file format:
 *  - GetStreamHeader()    reads stream header from file
 *  - StreamHeaderReady()  passes stream header object to RMA core
 *
 *  While playing your RealMedia data type:
 *  - GetPacket()          called each time the RMA core requests a packet
 *  - PacketReady()        passes packet object to RMA core
 *  - Seek()               called to move to a different part of the timeline
 *
 *
 *  Note:
 *  The routines are listed roughly in the order in which they are called.
 */


/****************************************************************************
 * Defines
 */
#define   INITGUID     /* Interface ID's */


/****************************************************************************
 * Includes
 */
#include <string.h>    /* strrchr */
#include <stdlib.h>    /* atol */
#include <stdio.h>     /* sprintf */

#include "pntypes.h" 
#include "pncom.h"     /* IUnknown */
#include "rmacomm.h"   /* IRMACommonClassFactory */
#include "rmapckts.h"  /* IRMABuffer, IRMAPacket, IRMAValues */
#include "rmaplugn.h"  /* IRMAPlugin */
#include "rmafiles.h"  /* IRMAFileResponse */
#include "rmaformt.h"  /* IRMAFileFormatObject */

#include "filefmt1.h"  /* CMyExampleFileFormat */



/****************************************************************************
 *  RMACreateInstance                                        ref:  rmaplugn.h
 *
 *  This routine creates a new instance of the CMyExampleFileFormat class.
 *  It is called when the RMA core application is launched, and whenever
 *  an URL associated with this plug-in is opened.
 */
STDAPI
RMACreateInstance(IUnknown** ppExFileFormatObj)
{
	*ppExFileFormatObj = (IUnknown*)(IRMAPlugin*)new CMyExampleFileFormat();
	if (*ppExFileFormatObj != NULL)
	{
		(*ppExFileFormatObj)->AddRef();
		return PNR_OK;
	}

	return PNR_OUTOFMEMORY;
}


/****************************************************************************
 *  CMyExampleFileFormat static variables                    ref:  filefmt1.h
 *
 *  These variables are passed to the RMA core to provide information about
 *  this plug-in. They are required to be static in order to remain valid
 *  for the lifetime of the plug-in.
 */
const char* CMyExampleFileFormat::zm_pDescription = MY_DESCRIPTION;
const char* CMyExampleFileFormat::zm_pCopyright   = MY_COPYRIGHT;
const char* CMyExampleFileFormat::zm_pMoreInfoURL = MY_MORE_INFO_URL;
const char* CMyExampleFileFormat::zm_pFileMimeTypes[]  = MY_FILE_MIME_TYPES;
const char* CMyExampleFileFormat::zm_pFileExtensions[] = MY_FILE_EXTENSIONS;
const char* CMyExampleFileFormat::zm_pFileOpenNames[]  = MY_FILE_OPEN_NAMES;


/****************************************************************************
 *  CMyExampleFileFormat::CMyExampleFileFormat               ref:  filefmt1.h
 *
 *  Constructor
 */
CMyExampleFileFormat::CMyExampleFileFormat(void)
	: m_RefCount      (0),
	  m_pClassFactory (NULL),
      m_pFileObj      (NULL),
	  m_pStatus       (NULL),
	  m_State         (Ready),
	  m_NextPacketDeliveryTime(0)
{
}


/****************************************************************************
 *  IRMAPlugin::GetPluginInfo                                ref:  rmaplugn.h
 *
 *  This routine returns descriptive information about the plug-in, most
 *  of which is used in the About box of the user interface. It is called
 *  when the RMA core application is launched.
 */
STDMETHODIMP
CMyExampleFileFormat::GetPluginInfo
(
	REF(BOOL)        bLoadMultiple,
	REF(const char*) pDescription,
	REF(const char*) pCopyright,
	REF(const char*) pMoreInfoURL,
    REF(UINT32)      versionNumber
)
{
	// File Format plug-ins MUST be able to be multiply instantiated
	bLoadMultiple = TRUE;

	pDescription  = zm_pDescription;
	pCopyright	  = zm_pCopyright;
	pMoreInfoURL  = zm_pMoreInfoURL;
	versionNumber = MY_PLUGIN_VERSION;

	return PNR_OK;
}


/****************************************************************************
 *  IRMAFileFormatObject::GetFileFormatInfo                  ref:  rmaformt.h
 *
 *  This routine returns crucial information required to associate this
 *  plug-in with a given MIME type. This information tells the core which
 *  File Format plug-in to use for a particular URL. The routine is called
 *  when the RMA core application is launched.
 */
STDMETHODIMP
CMyExampleFileFormat::GetFileFormatInfo
(
	REF(const char**) pFileMimeTypes,
	REF(const char**) pFileExtensions,
	REF(const char**) pFileOpenNames
)
{
	pFileMimeTypes  = zm_pFileMimeTypes;
	pFileExtensions = zm_pFileExtensions;
	pFileOpenNames  = zm_pFileOpenNames;

	return PNR_OK;
}


/****************************************************************************
 *  IRMAPlugin::InitPlugin                                   ref:  rmaplugn.h
 *
 *  This routine performs initialization steps such as determining whether
 *  required interfaces are available. It is called when the RMA core 
 *  application is launched, and whenever an URL associated with this
 *  plug-in is opened.
 */
STDMETHODIMP
CMyExampleFileFormat::InitPlugin(IUnknown* pRMACore)
{
	/*
	 * Store a reference to the IRMACommonClassFactory interface which is
	 * used to create commonly used RMA objects such as IRMAPacket, 
	 * IRMAValues, and IRMABuffers.
	 */
	pRMACore->QueryInterface(IID_IRMACommonClassFactory, 
                              (void**)&m_pClassFactory);
	if (m_pClassFactory == NULL)
	{
		return PNR_NOTIMPL;
	}

	/* 
	 * Note that QueryInterface() takes care of adding a reference to the
	 * interface for us. You are however responsible for releasing the
	 * reference to the interface when you are done using it by calling
	 * the Release() routine.
	 */

	return PNR_OK;
}


/****************************************************************************
 *  IRMAFileFormatObject::InitFileFormat                     ref:  rmaformt.h
 *
 *  This routine initializes the file, and stores references to objects used
 *  throughout the example. It is called whenever an URL associated with this
 *  plug-in is opened.
 */
STDMETHODIMP
CMyExampleFileFormat::InitFileFormat
(
	IRMARequest*     /* pRequest */,
	IRMAFormatResponse* pFormatResponse,
	IRMAFileObject*     pFileObject
)
{
	// The format response object is used to notify RMA core of our status
	m_pStatus = pFormatResponse;
	if (m_pStatus != NULL)
	{
		m_pStatus->AddRef();
	}

	// The file object is used to handle file I/O
	m_pFileObj = pFileObject;
	if (m_pFileObj != NULL)
	{
		m_pFileObj->AddRef();
		
		/*
		 * Initialize and check validity of file. Also, associate the file
		 * with a response object which is notified when the file operation is
		 * complete.
		 */
		m_State = InitPending;
		m_pFileObj->Init(PN_FILE_READ | PN_FILE_BINARY, this);	// asynchronous

		/* 
		 * Since this class was designated at the response object, this
		 * class's IRMAFileResponse::InitDone method will be called when
		 * the initialization of the file is complete. (See InitDone).
		 */
	}

	return PNR_OK;
}


/****************************************************************************
 *  IRMAFileResponse::InitDone                               ref:  rmafiles.h
 *
 *  This routine notifies the RMA core that the initialization of the file is
 *  done. It is called automatically when the initialization of the file
 *  has completed.
 */
STDMETHODIMP
CMyExampleFileFormat::InitDone(PN_RESULT status)
{
	if (m_State != InitPending)
	{
		return PNR_UNEXPECTED;
	}

	// Notify RMA core that intialization started in InitFileFormat is done
	m_State = Ready;
	m_pStatus->InitDone(status);

	return PNR_OK;
}


/****************************************************************************
 *  IRMAFileFormatObject::GetFileHeader                      ref:  rmaformt.h
 *
 *  This routine returns to the RMA core an object containing the file
 *  header information. Several routines are actually required to complete
 *  the process due to the asynchronous nature of the RMA file system. This
 *  method is called by the RMA core after the file has been initialized.
 */
STDMETHODIMP
CMyExampleFileFormat::GetFileHeader(void)
{
	/* The asynchronous nature of the RMA file system somewhat obscures the
	 * sequence of events. The basic procedure is as follows:
	 *
	 *  1. Move to the beginning of the header data in the file        (Seek)
	 *  2. Wait until move done                                    (SeekDone)
	 *  3. Read the file header data from the file                     (Read)
	 *  4. Wait until read done                                    (ReadDone)
	 *  5. Create new object containing header data   (MyCreateFileHeaderObj)
	 *  6. Notify RMA core when header object is ready(MyCreateFileHeaderObj)
	 */

	if (m_State != Ready)
	{
		return PNR_UNEXPECTED;
	}

	/*
	 * Since Seek() is asynchronous we need to note our state in order to
	 * correctly respond in SeekDone() when finished.
	 */
	m_State = GetFileHeaderSeekPending;

	// Step 1: Move to the beginning of the header data in the file
	const BOOL bIsRelative = FALSE;
	m_pFileObj->Seek(MY_FILE_HEADER_START, bIsRelative);

	// Step 2: Wait until move is done

	// See SeekDone() for next "step" of GetFileHeader() procedure

	return PNR_OK;
}


/****************************************************************************
 *  CMyExampleFileFormat::MyCreateFileHeaderObj              ref:  filefmt1.h
 *
 *  This routine creates a "file header" object and passes it off to the
 *  RMA core, which in turn transports it to the renderer. This object must
 *  contain a property stating the number of streams contained in this file
 *  format. Any additional header information read from the file can also be
 *  placed in this object. This method is called after the header data from
 *  the file has been completely read.
 */
void
CMyExampleFileFormat::MyCreateFileHeaderObj
(
	PN_RESULT   status,
	IRMABuffer* pHeaderDataReadFromFile
)
{	/*
	 * We are in the process of handling the GetFileHeader() request...
	 * See GetFileHeader(), SeekDone(), and ReadDone() for previous steps.
	 */
	m_State = Ready;

	// Step 5: Create new object containing the header data
	IRMAValues* pHeaderObj = NULL;
	m_pClassFactory->CreateInstance(CLSID_IRMAValues, (void**)&pHeaderObj);
	if (pHeaderObj != NULL)
	{
		/*
		 * REQUIRED Property: "StreamCount"
		 * Number of streams contained in this file format
		 */
		pHeaderObj->SetPropertyULONG32("StreamCount",  1);

		/*
		 * Optional Property: "OpaqueData"
		 * Any other user defined data 
		 */
		pHeaderObj->SetPropertyBuffer("OpaqueData", pHeaderDataReadFromFile);


		// Step 6: Notify the RMA core that header object is ready
		m_pStatus->FileHeaderReady(status, pHeaderObj);

		// Release the object since we are done with it
		pHeaderObj->Release();
	}
}


/****************************************************************************
 *  IRMAFileFormatObject::GetStreamHeader                    ref:  rmaformt.h
 *
 *  This routine returns to the RMA core an object containing the stream
 *  header information for a particular stream. Several routines are actually
 *  required to complete the process due to the asynchronous nature of the
 *  RMA file system. This method is called (after the file header has been
 *  read) by the RMA core for each stream in the file format.
 */
STDMETHODIMP CMyExampleFileFormat::GetStreamHeader(UINT16 streamNo)
{
	/* Again, the asynchronous nature of the RMA file system somewhat
	 * obscures the sequence of events. The basic procedure is as follows:
	 *
	 *  1. Move to the beginning of the particular stream header       (Seek)
	 *  2. Wait until move done                                    (SeekDone)
	 *  3. Read the stream header data from the file                   (Read)
	 *  4. Wait until read done                                    (ReadDone)
	 *  5. Create new object containing stream header data
	 *                                              (MyCreateStreamHeaderObj)
	 *  6. Notify RMA core when stream header object is ready
	 *                                              (MyCreateStreamHeaderObj)
	 */


	if ((m_State != Ready) || (streamNo != MY_STREAM_NO))
	{
		return PNR_UNEXPECTED;
	}

	/*
	 * Again, since asynchronous, note our state in order to correctly
	 * respond in SeekDone().
	 */
	m_State = GetStreamHeaderSeekPending;

	// Step 1: Move to the beginning of the stream header data in the file
	const BOOL bIsRelative = FALSE;
	m_pFileObj->Seek(MY_STREAM_HEADER_START, bIsRelative);

	// Step 2: Wait until move is done

	// See SeekDone() for next "step" of GetStreamHeader() procedure

	return PNR_OK;
}


/****************************************************************************
 *  CMyExampleFileFormat::MyCreateStreamHeaderObj            ref:  filefmt1.h
 *
 *  This routine creates a "stream header" object and passes it off to the
 *  RMA core, which in turn transports it to the renderer. This object must
 *  contain certain properties required to characterize the stream. Any
 *  additional stream information read from the file can also be placed in
 *  this object. This method is called after the stream header data from the
 *  file has been completely read.
 */
void
CMyExampleFileFormat::MyCreateStreamHeaderObj
(
	PN_RESULT   status,
	IRMABuffer* pStreamHeaderDataReadFromFile
)
{	/*
	 * We are in the process of handling the GetStreamHeader() request...
	 * See GetStreamHeader(), SeekDone(), and ReadDone() for previous steps.
	 */
	m_State = Ready;

	// Step 5: Create new object containing the stream header data
	IRMAValues* pStreamHeaderObj = NULL;
	m_pClassFactory->CreateInstance(CLSID_IRMAValues,
                                     (void**)&pStreamHeaderObj);
	if (pStreamHeaderObj != NULL)
	{
		/*
		 * REQUIRED Properties:
		 */

		//..."StreamNumber": stream number identifying this stream
		pStreamHeaderObj->SetPropertyULONG32("StreamNumber", MY_STREAM_NO);

		//..."AvgBitRate": average number of bits streamed per second
		UINT32 avgBitRate= 8*MY_PACKET_LENGTH*(MY_TIME_BETWEEN_PACKETS/1000);
		pStreamHeaderObj->SetPropertyULONG32("AvgBitRate", avgBitRate);

		//..."Preroll": amount to buffer before starting play
		pStreamHeaderObj->SetPropertyULONG32("Preroll", 0);

		//..."Duration": total duration in milliseconds of this stream
		// Extract the total number of packets from the stream header
		UINT32 totNumPackets  = 0;
		UCHAR* pStreamHdrData = pStreamHeaderDataReadFromFile->GetBuffer();
		if (pStreamHdrData != NULL)
		{
			char streamHeaderStr[ MY_STREAM_HEADER_LENGTH + 1 ];
			::sprintf(streamHeaderStr, "%.*s", MY_STREAM_HEADER_LENGTH,
		              pStreamHdrData);
		              
			char* numPacketsStr = ::strrchr(streamHeaderStr, '=');
			if (numPacketsStr != NULL)
			{
				numPacketsStr++; // Move past the equals sign
				totNumPackets = ::atol(numPacketsStr);
			}
		}
		UINT32 duration = totNumPackets * MY_TIME_BETWEEN_PACKETS;
		pStreamHeaderObj->SetPropertyULONG32("Duration", duration);

		//..."MimeType": this stream's MIME type
		//    This associates this stream type with a particular renderer
		IRMABuffer* pStringObj = NULL;
		m_pClassFactory->CreateInstance(CLSID_IRMABuffer,
                                         (void**)&pStringObj);
		if (pStringObj != NULL)
		{
			char* streamMimeType = MY_STREAM_MIME_TYPE;
			pStringObj->Set((const BYTE*)streamMimeType,
                             strlen(streamMimeType) + 1);
			pStreamHeaderObj->SetPropertyCString("MimeType", pStringObj);
			pStringObj->Release();
		}

		/*
		 * Optional Properties:
		 */

		// ..."OpaqueData": any other user defined data
		pStreamHeaderObj->SetPropertyBuffer("OpaqueData",
		                                     pStreamHeaderDataReadFromFile);


		// Step 6: Notify the RMA core that stream header object is ready
		m_pStatus->StreamHeaderReady(status, pStreamHeaderObj);

		// Release the object since we are done with it
		pStreamHeaderObj->Release();
	}
}


/****************************************************************************
 *  IRMAFileFormatObject::GetPacket                          ref:  rmaformt.h
 *
 *  This routine returns to the RMA core an object containing the packet
 *  data for a particular stream. Several routines are actually required to
 *  complete the process due to the asynchronous nature of the RMA file
 *  system. This method is called by the RMA core each time it needs another
 *  packet.
 */
STDMETHODIMP
CMyExampleFileFormat::GetPacket(UINT16 streamNo)
{
	/* Again, the asynchronous nature of the RMA file system somewhat
	 * obscures the sequence of events. The basic procedure is as follows:
	 *
	 *  1. Move to the beginning of the next packet in the file        (Seek)
	 *  2. Wait until move done                                    (SeekDone)
	 *  3. Read the packet data from the file                          (Read)
	 *  4. Wait until read done                                    (ReadDone)
	 *  5. Create new object containing packet data       (MyCreatePacketObj)
	 *  6. Notify RMA core when packet object is ready    (MyCreatePacketObj)
	 *  7. Notify RMA core when no more packets are avail.(MyCreatePacketObj)
	 */

	if ((m_State != Ready) || (streamNo != MY_STREAM_NO))
	{
		return PNR_UNEXPECTED;
	}
	
	/*
	 * Steps 1 and 2 are called automatically by the RMA core while playing
	 * your RealMedia data type.
	 */
	 
	/*
	 * Since Read() is asynchronous, note our state in order to correctly
	 * respond in ReadDone().
	 */
	m_State = GetPacketReadPending;

	// Step 3: Read the packet data from the file
	m_pFileObj->Read(MY_PACKET_LENGTH);

	// Step 4: Wait until Read() is done

	// See ReadDone() for next "step" of GetPacket() procedure

	return PNR_OK;
}


/****************************************************************************
 *  CMyExampleFileFormat::MyCreatePacketObj                  ref:  filefmt1.h
 *
 *  This routine creates a packet object and passes it off to the RMA core,
 *  which in turn transports it to the renderer. The object contains the
 *  packet data read from the file along with the time when it should be
 *  delivered to the renderer. If there are no more packets available for
 *  this stream, the RMA core is notified. This method is called after the
 *  packet data from the file has been read.
 */
void
CMyExampleFileFormat::MyCreatePacketObj
(
	PN_RESULT   status,
	IRMABuffer* pPacketDataReadFromFile
)
{	/*
	 * We are in the process of handling the GetPacket() request...
	 * See GetPacket() and ReadDone() for previous steps.
	 */
	m_State = Ready;

	if (status == PN_STATUS_OK)
	{
		// Step 5: Create new object containing the packet data
		IRMAPacket* pPacketObj = NULL;
		m_pClassFactory->CreateInstance(CLSID_IRMAPacket,
                                         (void**)&pPacketObj);
		if (pPacketObj != NULL)
		{
			// Fill in the packet
			UINT32 deliveryTime = m_NextPacketDeliveryTime;
			UINT16 streamNo	    = MY_STREAM_NO;
			UINT8  ASMFlags     = RMA_ASM_SWITCH_ON;
			UINT16 ASMRuleNo    = 0;
			
			pPacketObj->Set(pPacketDataReadFromFile, deliveryTime, streamNo,
                            ASMFlags, ASMRuleNo);
			/* 
			 * Set the delivery time for the next packet. Note that each
			 * packet is stamped with a delivery time which is used to pace
			 * the delivery of packets to the rendering plug-ins.
			 */
			m_NextPacketDeliveryTime += MY_TIME_BETWEEN_PACKETS;


			// Step 6: Notify the RMA core that the packet object is ready
			m_pStatus->PacketReady(status, pPacketObj);

			// Release the object since we are done with it
			pPacketObj->Release();
		}
	}
	else	// No more packets are available for this stream
	{
		m_NextPacketDeliveryTime = 0;

		// Step 7: Notify the RMA core that the stream is done
		m_pStatus->StreamDone(0);
	}
}


/****************************************************************************
 *  IRMAFileResponse::SeekDone                               ref:  rmafiles.h
 *
 *  This routine is automatically called when the Seek() operation in the
 *  file is complete. Other actions are then taken dependent on the current
 *  processing state.
 */
STDMETHODIMP
CMyExampleFileFormat::SeekDone(PN_RESULT status)
{
	switch (m_State)
	{
		case SeekSeekPending:
		{
			// We are in the process of handling a Seek() request...
			m_State = Ready;

			// Notify the RMA core that the Seek() has completed
			m_pStatus->SeekDone(status);
		}
		break;

		case GetFileHeaderSeekPending:
		{	/*
			 * We are in the process of handling the GetFileHeader() request...
			 * See GetFileHeader() for previous steps.

			 * The Seek() has completed, indicating that we have moved to the
			 * start of the header data, and are now ready to read the header.
			 */

			// Again, since Read() is asynchronous, set up our state
			m_State = GetFileHeaderReadPending;

			// Step 3: Read the file header data from the file
			m_pFileObj->Read(MY_FILE_HEADER_LENGTH);

			// Step 4: Wait until read is done

			// See ReadDone() for next "step" of GetFileHeader() procedure
		}
		break;

		case GetStreamHeaderSeekPending:
		{	/* We are in the process of handling the GetStreamHeader request.
			 * See GetStreamHeader() for previous steps.

			 * The Seek() has completed, indicating that we have moved to the
			 * start of the stream header data, and are now ready to read the
			 * header.
			 */

			// Again, since Read() is asynchronous, set up our state
			 m_State = GetStreamHeaderReadPending;

			// Step 3: Read the stream header data from the file
			m_pFileObj->Read(MY_STREAM_HEADER_LENGTH);

			// Step 4: Wait until read is done

			// See ReadDone() for next "step" of GetStreamHeader() procedure
		}
		break;

		default:
		{
			return PNR_UNEXPECTED;
		}
	}

	return PNR_OK;
}


/****************************************************************************
 *  IRMAFileResponse::ReadDone                               ref:  rmafiles.h
 *
 *  This routine is automatically called when the Read() from the file is
 *  done. Other actions are then taken dependent on the current processing
 *  state.
 */
STDMETHODIMP
CMyExampleFileFormat::ReadDone
(
	PN_RESULT   status,
	IRMABuffer* pBufferRead
)
{
	switch (m_State)
	{
		case GetFileHeaderReadPending:
		{	/*
			 * We are in the process of handling the GetFileHeader() request.
			 * See GetFileHeader() and SeekDone() for previous steps.
			 */

			/*
			 * The Read() has completed, and we are now ready to create on
			 * object containing the header data.
			 */
			MyCreateFileHeaderObj(status, pBufferRead);
		}
		break;

		case GetStreamHeaderReadPending:
		{	/*
			 * We are in the process of handling the GetStreamHeader request.
			 * See GetStreamHeader() and SeekDone() for previous steps.
			 */

			/* The Read() has completed, and we are now ready to create on
			 * object containing the stream header data.
			 */
			MyCreateStreamHeaderObj(status, pBufferRead);
		}
		break;

		case GetPacketReadPending:
		{	/*
			 * We are in the process of handling the GetPacket() request.
			 * See GetPacket() for previous steps.
			 */

			/* The Read() has completed, and we are now ready to create on
			 * object containing the packet data.
			 */
			MyCreatePacketObj(status, pBufferRead);
		}
		break;

		default:
		{
			return PNR_UNEXPECTED;
		}
		break;
	}

	return PNR_OK;
}


/****************************************************************************
 *  IRMAFileFormatObject::Seek                               ref:  rmaformt.h
 *
 *  This routine moves to the packet nearest to the requested time in the
 *  file. It is called, for example, in response to the user moving the
 *  slider to a different location in the playback timeline.
 */
STDMETHODIMP
CMyExampleFileFormat::Seek(UINT32 requestedTime)
{

	// Find nearest packet to the requested time
	UINT32 packetNo    = requestedTime / MY_TIME_BETWEEN_PACKETS;
	UINT32 packetStart = (packetNo * MY_PACKET_LENGTH) +
	                     MY_PACKET_DATA_START;

	// Update packet's delivery time stamp
	m_NextPacketDeliveryTime = packetNo * MY_TIME_BETWEEN_PACKETS;

	// Note state, since Seek() is asynchronous
	m_State = SeekSeekPending;

	// Move to the requested location in the file
	const BOOL bIsRelative = FALSE;
	m_pFileObj->Seek(packetStart, bIsRelative);

	// Wait until move is done

	// See SeekDone() for next "step" of the Seek() process.

	return PNR_OK;
}


/****************************************************************************
 *  IRMAFileFormatObject::Close                              ref:  rmaformt.h
 *
 *  This routine handles any cleanup required before closing the plug-in. All
 *  references to objects should be released and memory deallocated. This is
 *  called by the RMA core when the playback is finished or stopped.
 */
STDMETHODIMP
CMyExampleFileFormat::Close(void)
{
	if (m_pFileObj != NULL)
	{
		m_pFileObj->Close();
		m_pFileObj->Release();
		m_pFileObj = NULL;
	}

	if (m_pStatus != NULL)
	{
		m_pStatus->Release();
		m_pStatus = NULL;
	}

	if  (m_pClassFactory != NULL)
	{
		m_pClassFactory->Release();
		m_pClassFactory = NULL;
	}

	return PNR_OK;
}


/****************************************************************************
 *  IRMAFileResponse::CloseDone, WriteDone                   ref:  rmafiles.h
 *
 *  This routines are not used in this example, but are required to complete
 *  the IRMAFileResponse interface.
 */
STDMETHODIMP CMyExampleFileFormat::CloseDone(PN_RESULT /* result */)
{
	return PNR_OK;
}

STDMETHODIMP CMyExampleFileFormat::WriteDone(PN_RESULT /* result */)
{
	return PNR_OK;
}


/****************************************************************************
 *  CMyExampleFileFormat::~CMyExampleFileFormat              ref:  filefmt1.h
 *
 *  Destructor
 */
CMyExampleFileFormat::~CMyExampleFileFormat(void)
{
	Close();
}


// IUnknown COM Interface Methods

/****************************************************************************
 *  IUnknown::AddRef                                            ref:  pncom.h
 *
 *  This routine increases the object reference count in a thread safe
 *  manner. The reference count is used to manage the lifetime of an object.
 *  This method must be explicitly called by the user whenever a new
 *  reference to an object is used.
 */
STDMETHODIMP_(UINT32)
CMyExampleFileFormat::AddRef(void)
{
	return InterlockedIncrement(&m_RefCount);
}


/****************************************************************************
 *  IUnknown::Release                                           ref:  pncom.h
 *
 *  This routine decreases the object reference count in a thread safe
 *  manner, and deletes the object if no more references to it exist. It must
 *  be called explicitly by the user whenever an object is no longer needed.
 */
STDMETHODIMP_(UINT32)
CMyExampleFileFormat::Release(void)
{
	if (InterlockedDecrement(&m_RefCount) > 0)
	{
		return m_RefCount;
	}

	delete this;
	return 0;
}


/****************************************************************************
 *  IUnknown::QueryInterface                                    ref:  pncom.h
 *
 *  This routine indicates which interfaces this object supports. If a given
 *  interface is supported, the object's reference count is incremented, and
 *  a reference to that interface is returned. Otherwise a NULL object and
 *  error code are returned. This method is called by other objects to
 *  discover the functionality of this object.
 */
STDMETHODIMP
CMyExampleFileFormat::QueryInterface
(
	REFIID interfaceID,
	void** ppInterfaceObj
)
{
	// By definition all COM objects support the IUnknown interface
	if (IsEqualIID(interfaceID, IID_IUnknown))
	{
		AddRef();
		*ppInterfaceObj = (IUnknown*)(IRMAPlugin*)this;
		return PNR_OK;
	}

	// IRMAPlugin interface is supported
	else if (IsEqualIID(interfaceID, IID_IRMAPlugin))
	{
		AddRef();
		*ppInterfaceObj = (IRMAPlugin*)this;
		return PNR_OK;
	}

	// IRMAFileResponse interface is supported
	else if (IsEqualIID(interfaceID, IID_IRMAFileResponse))
	{
		AddRef();
		*ppInterfaceObj = (IRMAFileResponse*)this;
		return PNR_OK;
	}

	// IRMAFileFormatObject interface is supported
	else if (IsEqualIID(interfaceID, IID_IRMAFileFormatObject))
	{
		AddRef();
		*ppInterfaceObj = (IRMAFileFormatObject*)this;
		return PNR_OK;
	}

	// No other interfaces are supported
	*ppInterfaceObj = NULL;
	return PNR_NOINTERFACE;
}
